/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
/** 
    \mainpage O2scl User's Guide

    \htmlonly
    <div align="right">
    <a href="dl_page.html"><b>Download o2scl
    </b></a></div>
    \endhtmlonly

    \o2 is a C++ class library for object-oriented numerical
    programming. It includes
    - Classes based on numerical routines from GSL and CERNLIB
    - Vector and matrix classes which are fully compatible with
    \c gsl_vector and \c gsl_matrix, yet offer indexing with
    \c operator[] and other object-oriented features
    - The CERNLIB-based classes are rewritten in C++ and are often 
    faster than their GSL counterparts
    - Classes which require function inputs are designed to accept
    (public or private) member functions, even if they are virtual.
    - Classes use templated vector types, which allow the use of 
    object-oriented vectors or C-style arrays. Because of this,
    the \o2 versions of GSL algorithms can be signficantly faster
    than the original.
    - Highly compatible - Recent versions have been tested on 
    Linux (32- and 64-bit systems, with Intel and AMD chips), 
    Windows XP with Cygwin, and MacOSX.
    - Free! \o2 is provided under Version 3 of the GNU 
    Public License
    - Two mini-libraries
    <ul>
    <li>Thermodynamics of ideal and nearly-ideal particles with
    quantum statistics</li>
    <li>Equations of state for finite density relevant for 
    neutron stars</li>
    </ul>

    See licensing information at \ref license_section. 

    This is a \b pre- \f$ \beta \f$ \b version. While I have released
    it to anyone who finds it useful, there may still be some bugs
    lurking around. I use many of these classes myself frequently and
    most things have been tested.

    \htmlonly
    <hr />
    <a href="dl_page.html">Download o2scl</a>
    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
    <a href="../latex/refman.pdf">PDF documentation</a>
    \endhtmlonly

    \hline
    \section toc_section Quick Reference to User's Guide
    
    - \ref install_section
    - \ref usage_section
    - \ref examples_section
    - \ref cplxnum_section
    - \ref vecmat_section
    - \ref permute_section
    - \ref intp_section
    - \ref const_section
    - \ref funct_section
    - \ref table_section
    - \ref string_section
    - \ref diff_section
    - \ref inte_section
    - \ref poly_section
    - \ref solve_section
    - \ref min_section
    - \ref conmin_section
    - \ref mcarlo_section
    - \ref anneal_section
    - \ref fit_section
    - \ref ode_section
    - \ref rng_section
    - \ref tintp_section
    - \ref other_section
    - \ref lset_section
    - \ref io_section
    - \ref design_section
    - \ref license_section
    - \ref ack_section
    - \ref ref_section
    - \ref todo 
    - \ref bug

    \htmlonly
    <UL><LI>Other mini-libraries
    <UL>
    <LI><B><a href="../part/html/index.html">Particles</a></b>
    <LI><B><a href="../eos/html/index.html">Equations of State</a></b>
    </UL>
    </LI>
    <br />
    <li>Add-on library: There is also a related library,
    <a href="../../o2scl_ext/html/index.html">O2scl_ext</a>,
    which has a separate source distribution.
    </li>
    </UL>
    \endhtmlonly

    \hline
    \section motiv_section Motivation

    I wanted a library for numerical programming in C++. This
    immediately suggests the following question: Why C++? Part of the
    answer to this question is practical; I am much more familiar with
    C/C++ so learning a different language (e.g. Fortran) never seemed
    like it made much sense. A related question might be: Why not an
    interpreted language, like python? Answer: faster
    execution. Finally, why C++ and not C? Answer: The object-oriented
    approach arguably allows for simpler and faster development while
    not appreciably removing flexibility. The object-orientation also
    encourages library development.  Furthermore, I find the syntactic
    simplicity provided by C++ to be very convient at times.

    In regards to the GNU Scientific Library (GSL), GSL is not simple
    to use for C++ development because it makes it very difficult to
    utilize member functions in those routines which take functions as
    inputs.

    I have also allowed the classes to hide their memory allocation
    structure, even though the user is frequently allowed to specify
    when it is allocated or deallocated. In light of the philosophy of
    'encapsulation' this makes sense, as the memory requirements for
    the classes are not generic. Users who do not know about the
    details of the algorithm need not know about the details of the
    memory requirements, and users who need to know about the details
    of the memory requirements most often also want to know the
    details about the algorithm.

    \hline
    \section install_section Installation

    The rules for installation are generally the same as that for
    other GNU libraries. The file \c INSTALL has some details on this
    procedure. Generally, you should be able to run \c ./configure and
    then type \c make and \c make \c install. The documentation is
    automatically installed by \c make \c install. 

    After \c make \c install, you may test the library by \c make \c
    o2scl-test. This should output a summary which is reproduced in
    the file \c test-summary.text. If the compilation of the test
    programs doesn't work, you may have to add the correct directory
    to the \c LD_LIBRARY_PATH environment variable between
    installation and testing.

    This library requires GSL and is designed to work with GSL
    versions 1.9 or 1.10. Some classes may work with older versions of GSL,
    but this cannot be guaranteed. 

    Range-checking for vectors and matrices is performed similar to
    the GSL approach, and is turned off by default. You can 
    enable range-checking by defining \c GSL_RANGE_CHECK=1, e.g.
    \code
    CPPFLAGS="-DGSL_RANGE_CHECK=1" ./configure
    \endcode

    The separate libraries o2scl_eos and o2scl_part are installed
    by default. To disable these libraries, configure with 
    \c --disable-eoslib or \c --disable-partlib.
    Note that o2scl_eos depends on o2scl_part so using \c --disable-partlib
    without \c --disable-eoslib may not work. 

    There are several warning flags that are useful when configuring
    and compiling with \c g++. (See the GSL documentation for an 
    excellent discussion.) For running \c configure, I use
    something like
    \code
    CFLAGS="" CXXFLAGS="" CPPFLAGS="-ansi -pedantic -Wall -W           \
    -Wconversion -Wno-unused -Wshadow -Wpointer-arith -Wcast-align     \
    -Wwrite-strings -fshort-enums -ggdb -O3 -DGSL_RANGE_CHECK=0	       \
    -DHAVE_INLINE -I/home/asteiner/install/include"                    \
    LDFLAGS="-L/home/asteiner/install/lib" ./configure -C	       \
    --enable-readline --prefix=/home/asteiner/install
    \endcode
    with the references to the directory \c /home/asteiner/install 
    in order to install somewhere in my user directory, and because
    my copy of GSL is installed there, rather than in \c /usr/local.
    
    The documentation is generated with Doxygen. In principle, the 
    documentation can be regenerated by the end-user, but this is 
    not supported and may require several external applications
    not included in the distribution.

    \b Un-installation: While there is no explicit "uninstall"
    procedure, there are only a couple places to check. Installation
    creates directories named \c o2scl in the include, doc and shared
    files directory (which default to \c /usr/local/include, \c
    /usr/local/doc, and \c /usr/local/share) which can be
    removed. Finally, all of the libraries are named with the prefix
    \c libo2scl and are created by default in /usr/local/lib. As
    configured with the settings above, the files are in \c
    /home/asteiner/install/include/o2scl, \c /home/asteiner/install/lib,
    \c /home/asteiner/install/share/o2scl, and \c
    /home/asteiner/install/doc/o2scl.

    \hline
    \section usage_section General Usage

    \b Namespaces
    
    Most of the classes reside in the namespace \c o2scl (removed from
    the documentation). Numerical and physical constants (many of them
    based on the GSL constants) are placed in separate
    namespaces. There are a couple other (mostly internal) namespaces
    which are documented separately.

    \b Documentation \b conventions

    In the following documentation, function parameters are denoted by
    \c parameter, except when used in mathematical formulas as in \f$
    \mathrm{variable} \f$ .
    
    \b Nomenclature

    Classes directly derived from the GNU Scientific Library are
    preceeded by the prefix \c gsl_ and classes derived from CERNLIB
    are preceeded by the prefix \c cern_. Some of those classes 
    derived from GSL and CERNLIB operate slightly differently 
    from the original versions. The differences are detailed
    in the corresponding class documentation.
    
    \b Error \b handling

    Error handling is GSL-like with functions returning 0 for success
    and calling a GSL-like error handler. The error handler, 
    \ref err_hnd is a global pointer to an object of type \ref err_class.
    
    The default behaviour for all errors is simply store the error
    code and information. You can require the default error handler to
    print out any error (or even exit) using \ref
    err_class::set_mode(). Also, if O2SCL_ARRAY_ABORT is defined,
    then \c exit() will be called any time a \ref gsl_index error is
    called, e.g. an out-of-bounds array access.

    Errors can be set through the macros \ref set_err, \ref set_err_ret, 
    \ref add_err, and \ref add_err_ret, which are defined in the
    file err_hnd.h.

    Functionality similar to assert() is provided with the macro
    \ref err_assert, which exits if its argument is non-zero, and 
    \ref bool_assert which exits if its argument is false.

    Note that the library functions do not typically try to 
    reset the error handler. For this reason the user may 
    need to reset the error handler before calling functions 
    which do not return an integer error value so that they
    can easily test to see if an error occured, e.g. using
    err_hnd::get_errno().

    The default error handler can be replaced by simply assigning
    the address of a descendant of \ref err_class to \ref err_hnd.

    \b Library \b dependencies

    All of the libraries need \c libo2scl_base, \c libgsl, and either \c
    libgslcblas or some externally-specified \c libcblas to
    operate. Other additional dependencies are listed below:
    - \c libo2scl_eos needs \c libo2scl_nuclei, \c libo2scl_part, 
    \c libo2scl_other, \c libo2scl_minimize, \c libo2scl_inte, and 
    \c libo2scl_root
    - \c libo2scl_nuclei needs \c libo2scl_part, \c libo2scl_other, 
    \c libo2scl_inte, and \c libo2scl_root.
    - \c libo2scl_part needs \c libo2scl_other, \c libo2scl_inte, and 
    \c libo2scl_root.

    \hline
    \section examples_section Example programs
    
    A few example programs are in the \c examples directory.
    After installation, they can be compiled and 
    executed by running \c make \c o2scl-examples in that directory. 

    Also, the testing code for each class is frequently useful for
    providing examples of their usage.  The testing source code for
    each source file is named with an \c _ts.cpp prefix in the same
    directory as the class source.
      
    \hline
    \section cplxnum_section Complex Numbers
    
    Some rudimentary arithmetic operators for \c gsl_complex are
    defined in cx_arith.h, but no constructor has been written. The
    object \c gsl_complex is still a \c struct, not a \c class. For
    example,
    \code
    gsl_complex a={{1,2}}, b={{3,4}};
    gsl_complex c=a+b;
    cout << GSL_REAL(c) << " " << GSL_IMAG(C) << endl;
    \endcode
    In case the user needs to convert between \c gsl_complex and \c
    std::complex<double>, two conversion functions gsl_to_complex()
    and complex_to_gsl() are provided in \ref ovector_cx_tlate.h.

    \hline
    \section vecmat_section Vectors, Matrices and Tensors

    \b Introduction

    Vectors and matrices are designed using the templates \ref
    ovector_tlate and \ref omatrix_tlate, which are compatible with \c
    gsl_vector and \c gsl_matrix. Vectors and matrices with unit
    stride are provided in \ref uvector_tlate and \ref
    umatrix_tlate. The most commonly used double-precision versions of
    these template classes are \ref ovector, \ref omatrix, \ref
    uvector and \ref umatrix, and their associated "views" (analogous
    to GSL vector and matrix views) which are named with a \c _view
    suffix. The classes \ref ovector_tlate and \ref omatrix_tlate
    offer the syntactic simplicity of array-like indexing, which is
    easy to apply to vectors and matrices created with GSL.

    The following sections primarily discuss the operation objects of
    type \ref ovector. The generalizations to objects of type \ref
    uvector, \ref omatrix, and the complex vector and matrix objects
    \ref ovector_cx, \ref omatrix_cx, \ref uvector_cx, and \ref
    umatrix_cx are straightforward.

    \b Views

    Vector and matrix views are provided as parents of the vector and
    matrix classes which do not have methods for memory allocation. As
    in GSL, it is simple to "view" normal C-style arrays or pointer
    arrays (see \ref ovector_array), parts of vectors (\ref
    ovector_subvector), and rows and columns of matrices (\ref
    omatrix_row and \ref omatrix_col). Several operations are defined,
    including addition, subtraction, and dot products.

    \b Typedefs

    Several typedefs are used to give smaller names to often used
    templates. Vectors and matrices of double-precision numbers all
    begin with the prefixes \ref ovector and \ref omatrix. Integer
    versions begin with the prefixes \ref ovector_int and \ref
    omatrix_int. Complex versions have an additional \c "_cx",
    e.g. \ref ovector_cx. See \ref ovector_tlate.h, \ref
    ovector_cx_tlate.h, \ref omatrix_tlate.h, and \ref
    omatrix_cx_tlate.h.

    \b Unit-stride \b vectors

    The \ref uvector_tlate objects are naturally somewhat faster
    albeit less flexible than their finite-stride counterparts.
    Conversion to GSL vectors and matrices is not trivial for \ref
    uvector_tlate objects, but demands copying the entire
    vector. Vector views, operators, and the nomenclature is similar
    to that of \ref ovector. 

    These unit-stride vectors are very useful to help optimize the
    internal workings of functions that do not need a finite stride.

    \b Memory \b allocation

    Memory for vectors can be allocated using ovector::allocate() and
    ovector::free(). Allocation can also be performed by the
    constructor, and the destructor automatically calls free() if
    necessary. In contrast to \c gsl_vector_alloc(),
    ovector::allocate() will call ovector::free(), if necessary, to
    free previously allocated space. Allocating memory does not clear
    the recently allocated memory to zero.  You can use \ref
    ovector::set_all() with an argument of \c 0.0 to clear a vector
    (and similarly for a matrix).

    If the memory allocation fails, either in the constructor or in
    allocate(), then the error handler will be called, partially
    allocated memory will be freed, and the size will be reset to
    zero. You can either use the error handler or test to see if the
    allocation succeeded by examining the size argument, e.g.
    \code
    const size_t n=10;
    ovector x(10);
    if (x.size()==0) cout << "Failed." << endl;
    \endcode
    
    \b Get \b and \b set
    
    Vectors and matrices can be modified using ovector::get() and
    ovector::set() methods analogous to \c gsl_vector_get() and \c
    gsl_vector_set(), or they can be modified through
    ovector::operator[] (or ovector::operator() ), e.g.
    \code
    ovector a(4);
    a.set(0,2.0);
    a.set(1,3.0);
    a[2]=4.0;
    a[3]=2.0*a[1];
    \endcode
    
    If you want to set all of the values in an \ref ovector or an \ref
    omatrix at the same time, then use ovector::set_all() or
    omatrix::set_all().

    \b Range \b checking
    
    Range checking is performed depending on whether or not \c
    O2SCL_NO_RANGE_CHECK is defined. It can be defined in the
    arguments to \c ./configure upon installation to turn off range
    checking. Note that this is completely separate from the GSL range
    checking mechanism, so range checking may be on in \o2 even if
    it has been turned off in GSL.  Range checking is used primarily
    in the vector, matrix, and tensor \c get() and \c set() methods.

    To see if range checking was turned on during installation
    (without calling the error handler), use
    lib_settings_class::range_check().

    Note that range checking in \o2 code is present in header files,
    rather than in source code. This means that range checking can be
    turned on or off in user-defined functions separately from whether
    or not it was used in the library classes and functions.

    \b Shallow \b and \b deep \b copy

    Copying \o2 vectors using constructors or the \c = operator
    is performed according to what kind of object is
    on the left-hand side (LHS) of the equals sign. If the LHS is a
    view, then a shallow copy is performed, and if the LHS is a \ref
    ovector, then a deep copy is performed. If an attempt is made to
    perform a deep copy onto a vector that has already been allocated,
    then that previously allocated memory is automatically freed. 
    The user must be careful to ensure that information is not lost
    this way, even though no memory leak will occur.
    
    For generic deep vector and matrix copying, you can use the
    template functions \ref vector_copy(), \ref matrix_copy(), \ref
    vector_cx_copy(), and \ref matrix_cx_copy() defined in the 
    o2scl_arith namespace. These would allow you, for example, to 
    copy an \ref ovector to a \c std::vector<double> object (assuming
    the memory allocation has already been taken care of). 

    \b Compatibility \b with \b MV++
    
    There is also a moderate amount of compatibility between
    \ref ovector_tlate, \ref ovector_view_tlate, \ref uvector_tlate, 
    \c std::vector<> and the vectors from \c MV++. They all offer
    - A blank constructor to create an empty vector
    - A constructor with a "size" argument (\c size_t or \c int) to
    create a vector with a specified size.
    - \c operator=() - Copy constructor
    - \c operator[] - Array-indexing
    - \c size() - Get the "size" of the vector

    There is a slight difference between how this works in comparison
    to MV++. The function allocate() operates a little differently
    than newsize(), as it will feel free to allocate new memory when
    owner is false, i.e. it will allocate memory for a new vector even
    if it previously pointed to a vector whose memory it did not
    own. This should not be an issue, however, since it is not
    possible to create an \ref ovector_tlate with a value of \c owner
    equal to zero, unless the \ref ovector_tlate class is overloaded
    improperly.
    
    \b Vector \b and \b matrix \b arithmetic
    
    Several operators are available as member functions of the
    corresponding template:
    
    Vector_view unary operators:
    - vector_view += vector_view
    - vector_view -= vector_view
    - vector_view += scalar
    - vector_view -= scalar
    - vector_view *= scalar
    - scalar = norm(vector_view)

    Matrix_view unary operators:
    - matrix += matrix
    - matrix -= matrix
    - matrix += scalar
    - matrix -= scalar
    - matrix *= scalar

    Binary operators like addition, subtraction, and matrix
    multiplication are also defined for \ref ovector, \ref uvector,
    and related objects in the \ref o2scl_arith namespace. The generic
    template for a binary operator, e.g.
    \code
    template<class vec_t> vec_t &operator+(vec_t &v1, vec_t &v2);
    \endcode
    is difficult because the compiler has no way of distinguishing vector 
    and non-vector classes. At the moment, this is solved by 
    creating a define macro for the binary operators. In addition
    to the predefined operators for native classes, the user may 
    also define binary operators for other classes using the same
    macros. For example, 
    \code
    O2SCL_OP_VEC_VEC_ADD(o2scl::ovector,std::vector<double>,
    std::vector<double>)
    \endcode
    would provide an addition operator for \ref ovector and vectors
    from the Standard Template Library. The macros are detailed
    in \ref vec_arith.h.

    The GSL BLAS routines can also be used directly with \ref ovector
    and \ref omatrix objects.

    Note that some of these arithmetic operations succeed even with
    non-matching vector and matrix sizes. For example, adding a 3x3
    matrix to a 4x4 matrix will result in a 3x3 matrix and the 7 outer
    elements of the 4x4 matrix are ignored.

    \b Converting \b to \b and \b from \b GSL \b forms

    Because of the way \ref ovector is constructed, you may use
    type conversion to convert to and from objects of type \c gsl_vector.
    \code
    ovector a(2);
    a[0]=1.0;
    a[1]=2.0;
    gsl_vector *g=(gsl_vector *)(&a);
    cout << gsl_vector_get(g,0) << " " << gsl_vector_get(g,1) << endl;
    \endcode
    Or,
    \code
    gsl_vector *g=gsl_vector_alloc(2);
    gsl_vector_set(0,1.0);
    gsl_vector_set(1,2.0);
    ovector &a=(ovector &)(*g);
    cout << a[0] << " " << a[1] << endl;
    \endcode
    This sort of type-casting is discouraged among unrelated classes,
    but is permissible here because ovector_tlate is a descendant of
    gsl_vector.  In particular, this will not generate "type-punning"
    warnings in later gcc versions.  If this bothers your
    sensibilities, however, then you can use the following approach:
    \code
    ovector a(2);
    gsl_vector *g=a.get_gsl_vector();
    \endcode
    The ease of converting between these two kind of objects makes it
    easy to use gsl functions on objects of type \ref ovector, i.e.
    \code
    ovector a(2);
    a[0]=2.0;
    a[1]=1.0;
    gsl_vector_sort((gsl_vector *)(&a));
    cout << a[0] << " " << a[1] << endl;
    \endcode

    \b Converting \b from \b STL \b form

    To "view" a \c std::vector<double>, you can use \ref ovector_array
    \code
    std::vector<double> d;
    d.push_back(1.0);
    d.push_back(3.0);
    ovector_array aa(d.size,&(d[0]));
    cout << aa[0] << " " << aa[1] << endl;
    \endcode

    However, you should note that if the memory for the \c std::vector
    is reallocated (for example because of a call to \c push_back()),
    then a previously created \ref ovector_view will be incorrect.

    \b push_back() \b and \b pop() \b methods
    
    These two functions give a behavior similar to the corresponding
    methods for \c std::vector<>. This will work in \o2 classes,
    but may not be compatible with all of the GSL functions. This 
    will break if the address of a \ref ovector_tlate is given
    to a GSL function which accesses the \c block->size parameter
    instead of the \c size parameter of a \c gsl_vector. Please 
    contact the author of \o2 if you find a GSL function with this
    behavior. 
    
    \b Views

    Views are slightly different than in GSL in that they are now
    implemented as parent classes. The code
    \code
    double x[2]={1.0,2.0};
    gsl_vector_view_array v(2,x);
    gsl_vector *g=&(v.vector);
    gsl_vector_set(g,0,3.0);
    cout << gsl_vector_get(g,0) << " " << gsl_vector_get(g,1) << endl;
    \endcode
    can be replaced by
    \code 
    double x[2]={1.0,2.0};
    ovector_array a(2,x);
    a[0]=3.0;
    cout << a[0] << " " << a[1] << endl;
    \endcode

    \b Passing \b ovector \b parameters
    
    It is often best to pass an \ref ovector as a const reference
    to an \ref ovector_view, i.e.
    \code
    void function(const ovector_view &a);
    \endcode
    If the function may change the values in the \ref ovector,
    then just leave out \c const
    \code
    void function(ovector_view &a);
    \endcode
    This way, you ensure that the function is not allowed to modify
    the memory for the vector argument.

    If you intend for a function (rather than the user) to handle 
    the memory allocation, then some care is necessary. The
    following code
    \code
    class my_class {
    int afunction(ovector &a) {
    a.allocate(1);
    // do something with a
    return 0;
    }
    };
    \endcode
    is confusing because the user may have already allocated memory
    for \c a. To avoid this, you may want to ensure that the user
    sends an empty vector. For example,
    \code
    class my_class {
    int afunction(ovector &a) {
    if (a.get_size()>0 && a.is_owner()==true) {
    set_err("Unallocated vector not sent.",1);
    return 1;
    } else {
    a.allocate(1);
    // do something with a
    return 0;
    }
    }
    };
    \endcode
    In lieu of this, it is often preferable to use a local variable for
    the storage and offer a \c get() function, 
    \code
    class my_class {
    protected:
    ovector a;
    public:
    int afunction() {
    a.allocate(1);
    // do something with a
    return 0;
    }
    int get_result(const ovector_view &av) { av=a; return 0; }
    };
    \endcode
    The \o2 classes run into this situation quite frequently, but
    the vector type is specified through a template
    \code
    template<class vec_t> class my_class {
    protected:
    vec_t a;
    public:
    int afunction(vec_t &a) {
    a.allocate(1);
    // do something with a
    return 0;
    }
    };
    \endcode

    \b Vectors \b and \b operator=()

    An "operator=(value)" method for setting all vector elements to
    the same value is not included because it leads to confusion
    between, \c ovector_tlate::operator=(const data_t &val) and
    ovector_tlate::ovector_tlate(size_t val) For example, after
    implementing \c operator=() and executing the following
    \code
    ovector_int o1=2;
    ovector_int o2;
    o2=2;
    \endcode
    \c o1 will be a vector of size two, and \c o2 will be an
    empty vector! 

    To set all of the vector elements to the same value, use
    ovector_tlate::set_all().  Because of the existence of
    constructors like ovector_tlate::ovector_tlate(size_t val), the
    following code
    \code
    ovector_int o1=2;
    \endcode
    still compiles, and is equivalent to 
    \code
    ovector_int o1(2);
    \endcode
    while the code
    \code
    ovector_int o1;
    o1=2;
    \endcode
    will not compile. As a matter of style, \c ovector_int \c o1(2);
    is preferable to \c ovector_int \c o1=2; to avoid confusion.

    \b Matrix \b structure
    
    The matrices from \ref omatrix_tlate
    are structured in exactly the same way as in GSL.
    For a matrix with 2 rows, 4 colums, and a "tda" or "trailing
    dimension" of 7, the memory for the matrix is structured in the
    following way:
    
    \verbatim
    00 01 02 03 XX XX XX
    10 11 12 13 XX XX XX
    \endverbatim
    
    where \c XX indicates portions of memory that are unused.  The tda
    can be accessed through, for example, the method
    omatrix_view_tlate::tda(). The \c get(size_t,size_t) methods
    always take the row index as the first argument and the column
    index as the second argument. The matrices from \ref umatrix_tlate
    have a trailing dimension which is always equal to the number of
    columns.

    \b Reversing the order of vectors

    You can get a reversed vector view from \ref
    ovector_reverse_tlate, or uvector_reverse_tlate. For these
    classes, \c operator[] and related methods are redefined to
    perform the reversal. If you want to make many calls to these
    indexing methods for a reversed vector, then simply copying the
    vector to a reversed version may be faster.
    
    \b Const-correctness with vectors
    
    There are several classes named with \c "_const" to provide different
    kinds of const
    views of const vectors. The keyword \c const still ought to be included
    to ensure that the object is treated properly. For example,
    \code
    ovector o(2);
    o[0]=3.0;
    o[1]=-1.0;
    const ovector_const_subvector ocs(o,1,1);
    \endcode
    
    At present, const-correctness in \o2 can be improperly removed,
    if the \c const keyword is not properly included. For example, the
    following code will compile, violated the const-correctness of 
    the \c ocs variable.
    \code
    ovector o(2);
    o[0]=3.0;
    o[1]=-1.0;
    ovector_const_subvector ocs(o,1,1);
    ovector_view ov(ocs);
    ov[0]=2.0;
    \endcode
    
    \b Tensors

    Some preliminary support is provided for tensors of arbitrary rank
    and size in the class \ref tensor. Classes \ref tensor1, \ref
    tensor2, \ref tensor3, and \ref tensor4 are rank-specific versions
    for 1-, 2-, 3- and 4-rank tensors. For n-dimsional data defined on
    a grid, \ref tensor_grid provides a space to define a hyper-cubic
    grid in addition to the the tensor data. This class \ref
    tensor_grid also provides n-dimensional interpolation of the data
    defined on the specified grid.

    \hline
    \section permute_section Permutations

    Permutations are implemented through the \ref permutation class.
    This class is fully compatible with gsl_permutation objects since
    it is inherited from gsl_permutation_struct. The class also
    contains no new data members, so upcasting and downcasting can
    always be performed. It is perfectly permissible to call
    GSL permutation functions from \ref permutation objects by 
    simply passing the address of the permutation, i.e.
    \code
    permutation p(4);
    p.init();
    gsl_permutation_swap(&p,2,3);
    \endcode

    The functions which apply a permutation to a
    user-specified vector are member template functions in the \ref
    permutation class (see \ref permutation::apply() ).

    Memory allocation/deallocation between the class and the
    gsl_struct is compatible in many cases, but mixing these forms is
    strongly discouraged, i.e. avoid using
    <tt>gsl_permutation_alloc()</tt> on a \ref permutation object, but
    rather use \ref permutation::allocate() instead.  The use of \ref
    permutation::free() is encouraged, but any remaining memory is
    deallocated in the object destructor.

    \hline
    \section intp_section Interpolation
    
    The classes o2scl_interp and o2scl_interp_vec allow basic
    interpolation, lookup, differentiation, and integration of data
    given in two ovectors or ovector views. In contrast to the GSL
    routines, data which is presented with a decreasing independent
    variable is handled automatically. For interpolation with
    arrays rather than ovectors, use array_interp or array_interp_vec.

    For fast interpolation without error-checking, you can directly
    use the children of \ref base_interp. 

    \b The \b two \b interpolation \b interfaces

    The difference between the two classes, o2scl_interp and
    o2scl_interp_vec, analogous to the difference between using \c
    gsl_interp_eval() and \c gsl_spline_eval() in \c GSL. You can
    create a o2scl_interp object and use it to interpolate among any
    pair of chosen vectors, i.e.
    \code
    ovector x(20), y(20);
    // fill x and y with data
    o2scl_interp gi;
    double y_o2sclf=gi.interp(0.5,20,x,y);
    \endcode
    Alternatively, you can create a o2scl_interp_vec object which can be
    optimized for a pair of vectors that you specify in advance
    \code
    ovector x(20), y(20);
    // fill x and y with data
    o2scl_interp_vec gi(20,x,y);
    double y_o2sclf=gi.interp(0.5);
    \endcode

    \b Lookup \b and \b binary \b search

    The class search_vec contains a searching functions for objects of
    type \ref ovector which are monotonic. Note that if you want to
    find the index of an \ref ovector where a particular value is
    located without any assumptions with regard to the ordering, you
    can use ovector::lookup() which performs an exhaustive search.

    \b "Smart" \b interpolation
    
    The classes smart_interp and smart_interp_vec allow interpolation,
    lookup, differentiation, and integration of data which is
    non-monotonic or multiply-valued outside the region of interest.
    As with o2scl_interp above, the corresponding array versions are
    given in sma_interp and sma_interp_vec.

    \note The classes smart_interp and smart_interp_vec are still
    a bit experimental. 

    \b Two \b and \b higher \b dimensional \b interpolation

    Preliminary support for two-dimensional interpolation is 
    given in \ref twod_intp, and n-dimensional interpolation
    in \ref tensor_grid.

    \hline
    \section const_section Physical constants
    
    The constants from GSL are reworked with the type \c const \c
    double and placed in namespaces called gsl_cgs, gsl_cgsm, gsl_mks,
    gsl_mksa, and gsl_num. Some additional constants are given in the
    namespace o2scl_const.

    \hline
    \section funct_section Function Objects

    Functions are passed to numerical routines using template-based
    function classes. There are several basic "kinds" of function
    objects:
    - \ref funct : One function of one variable
    - \ref multi_funct : One function of several variables
    - \ref mm_funct : \c n functions of \c n variables
    - \ref fit_funct : One function of one variable with \c n 
    fitting parameters
    - \ref ode_funct : \c n derivatives as a function of \c n function
    values and the value of the independent variable

    For each of these classes, there is a version named \c _vfunct
    instead of \c _funct which is designed to be used with C-style
    arrays instead. 

    The class name suffixes denote children of a generic function type
    which are created using different kinds of inputs:
    - _fptr: function pointer for a static or global function
    - _gsl: GSL-like function pointer 
    - _mfptr: function pointer template for a class member function
    - _strings: functions specified using strings, e.g. "x^2-2"
    - _noerr: (for \ref funct and multi_funct) a function which 
    directly returns the function value rather than returning
    an integer error value
    - _nopar: (for \ref funct) a function which has no parameters
    specified by a \c void \c * and directly returns the function
    value.

    There is a small overhead associated with the indirection: a
    "user" accesses the function class which then calls function which
    was specified in the constructor of the function class.  This
    overhead can always be avoided by inheriting directly from the
    function class and thus the user will make a direct call, or by
    specifying a new type for the template parameter in the class
    which will call the user-specified function. In many problems,
    the overhead associated with the indirection is small. In problems
    where the associated overhead is not small, there are usually
    other optimization issues (such as error handling and bounds
    checking in the "user" of the function object) which are still
    larger. 

    Note that virtual functions can be specified through 
    this mechanism as well. For example, if \ref cern_mroot
    is used to solve a set of equations specified as
    \code
    class my_type_t {
    virtual member_func();
    };
    my_type_t my_instance;
    class my_derived_type_t : public my_type_t {
    virtual member_func();
    };
    my_derived_type_t my_inst2;
    mm_funct_mfptr<my_type_t> func(&my_inst2,&my_instance::member_func);
    \endcode
    Then the solver will solve the member function in the derived
    type, not the parent type. 

    \hline
    \section table_section Data tables
    
    The class \ref table is a container to hold and perform operations
    on related columns of data. It supports column operations,
    interpolation, column reference by either name or index, binary
    searching (in the case of ordered columns), sorting, and fitting
    two columns to a user-specified function.

    \hline
    \section string_section String manipulation

    There are a couple classes and functions to help manipulate
    strings of text. Conversion routines for \c std::string 
    objects are given in \ref string_conv.h and include
    - \ref ptos() - pointer to string
    - \ref itos() - integer to string
    - \ref dtos() - double to string
    - \ref stoi() - string to integer
    - \ref stod() - string to double

    See also \ref size_of_exponent(), \ref double_to_latex(), 
    \ref double_to_html, and \ref double_to_ieee_string(). 

    There is a class called \ref columnify, which converts a set of
    strings into nicely formatted columns by padding with the
    necessary amount of spaces. This class operates on string objects
    of type \c std::string, and also works will for formatting columns
    of floating-point numbers.  This class is used to provide output
    for matrices in the functions \ref matrix_out(), \ref
    array_2d_out(), and \ref matrix_cx_out(). For output
    of vectors, see \ref vector_out() in \ref array.h. 

    A related function, \ref screenify(), reformats a column 
    of strings into many columns stored row-by-row in a new 
    string array. It operates very similar to
    the way the classic Unix command \c ls organizes files and
    directories in multiple columns in order to save screen space.
    
    The function \ref count_words() counts the number of "words"
    in a string, which are delimited by whitespace. 

    \hline
    \section diff_section Differentiation

    Differentiation is performed by descendants of \ref deriv and the
    classes are provided. These allow one to calculate either first,
    second, and third derivatives. The GSL approach is used in \ref
    gsl_deriv, and the CERNLIB routine is used in \ref
    cern_deriv. Both of these compute derivatives for a function
    specified using a descendant of \ref funct. For functions which
    are tabulated over equally-spaced abscissas, the class \ref
    eqi_deriv is provided which applies the formulas from Abramowitz
    and Stegun at a specified order.

    \b Warning: For gsl_deriv and cern_deriv, the second and third
    derivatives are calculated by naive repeated application of the
    code for the first derivative and can be particularly troublesome
    if the function is not sufficiently smooth. Error estimation is 
    also incorrect for second and third derivatives.

    \hline
    \section inte_section Integration

    Integration is performed by descendants of \ref inte and is
    provided in the library \c o2scl_inte.

    There are several routines for one-dimensional integration.
    - General integration over a finite interval: cern_adapt,
    cern_gauss, cern_gauss56, gsl_inte_qag, and gsl_inte_qng.
    - General integration from 0 to  \f$ \infty \f$ : gsl_inte_qagiu
    - General integration from  \f$ -\infty \f$ to 0: gsl_inte_qagil
    - General integration from  \f$ -\infty \f$ to \f$ \infty \f$ : 
    gsl_inte_qagi
    - Integration for a finite interval over an function with equally
    spaced abscissas provided in an array: eqi_inte
    - General integration over a finite interval for a function with
    singularities: gsl_inte_qags and gsl_inte_qagp
    - Cauchy principal value integration over a finite interval:
    cern_cauchy and gsl_inte_qawc
    - Integration over a function weighted by \c cos(x) or \c sin(x):
    \ref gsl_inte_qawo_cos and \ref gsl_inte_qawo_sin
    - Fourier integrals: \ref gsl_inte_qawf_cos and \ref
    gsl_inte_qawf_sin
    - Integration over a weight function
    \f[
    W(x)=(x-a)^{\alpha}(b-x)^{\beta}\log^{\mu}(x-a)\log^{\nu}(b-x)
    \f]
    is performed by \ref gsl_inte_qaws. 
    
    For the GSL-based integration routines, the variables inte::tolx
    and inte::tolf have the same role as the quantities usually
    denoted in the GSL integration routines by \c epsabs and \c
    epsrel. In particular, the integration classes attempt to ensure
    that
    \f[ 
    |\mathrm{result}-I| \leq \mathrm{Max}(\mathrm{tolx},
    \mathrm{tolf}|I|)
    \f]
    and returns an error to attempt to ensure that
    \f[
    |\mathrm{result}-I| \leq \mathrm{abserr} \leq
    \mathrm{Max}(\mathrm{tolx},\mathrm{tolf}|I|)
    \f]
    where \c I is the integral to be evaluated. Even when the
    corresponding descendant of inte::integ() returns success, these
    inequalities may fail for sufficiently difficult functions.

    The GSL routines were originally based on QUADPACK, which is
    available at http://www.netlib.org/quadpack. 
    
    Multi-dimensional hypercubic integration is performed by \ref
    composite_inte, the sole descendant of multi_inte. \ref
    composite_inte allows you to specify a set of one-dimensional
    integration routines (objects of type \ref inte) and apply them to
    a multi-dimensional problem.

    General multi-dimensional integration is performed by \ref
    comp_gen_inte, the sole descendant of \ref gen_inte.  The user is
    allowed to specify a upper and lower limits which are functions of
    the variables for integrations which have not yet been performed,
    i.e. the n-dimensional integral
    \f[ 
    \int_{x_0=a_0}^{x_0=b_0} f(x_0) \int_{x_1=a_1(x_0)}^{x_1=b_1(x_0)} 
    f(x_0, x_1) ...
    \int_{x_{\mathrm{n}-1}=a_{\mathrm{n}-1}(x_0,x_1,..,x_{\mathrm{n}-2})}^
    {x_{\mathrm{n}-1}=b_{\mathrm{n}-1}(x_0,x_1,..,x_{\mathrm{n}-2})} 
    f(x_0,x_1,...,x_{\mathrm{n-1}})~d x_{\mathrm{n}-1}~...~d x_1~d x_0
    \f]
    Again, one specifies a set of \ref inte objects to apply to
    each variable to be integrated over.

    Monte Carlo integration is also provided (see \ref mcarlo_section).

    \hline
    \section poly_section Roots of Polynomials

    Classes are provided for solving quadratic, cubic, and quartic
    equations as well as general polynomials. There is a standard
    nomenclature: classes which handle polynomials with real
    coefficients and real roots end with the suffix "_real"
    (quadratic_real, cubic_real and quartic_real), classes which
    handle real coefficients and complex roots end with the suffix
    "_real_coeff" (quadratic_real_coeff, cubic_real_coeff,
    quartic_real_coeff, and poly_real_coeff), and classes which handle
    complex polynomials with complex coefficients (quadratic_complex,
    cubic_complex, quartic_complex, and poly_complex). As a reminder,
    complex roots may not occur in conjugate pairs if the coefficients
    are not real. Most of these routines do not separately handle
    cases where the leading coefficient is zero.

    At present, the polynomial routines work with complex numbers as
    objects of type \c std::complex<double> and are located in library
    \c o2scl_other. 

    For quadratics, \ref gsl_quadratic_real_coeff is the best if the
    coefficients are real, while if the coefficients are complex, \ref
    quadratic_std_complex is the only option at present. For cubics
    with real coefficients, \ref cern_cubic_real_coeff is the best,
    while if the coefficients are complex, use \ref cubic_std_complex.
    
    For a quartic polynomial with real coefficients, \ref
    cern_quartic_real_coeff is the best, unless the coefficients of
    odd powers happen to be small, in which case, \ref
    gsl_quartic_real2 tends to work better. For quartics, generic
    polynomial solvers such as \ref gsl_poly_real_coeff can provide
    more accurate (but slower) results. If the coefficients are
    complex, then you can use \ref naive_quartic_complex.

    \hline
    \section solve_section Equation Solving

    Equation solving classes are stored in the library
    \c o2scl_root. 

    \b One-dimensional \b solvers

    Solution of one equation in one variable is accomplished by
    children of the class \ref root. This base class provides
    the structure for three different solving methods:
    - root::solve(double &x, void *pa, funct &func) which
    solves a function given an initial guess \c x
    - root::solve_bkt(double &x1, double x2, void *pa, funct &func)
    which solves a function given a solution bracketed between
    \c x1 and \c x2. The values of the function at \c x1 and
    \c x2 must have different signs.
    - root::solve_de(double &x, void *pa, funct &func, funct &df)
    which solves a function given an initial guess \c x and the
    derivative of the function \c df.  

    For one-dimensional solving, use cern_root or gsl_root_brent if
    you have the root bracketed, or gsl_root_stef if you have the
    derivative available. If you have neither a bracket or a
    derivative, you can use cern_mroot_root. 

    If not all of these three functions are overloaded, then the
    source code in \ref root is designed to try to automatically
    provide the solution using the remaining functions. Most of the
    one-dimensional solving routines, in their original form, are
    written in the second or third form above.  For example,
    gsl_root_brent is originally a bracketing routine of the form
    root::solve_bkt(), but calls to either root::solve() or
    root::solve_de() will attempt to automatically bracket the
    function given the initial guess that is provided. Also,
    gsl_root_stef is a "root-polishing" routine given
    derivatives of the form root::solve_de(). If either root::solve()
    or root::solve_bkt() are called, then root::solve_de() will be
    called with finite-differencing used to estimate the derivative.
    Of course, it is frequently most efficient to use the solver in
    the way it was intended.

    \todo Double check this documentation above

    \b Multi-dimensional \b solvers

    Solution of more than one equation is accomplished by descendants
    of the class \ref mroot. There are two basic functions
    - mroot::msolve(size_t n, ovector &x, void *pa, mm_funct &func) 
    which solves the \c n equations given in \c func with an initial
    guess \c x.

    For multi-dimensional solving, you can use either cern_mroot or
    gsl_mroot_hybrids. While cern_mroot does not use user-supplied
    derivatives, gsl_mroot_hybrids can use user-supplied derivative
    information (as in the GSL hybridsj method).

    \hline
    \section min_section Minimization

    One-dimensional minimization is performed by descendants of \ref
    minimize and provided in the library \c o2scl_minimize. There are
    two one-dimensional minimization algorithms, \ref cern_minimize
    and \ref gsl_min_brent, and they are both bracketing algorithms
    type where an interval and an initial guess must be provided.  If
    only an initial guess and no bracket is given, these two classes
    will attempt to find a suitable bracket from the initial
    guess. While the \ref minimize base class is designed to allow
    future descendants to optionally use derivative information, this
    is not yet supported for any one-dimensional minimizers.

    Multi-dimensional minimization is performed by descendants of
    multi_min: \ref gsl_mmin_simp, \ref gsl_mmin_conp, \ref
    gsl_mmin_conf, and \ref gsl_mmin_bfgs2. The class \ref
    multi_min_fix is a convenient way to perform a minimization while
    fixing some of the original parameters. The class \ref gsl_mmin_simp
    does not require or use any derivative information, but the other
    minimization classes are intended for use when derivatives
    are available. 

    Simulated annealing methods are also provided (see \ref
    anneal_section).

    It is important to note that not all of the minimization routines
    test the second derivative to ensure that it doesn't vanish to 
    ensure that we have found a true minimum. 

    A naive way of implementing constraints is to add a function to
    the original which increases the value outside of the allowed
    region. This can be done with the functions \ref constraint() and
    \ref lower_bound. There are two analogous functions, \ref
    cont_constraint() and \ref cont_lower_bound(), which continuous
    and differentiable versions. Where possible, it is better to use
    the constrained minimization routines described below.

    \hline
    \section conmin_section Constrained Minimization

    \o2 reimplements the Open Optimization Library (OOL) available at
    http://ool.sourceforge.net. The associated classes allow
    constrained minimization when the constraint can be expressed as a
    hyper-cubic constraint on all of the independent variables. The
    routines have been rewritten and reformatted for C++ in order to
    facilitate the use of member functions and user-defined vector
    types as arguments. The base class is ool_constr_mmin and there
    are two different constrained minimzation algorithms implemented
    in ool_mmin_pgrad, ool_mmin_spg. (The ool_mmin_gencan minimizer is
    not yet finished). The \o2 implementation should be essentially
    identical to the most recently released version of OOL.

    The constrained minimization classes operate in a similar way
    to the other multi-dimensional minimization classes (which are
    derived from multi_min). The constraints are specified with
    the function
    \code
    ool_constr_mmin::set_constraints(size_t nc, vec_t &lower, 
    vec_t &upper);
    \endcode
    and the minimization can be performed by calling either
    multi_min::mmin() or multi_min::mmin_de() (if the gradient is
    provided by the user). The "GENCAN" method requires a Hessian
    vector product and the user can specify this product for the
    minimization by using \ref ool_constr_mmin::mmin_hess(). The 
    Hessian product function can be specified as an object 
    of type \ref ool_hfunct or \ref ool_vhfunct in a similar way
    to the other function objects in \o2.

    There are five error codes defined in \ref ool_constr_mmin
    which are specific to the OOL classes. 

    \hline
    \section mcarlo_section Monte Carlo Integration

    Monte Carlo integration is performed by descendants of mcarlo_inte
    in the library \c o2scl_mcarlo (\ref gsl_monte, \ref gsl_miser,
    and \ref gsl_vegas). These routines are generally superior to the
    direct methods for integrals over regions with large numbers of
    spatial dimensions.

    \hline
    \section anneal_section Simulated Annealing

    Minimization by simulated annealing is performed by descendants of
    sim_anneal (see \ref gsl_anneal). The annealing schedule is given
    by a descendant of tptr_schedule (see \ref tptr_geoseries).

    \hline
    \section fit_section Non-linear Least-Squares Fitting

    Fitting is performed by descendants of \ref fit_base and fitting
    functions can be specifed using \ref fit_funct.  The GSL fitting
    routines (scaled and unscaled) are implemented in gsl_fit. A
    generic fitting routine using a minimizer object specified as a
    child of multi_min is implemented in min_fit. When the multi_min
    object is (for example) a sim_anneal object, min_fit can avoid
    local minima which can occur when fitting noisy data.

    \hline
    \section ode_section Solution of Ordinary Differential Equations

    Classes for non-adaptive integration are provided as descendents
    of \ref odestep and classes for adaptive integration are
    descendants of \ref adapt_step. To specify a set of functions
    to these classes, use a child of ode_funct for a generic 
    vector type or a child of ode_vfunct when using arrays.

    Solution of simple initial value problems is performed by \ref
    ode_iv_solve.

    Preliminary support for boundary value problems is given in
    children of \ref ode_bv_solve.

    \hline
    \section rng_section Random Number Generation

    Random number generators are descendants of \ref rnga and are
    provided in the library \c o2scl_rnga. While the base object \ref
    rnga is created to allow user-defined random number generators,
    the only random number generator presently included are from
    GSL. The GSL random number generator code is reimplemented in the
    class \ref gsl_rnga to avoid an additional performance
    penalty. This may not be a truly "object-oriented" interface in
    that it does not use virtual functions, but it avoids any possible
    performance penalty. Random number generators are implemented as
    templates in \ref sim_anneal and \ref mcarlo_inte. In these
    classes, the random number generator is a template type, rather
    than a member data pointer, in order to ensure fast execution.

    \hline
    \section tintp_section Two-dimensional Interpolation

    Successive use of \ref smart_interp is implemented in \ref twod_intp. 
    Also, see \ref planar_intp and \ref quad_intp and the 
    computation of contour lines in \ref contour. These latter three
    classes are somewhat experiemental at present.

    \hline
    \section other_section Other Routines

    (These are all experimental)

    \b Fourier \b transforms - see \ref gsl_fft

    \b Series \b acceleration - see \ref gsl_series

    \b Chebyshev \b approximations - see \ref gsl_chebapp

    \b Timing \b execution - see \ref timer_gettod and \ref timer_clock

    \b Polylogarithms - see \ref polylog

    \hline
    \section lset_section Library settings

    There are a couple library settings which are handled by 
    a global object \ref lib_settings of type \ref lib_settings_class.

    There are several data files that are used by various classes in
    the library. The installation procedure should ensure that these
    files are automatically found. However, if these data files are
    moved after installation, then a call to
    lib_settings_class::set_data_dir() can adjust the library to use
    the new directory. It is assumed that the directory structure
    within the data directory has not changed.

    \hline
    \section io_section Object I/O

    The I/O portion of the library is still experimental.
    
    Collections of objects can be stored in a \ref collection class,
    and these collections can be written to or read from text or
    binary files. User-defined classes may be added to the collections
    and may be read and written to files as long as a descendant of
    \ref io_base is provided.

    Every type has an associated I/O type which is a descendant of
    \ref io_base. In order to perform any sort of input/output on any
    type, an object of the corresponding I/O type must be instantiated
    by the user. This is not done automatically by the library. (Since
    it doesn't know which objects are going to be used ahead of time,
    the library would have to instantiate <em>all</em> of the I/O
    objects, which is needlessly slow.) This makes the I/O slightly
    less user-friendly, but much more efficient. For convenience, each
    subsection of the library has a class (named with an \c _ioc
    suffix) which will automatically allocate all I/O types for that
    subsection.
    
    <b>Level 1 functions:</b> Functions that input/output data from
    library-defined objects and internal types from files and combine
    these objects in collections. These are primarily member functions
    of the class \ref collection.

    <b>Level 2 functions:</b> Functions which are designed to allow
    the user to input or output data for user-generated objects. These
    are primarily member functions of classes \ref cinput and \ref
    coutput.
    
    <b>Level 3 functions:</b> Functions which allow low-level
    modifications on how input and output is performed. Usage of 
    level 3 functions is not immediately recommended for the casual
    user.
    
    <b>Level 1 usage</b>:

    For adding an object to a collection when you have a pointer
    to the I/O object for the associated type:
    \code
    int collection::add(std::string name, io_base *tio, void *vec, 
    int sz=0, int sz2=0, bool overwrt=true, bool owner=false);
    \endcode
    For adding an object to a collection otherwise:
    \code
    int collection::add(std::string name, std::string stype,
    void *vec, int sz=0, int sz2=0,
    bool overwrt=true, bool owner=false);
    \endcode

    To retrieve an object as a <pre>void *</pre> from a collection 
    use one of:
    \code
    int get(std::string tname, void *&vec);
    int get(std::string tname, void *&vec, int &sz);
    int get(std::string tname, void *&vec, int &sz, int &sz2);
    int get(std::string tname, std::string &stype, void *&vec);
    int get(std::string tname, std::string &stype, void *&vec, int &sz);
    int get(std::string tname, std::string &stype, void *&vec, int &sz,
    int &sz2);
    \endcode
    When retrieving a scalar object without error- and type-checking you
    can use the shorthand version:
    \code
    void *get(std::string name);
    \endcode

    To output one object to a file:
    \code
    int collection::out_one(out_file_format *outs, std::string stype, 
    std::string name, void *vp, int sz=0, int sz2=0);
    \endcode

    To input one object from a file with a given type and name:
    \code
    int collection::in_one_name(in_file_format *ins, std::string stype, 
    std::string name, void *&vp, int &sz, int &sz2);
    \endcode
    To input the first object of a given type from a file:
    \code
    int collection::in_one(in_file_format *ins, std::string stype, 
    std::string &name, void *&vp, int &sz, int &sz2);
    \endcode

    <b>Level 2 usage (string-based)</b>:

    If you don't have a pointer to the io_base child object corresponding
    to the type of subobject that you are manipulating, then you can use
    the following functions, which take the type name as a string.

    To input a sub-object in an io_base template for which memory
    has already been allocated use one of:
    \code
    int collection::object_in(std::string type, in_file_format *ins, void *vp,
    std::string &name);
    int collection::object_in(std::string type, in_file_format *ins, void *vp,
    int sz, std::string &name);
    int collection::object_in(std::string type, in_file_format *ins, void *vp,
    int sz, int sz2, std::string &name);
    \endcode
    
    To automatically allocate memory and input a sub-object of a 
    io_base template use one of:
    \code
    int collection::object_in_mem(std::string type, in_file_format *ins, 
    void *vp, std::string &name);
    int collection::object_in_mem(std::string type, in_file_format *ins, 
    void *vp, int sz, std::string &name);
    int collection::object_in_mem(std::string type, in_file_format *ins, 
    void *vp, int sz, int sz2, std::string &name);
    \endcode

    To output a subobject in an io_base template use:
    \code
    int collection::object_out(std::string type, out_file_format *outs,
    void *op, int sz=0, int sz2=0, std::string name="");
    \endcode
    
    <b>Level 2 usage (with io_base pointer)</b>:

    To input a sub-object in an io_base template for which memory
    has already been allocated use one of:
    \code
    virtual int object_in(cinput *cin, in_file_format *ins, object *op,
    std::string &name);
    virtual int object_in(cinput *cin, in_file_format *ins, object *op,
    int sz, std::string &name);
    virtual int object_in(cinput *cin, in_file_format *ins, object **op,
    int sz, int sz2, std::string &name);
    template<size_t N>
    int object_in(cinput *co, in_file_format *ins,
    object op[][N], int sz, std::string &name);
    \endcode
    
    To automatically allocate memory and input a sub-object of a 
    io_base template use one of:
    \code
    virtual int object_in_mem(cinput *cin, in_file_format *ins,
    object *&op, std::string &name);
    virtual int object_in_mem(cinput *cin, in_file_format *ins, object *&op,
    int &sz, std::string &name);
    virtual int object_in_mem(cinput *cin, in_file_format *ins,
    object **&op, int &sz, int &sz2,
    std::string &name);
    template<size_t N>
    int object_in_mem(cinput *co, in_file_format *ins,
    object op[][N], int &sz, std::string &name);
    \endcode
    
    To output a subobject in an io_base template use:
    \code
    virtual int object_out(coutput *cout, out_file_format *outs, object *op,
    int sz=0, std::string name="");
    virtual int object_out(coutput *cout, out_file_format *outs, object **op,
    int sz, int sz2, std::string name="");
    template<size_t N>
    int object_out(coutput *cout, out_file_format *outs,
    object op[][N], int sz, std::string name="");
    \endcode

    To automatically allocate/deallocate memory for an object, use:
    \code
    virtual int mem_alloc(object *&op);
    virtual int mem_alloc_arr(object *&op, int sz);
    virtual int mem_alloc_2darr(object **&op, int sz, int sz2);
    virtual int mem_free(object *op);
    virtual int mem_free_arr(object *op);
    virtual int mem_free_2darr(object **op, int sz);
    \endcode
    
    \b Usage \b of \b io_tlate

    The functions io_tlate::input() and io_tlate::output() need to be
    implemented for every class has information for I/O. For
    subobjects of the class, cinput::object_in() and
    cinput::object_out() can be called to input or output the
    information associated with the subobject. For input,
    cinput::object_in_name(), cinput::object_in_mem(), and
    cinput::object_in_mem_name() allow the freedom to input an object
    with a name or with memory allocation. The function
    coutput::object_out_name() allows one to output an object with a
    name.  If the class contains a pointer to the subobject, then
    io_base::pointer_in() or io_base::pointer_out() can be used.

    \hline
    \section design_section Design Considerations

    The design goal is to create an object-oriented computing library
    with classes that perform common numerical tasks. The most
    important principle is that the library should add functionality
    to the user while at the same time retaining as much freedom for
    the user as possible and allowing for ease of use and extensibility. 
    To that end, 
    - The classes which utilize user-specified functions
    should be able to operate on member functions without requiring
    a particular inheritance structure,
    - The interfaces ought to be generic so that the user
    can create new classes which perform related numerical
    tasks through inheritance,
    - The classes should not use static variables or functions
    - Const-correctness and type-safety should be used wherever possible, and
    - The design should be somewhat compatible with GSL.
    Also, the library provides higher-level routines
    for situations which do not require lower-level access. 

    <b>Header file dependencies</b>
    
    For reference, it's useful to know how the top-level header files
    depend on each other, since it can be difficult to trace everything
    down. In the \c base directory, the following are the most "top-level"
    header files and their associated dependencies within \o2 (there
    are other dependencies on GSL and the C standard library not
    listed here). 
    \code
    err_hnd.h : (none)
    sring_conv.h : (none)
    lib_settings.h : (none)
    array.h: err_hnd.h
    uvector_tlate.h: err_hnd.h
    ovector_tlate.h: uvector_tlate.h array.h err_hnd.h 
    misc.h : ovector_tlate.h uvector_tlate.h array.h lib_settings.h err_hnd.h
    test_mgr.h : misc.h ovector_tlate.h uvector_tlate.h 
    .             array.h lib_settings.h err_hnd.h
    \endcode

    <B>The use of templates</b>
    
    Templates are used extensively, and this makes for longer
    compilation times so any code that can be removed conveniently
    from the header files should be put into source code files
    instead. 

    \b Type-casting \b in \b vector \b and \b matrix \b design

    \o2 uses a GSL-like approach where viewing const double *
    arrays is performed by explicitly casting aways const'ness
    internally and then preventing the user from changing the data. 

    In gsl-1.6, the preprocessor output for \c vector/view_source.c is:
    \include gcv.cpp
    Note the explicit cast from const double * to double *. This
    is similar to what is done in \c src/base/ovector.cpp.

    \b Global \b objects

    There are three global objects that are created in
    libo2scl_base:
    - def_err_hnd is the default error handler
    - err_hnd is the pointer to the error handler (points to
    def_err_hnd by default)
    - lib_settings to control a few library settings

    All other global objects are to be avoided.

    \b Thread \b safety

    Most of the classes are thread-safe, meaning that two instances of
    the same class will not clash if their methods are called
    concurrently since static variables are only used for compile-time
    constants.  However, two threads cannot, in general, safely access
    the same instance of a class. In this respect, \o2 is no
    different from GSL.
    
    \b Documentation \b design
    
    
    The commands \\comment and \\endcomment delinate comments about
    the documentation that are present in the header files but don't
    ever show up in the HTML or LaTeX documentation.

    \b Release \b Procedure
    
    Internally, it's useful to keep track of the steps requred
    to create a full release (under construction)
    - svn update 
    - Update version numbers in configure.ac 
    - source aconfig.scr
    - make version-update
    - Edit version number in distribution link in doc/o2scl/main.h
    - make sinstall
    - make o2scl-test
    - Make the \c ex_ and \c bm_ files in the examples directory
    - Make sure acol.help and acol.usage are up to date
    - make o2scl-doc 
    - make o2scl-docp
    - make dist
    - svn commit 

    \hline
    \section license_section License Information
    
    \o2 (as well as CERNLIB and the Gnu Scientific Library (GSL)) is
    licensed under version 3 of the GPL as provided in the files \c
    COPYING and in \c doc/o2scl/extras/gpl_license.txt. After
    installation, it is included in the documentation in \c
    PREFIX/doc/extras/gpl_license.txt where the default \c PREFIX is
    \c /usr/local.
    
    \include gpl_license.txt

    This documentation is provided under the GNU Free Documentation
    License, as given below and provided in \c
    doc/o2scl/extras/fdl_license.txt. After installation, it is included
    in the documentation in \c PREFIX/doc/extras/fdl_license.txt where
    the default \c PREFIX is \c /usr/local.
    
    \include fdl_license.txt
    
    \hline
    \section ack_section Acknowledgements
    
    I would like to that the creators of GSL for their excellent work

    \hline
    \section ref_section Bibliography

    Some of the references which contain links should direct you to
    the work referred to directly through dx.doi.org.

    \anchor Bus75 Bus75:
    J.C.P. Bus and T.J. Dekker, ACM
    Trans. Math. Software 1 (1975) 330-345.

    \anchor Fletcher87 Fletcher87:
    R. Fletcher, Practical methods of optimization (John Wiley & Sons,
    Chichester 1987) 39.

    \anchor Krabs83 Krabs83:
    W. Krabs, Einfuhrung in die lieare und nictlineare Optimierung
    fur Ingenieure (BSB B.G. Teubner, Leipzig 1983) 84.

    \anchor Lewin83 Lewin83:
    L. Lewin, Polylogarithms and Associated Functions (North-Holland,
    New York, 1983).

    \anchor Longman58 Longman58:
    I.M. Longman, MTAC (later renamed Math. Comp.) \b 12 (1958) 205.

    \anchor More79 More79:
    J.J. More' and M.Y. Cosnard, ACM Trans. Math. Software,
    \b 5 (1979) 64-85.

    \anchor More80 More80:
    J.J. More' and M.Y. Cosnard, Algoright 554 BRENTM,
    Collected Algorithms from CACM (1980).

    \anchor Rutishauser63 Rutishauser63:
    H. Rutishauser, Ausdehnung des Rombergschen Prinzips
    (Extension of Romberg's Principle), Numer. Math. 5 (1963) 48-54.

    Other helpful references are:
    - The GSL reference manual
    http://www.gnu.org/software/gsl/manual/html_node/
    - The CERNLIB User's Guide at
    http://cernlib.web.cern.ch/cernlib/mathlib.html
    - The OOL reference manual at
    http://ool.sourceforge.net/
*/
/** 
    \page dl_page Download O2scl

    The present version is 0.803. The source distribution can 
    be obtained from \n
    - \c http://sourceforge.net/project/showfiles.php?group_id=206918
    
*/
/**
   \brief An empty class to add some items to the todo and bug lists
   
   \todo
   - Fix ex_file 
   - Improve ex_table
   - Fix problems with \c -ansi compilation on Cygwin
   - More examples and benchmarks
   - Document a list of all global functions and operators
   - Make sure we have a uvector_alloc, uvector_cx_alloc, 
   ovector_cx_const_reverse, ovector_cx_const_subvector_reverse, 
   uvector_reverse, uvector_const_reverse, uvector_subvector_reverse,
   uvector_const_subvector_reverse, omatrix_cx_diag, blah_const_diag, 
   umatrix_diag, and umatrix_cx_diag
   - ovector_cx_view::operator=(uvector_cx_view &) is missing
   - ovector_cx::operator=(uvector_cx_view &) is missing
   - uvector_c_view::operator+=(complex) is missing
   - uvector_c_view::operator-=(complex) is missing
   - uvector_c_view::operator*=(complex) is missing

   \future There may be a problem with const-correctness in vectors. 
   I'm not sure how it's best solved. It could be best to create two kinds of 
   ovector_view's: one const and one not. 10/19/07: I think it's the
   case that neither ovector_const_subvector, or const ovector_subvector 
   are truly const, but it's only const ovector_const_subvector that
   would be truly const. I'm not sure if this is related to the issue
   of constness in ovector_view discussed above.

   \future Consider breaking documentation up into sections? 
   
   \bug 
   - The file configure.ac does not correctly produce an error
   if the argument --disable-readline is not given when the 
   libncurses and libreadline libraries are not present. 
   - BLAS libraries not named \c libblas or \c libgslblas are not
   properly detected in ./configure and will have to be added
   manually. 
   - The -lm flag may not be added properly by ./configure

*/
class other_todos_and_bugs {
}

/** \brief The main namespace
    \nothing

    Most of the classes reside in this namespace, but it is removed
    from the documentation for clarity. 
    \htmlonly
    See <a href="annotated.html">Data Structures</a> for a full list.
    \endhtmlonly
*/
namespace o2scl {
}

/*
  Old example stuff

  \todo Maybe rewrite this as an example separate from ex_fptr below.

  \b An \b example

  \c ex_fptr gives an example of the how member functions and
  external parameters are supplied to numerical routines. In this
  case, a member function (which contains the equation  \f$ 
  -\sin(x-0.2)/|x+0.01|+c_1+c_2 =0 \f$  where  \f$ c_1 \f$  is a private
  data member and  \f$ c_2 \f$  is an externally-supplied parameter, is
  passed to the gsl_root_brent routine, which solves the equation.

  \b ex_fptr.cpp
  \include ex_fptr.cpp

  The output is \b ex_fptr.scr:
  \include ex_fptr.scr

  \hline
  \section examples_section Other Examples

  All of the following examples are in the \c examples subdirectory.

  \b I/O \b example

  This example demonstrates the ease of using text or binary files
  for input or output. A \ref collection is filled with data and
  output to several files. The data is read in from each of the
  files and written to \c cout.

  \b ex_file.cpp
  \include ex_file.cpp

  This produces the following output:

  \b ex_file.scr
  \include ex_file.scr

  Note that the binary files automatically store the full precision
  for floating-point numbers, but that the text files do not. The
  precision for text files can be modified, see \ref text_out_file
  for details.

  \b Differentiation

  \b ex_diff.cpp
  \include ex_diff.cpp

  The result of plotting \c err and \c err' from the 
  output file \c ex_diff.out is

  \image html ex_diff.gif "ex_diff.out"
  \image latex ex_diff.eps "ex_diff.out" width=9cm

  Note that the formula for equally-spaced abscissas provided
  by eq_interval is more accurate than interpolation with
  smart_interp at the boundaries, but less accurate otherwise.

*/
