/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_ANNEAL_H
#define O2SCL_GSL_ANNEAL_H

#include <o2scl/sim_anneal.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_siman.h>

#ifndef DOXYGENP
namespace o2scl {
#endif
  
  /** 
      \brief Multidimensional minimization by simulated annealing (GSL)
      
      Minimize by simulated annealing for an arbitrary temperature
      schedule and random number generator. A move is defined by
      \f[
      x_{i,\mathrm{new}} = \mathrm{step\_size}_i 
      (2 u_i - 1) + x_{i,\mathrm{old}}
      \f]
      where the \f$u_i\f$ are random numbers between 0 and 1.  The
      random number generator and temperature schedule are set in the
      parent, sim_anneal.  The variables minimize::tolx and
      minimize::tolf are not used

      The step size for each dimension is specified in \ref
      set_stepsize(). The number of stepsizes specified need not be
      the same as the number of dimensions. If \c nstep is the number
      of stepsizes, then the stepsize for dimension \c i is
      \code
      step_size[i % nstep]
      \endcode

      \comment 
      Note that I don't make a direct reference to "nstep" here
      since it's protected, and I want the documentation to 
      work even if the protected part is hidden.
      \endcomment

      \future Implement a more general simulated annealing routine
      which would allow the solution of discrete problems like the
      Traveling Salesman problem.
      \future Implement a method which automatically minimizes
      within some specified tolerance?
      
  */
  template<class param_t, class func_t, class vec_t=ovector_view,
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc,
    class rng_t=gsl_rnga> class gsl_anneal :
    public sim_anneal<param_t,func_t,vec_t,rng_t> {
      public:
      
      gsl_anneal() {
	nstep=1;
	step_sizes=new double[1];
	step_sizes[0]=10.0;
      }

      virtual ~gsl_anneal() {
	delete[] step_sizes;
      }
      
      /** \brief Calculate the minimum \c fmin of \c func w.r.t the 
	  array \c x0 of size \c nvar.
      */
      virtual int mmin(size_t nvar, vec_t &x0, double &fmin, param_t &pa,
		       func_t &func) 
      {
	
	fmin=0.0;

	allocate(nvar);

	double E, new_E, best_E, T;
	int i, iter=0;
	bool done;
	size_t j;

	for(j=0;j<nvar;j++) {
	  x[j]=x0[j];
	  best_x[j]=x0[j];
	}
    
	E=func(nvar,x,pa);
	best_E=E;
  
	T=this->tp->start(0.0,0,x0,NULL);
	done=false;

	while (!done) {
	  for (i=0;i<this->ntrial;++i) {
	    for (j=0;j<nvar;j++) new_x[j]=x[j];
      
	    step(new_x,nvar);
	    new_E=func(nvar,new_x,pa);
      
	    if(new_E<=best_E){
	      for(j=0;j<nvar;j++) best_x[j]=new_x[j];
	      best_E=new_E;
	    }
      
	    // Take the crucial step: see if the new point is accepted
	    // or not, as determined by the boltzman probability
	    if (new_E<E) {
	      for(j=0;j<nvar;j++) x[j]=new_x[j];
	      E=new_E;
	    } else if (this->def_rng.random() < exp(-(new_E-E)/(boltz*T)) ) {
	      for(j=0;j<nvar;j++) x[j]=new_x[j];
	      E=new_E;
	    }
	  }
    
	  if (this->verbose>0) {
	    this->print_iter(nvar,best_x,best_E,iter,T,"gsl_anneal");
	    iter++;
	  }
    
	  // apply the cooling schedule to the temperature
	  T=this->tp->next(0.0,0,x0,NULL);
	  if (this->tp->done(0.0,0,x0,NULL)) {
	    done=true;
	  }
	}
  
	for(j=0;j<nvar;j++) x0[j]=best_x[j];
	fmin=best_E;

	free();
  
	return 0;
      }
      
      /// Return string denoting type ("gsl_anneal")
      virtual const char *type() { return "gsl_anneal"; }

      /// Set the step
      template<class vec2_t> int set_stepsize(size_t n, vec2_t &ss) {
	if (n!=0) {
	  delete[] step_sizes;
	  step_sizes=new double[n];
	  for(size_t i=0;i<n;i++) {
	    step_sizes[i]=ss[i];
	  }
	  nstep=n;
	  return 0;
	}
	set_err_ret("Cannot set zero stepsizes in set_stepsize().",gsl_einval);
      }

      /// Boltzmann factor (default 1.0).
      double boltz;

#ifndef DOXYGEN_INTERNAL
      
      protected:
      
      /// \name Storage for present, next, and best vectors
      //@{
      alloc_vec_t x, new_x, best_x;
      //@}

      /// Allocation object
      alloc_t ao;

      /// Number of step sizes
      size_t nstep;
      
      /// Step sizes
      double *step_sizes;

      /** \brief Allocate memory for a minimizer over \c n dimensions
	  with stepsize \c step and Boltzmann factor \c boltz_factor 
      */
      virtual int allocate(size_t n, double boltz_factor=1.0) {
	ao.allocate(x,n);
	ao.allocate(new_x,n);
	ao.allocate(best_x,n);
	boltz=boltz_factor;
	return 0;
      }

      /// Free allocated memory
      virtual int free() {
	ao.free(x);
	ao.free(new_x);
	ao.free(best_x);
	return 0;
      }

      /// Make a step to a new attempted minimum
      virtual int step(vec_t &sx, int nvar) {
	for(int i=0;i<nvar;i++) {
	  double u=this->def_rng.random();
	  sx[i]=(2.0*u-1.0)*step_sizes[i%nstep]+sx[i];
	}
	return 0;
      }
  
#endif

    };

#ifndef DOXYGENP
}
#endif

#endif
