/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ARRAY_H
#define O2SCL_ARRAY_H

/** \file array.h
    \brief Various array classes

    This file contains the alloction classes
    - array_alloc
    - array_2d_alloc
    - pointer_alloc
    - pointer_2d_alloc
    the classes for the manipulation of arrays in smart_interp
    - array_reverse
    - array_subvector
    - array_subvector_reverse
    - array_const_reverse
    - array_const_subvector
    - array_const_subvector_reverse
    the array equivalent of omatrix_row or umatrix_row (see 
    usage proposed in src/ode/ode_it_solve_ts.cpp)
    - array_row
    an array output function
    - vector_out()

    For output of matrices rather than vectors, see matrix_out()
    and related functions documented in \ref columnify.h.

    Note that the classes 
    - array_reverse
    - array_subvector
    - array_subvector_reverse
    - array_const_reverse
    - array_const_subvector
    - array_const_subvector_reverse
    can be used with pointers or arrays, but array_alloc and
    pointer_alloc are \e not interchangable.

    \todo Ensure that array_row works, either here or in 
    src/ode/ode_it_solve_ts.cpp

    \todo Document vector_sort, etc.
*/

#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/err_hnd.h>
#include <gsl/gsl_ieee_utils.h>
#include <gsl/gsl_sort.h>

#ifndef DOXYGENP
namespace o2scl
{
#endif

  /** 
      \brief A simple class to provide an \c allocate() function
      for arrays
      
      The functions here are blank, as fixed-length arrays are
      automatically allocated and destroyed by the compiler. This
      class is present to provide an analog to \ref pointer_alloc and
      \ref ovector_alloc.
  */
  template<class vec_t> class array_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(vec_t &v, size_t i) {}
    /// Free memory
    void free(vec_t &v) {}
  };

  /** 
      \brief A simple class to provide an \c allocate() function
      for 2-dimensional arrays
      
      The functions here are blank, as fixed-length arrays are
      automatically allocated and destroyed by the compiler. This
      class is present to provide an analog to \ref pointer_2d_alloc
      and \ref omatrix_alloc
  */
  template<class mat_t> class array_2d_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(mat_t &v, size_t i, size_t j) {}
    /// Free memory
    void free(mat_t &v, size_t i) {}
  };

  /** 
      \brief A simple class to provide an \c allocate() function
      for pointers

      Uses \c new and \c delete.
   */
  template<class base_t> class pointer_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(base_t *&v, size_t i) { v=new base_t[i]; }
    /// Free memory
    void free(base_t *&v) { delete[] v; }
  };

  /** 
      \brief A simple class to provide an \c allocate() function
      for pointers

      Uses \c new and \c delete.
  */
  template<class base_t> class pointer_2d_alloc {
  public:
    /// Allocate \c v for \c i elements
    void allocate(base_t **&v, size_t i, size_t j) { 
      v=new base_t *[i]; 
      for(size_t m=0;m<i;m++) v[m]=new base_t[j];
    }
    /// Free memory
    void free(base_t **&v, size_t i) { 
      for(size_t m=0;m<i;m++) delete[] v[m];
      delete[] v; 
    }
  };

  /** 
      \brief A simple class which reverses the order of an array
   */
  template<size_t sz> class array_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_reverse(double *arr) {
      a=arr;
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[sz-1-i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[sz-1-i];
    }
    
  };

  /** 
      \brief A simple class which reverses the order of an array
   */
  template<size_t sz> class array_const_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_const_reverse(const double *arr) {
      a=(double *)arr;
    }

    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[sz-1-i];
    }
    
  };

  /** 
      \brief A simple subvector class for an array (without error checking)
   */
  class array_subvector {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;

    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_subvector(double *arr, size_t offset, size_t n) {
      a=arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[i];
    }
    
  };

  /** \brief A simple subvector class for a const array 
      (without error checking)
   */
  class array_const_subvector {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;

    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_const_subvector(const double *arr, size_t offset, size_t n) {
      a=(double *)arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[i];
    }
    
  };

  /** 
      \brief Reverse a subvector of an array
   */
  class array_subvector_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;
    
    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_subvector_reverse(double *arr, size_t offset, size_t n) {
      a=arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    double &operator[](size_t i) {
      return a[len-1-i];
    }
    
    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[len-1-i];
    }
    
  };

  /** 
      \brief Reverse a subvector of a const array
  */
  class array_const_subvector_reverse {
    
#ifndef DOXYGEN_INTERNAL

  protected:
    
    /// The array pointer
    double *a;

    /// The offset
    size_t off;

    /// The subvector length
    size_t len;

#endif

  public:

    /// Create a reversed array from \c arr of size \c sz
    array_const_subvector_reverse(const double *arr, size_t offset, size_t n) {
      a=(double *)arr+offset;
      off=offset;
      len=n;
    }

    /** 
	\brief Array-like indexing 
    */
    const double &operator[](size_t i) const {
      return a[len-1-i];
    }
    
  };

  /// Extract a row of a C-style 2d-array
  template <class data_t, class array_2d_t> class array_row {
  public:
    data_t *p;
    array_row(array_2d_t &a, size_t i) {
      p=&(a[i][0]);
    }
    data_t &operator[](size_t i) {
      return p[i];
    }
  };

  /** 
      \brief Output a vector to a stream
      
      No trailing space is output after the last element, and an
      endline is output only if \c endline is set to \c true.  If the
      parameter \c n is zero, this function silently does nothing.
  */
  template<class vec_t> 
    int vector_out(std::ostream &os, size_t n, vec_t &v, 
		   bool endline=false) {
    // This next line is important since n-1 is not well-defined if n=0
    if (n==0) return 0;
    for(size_t i=0;i<n-1;i++) os << v[i] << " ";
    os << v[n-1];
    if (endline) os << std::endl;
    return 0;
  }
  
  /** 
      \brief Provide a downheap() function for vector_sort()
  */
  template<class data_t, class vec_t>
    void sort_downheap(vec_t &data, const size_t N, size_t k) {
    
    data_t v=data[k];
    
    while (k<=N/2) {
      size_t j=2*k;
      
      if (j<N && data[j] < data[j+1]) j++;
      if (!(v < data[j])) break;
      data[k]=data[j];
      k=j;
    }
    
    data[k]=v;
  }

  /** 
      \brief Sort a vector

      This is a generic sorting template function. It will work for
      any types \c data_t and \c vec_t for which
      - \c data_t has an <tt>operator=</tt>
      - \c data_t has a less than operator to compare elements
      - <tt>vec_t::operator[]</tt> returns a reference
      to an object of type \c data_t 
      
      In particular, it will work with \ref ovector, \ref uvector, 
      \ref ovector_int, \ref uvector_int (and other related O2scl
      vector classes), the STL template class <tt>std::vector</tt>,
      and arrays and pointers of numeric, character, and string
      objects.

      For example,
      \code
      std::string list[3]={"dog","cat","fox"};
      vector_sort<std::string, std::string[3]>(3,list);
      \endcode
  */
  template<class data_t, class vec_t>
    int vector_sort(const size_t n, vec_t &data) {
    
    size_t N;
    size_t k;
    
    if (n==0) return 0;
    
    N=n-1;
    k=N/2;
    k++;
    do {
      k--;
      sort_downheap<data_t,vec_t>(data,N,k);
    } while (k > 0);
    
    while (N > 0) {
      data_t tmp=data[0];
      data[0]=data[N];
      data[N]=tmp;
      N--;
      sort_downheap<data_t,vec_t>(data,N,0);
    }
    
    return 0;
  }

  /** 
      \brief "Rotate" a vector so that the kth element is now the beginning

      This is a generic template function which will work for
      any types \c data_t and \c vec_t for which
      - \c data_t has an <tt>operator=</tt>
      - <tt>vec_t::operator[]</tt> returns a reference
      to an object of type \c data_t 
  */
  template<class data_t, class vec_t>
    int vector_rotate(const size_t n, vec_t &data, size_t k) {

    data_t *tmp=new data_t[n];
    for(size_t i=0;i<n;i++) {
      tmp[i]=data[(i+k)%n];
    }
    for(size_t i=0;i<n;i++) {
      data[i]=tmp[i];
    }
    delete[] tmp;
    
    return 0;
  }

  /// Compute the maximum of the first \c n elements of a vector
  template<class data_t, class vec_t>
    int vector_max(const size_t n, vec_t &data, data_t &max, size_t &ix) {
    
    if (n==0) {
      set_err_ret("Send size=0 to vector_max().",gsl_efailed);
    }
    ix=0;
    max=data[0];
    for(size_t i=1;i<n;i++) {
      if (data[i]>max) {
	max=data[i];
	ix=i;
      }
    }
    return 0;
  }

  /// Compute the minimum of the first \c n elements of a vector
  template<class data_t, class vec_t>
    int vector_min(const size_t n, vec_t &data, data_t &min, size_t &ix) {
    
    if (n==0) {
      set_err_ret("Send size=0 to vector_min().",gsl_efailed);
    }
    ix=0;
    min=data[0];
    for(size_t i=1;i<n;i++) {
      if (data[i]<min) {
	min=data[i];
	ix=i;
      }
    }
    return 0;
  }

  /** 
      \brief Compute the sum of the first \c n elements of a vector

      If \c n is zero, this will set \c avg to zero and return 
      \ref gsl_success.
   */
  template<class data_t, class vec_t>
    int vector_sum(const size_t n, vec_t &data, data_t &sum) {
    
    sum=0;
    for(size_t i=0;i<n;i++) {
      sum+=data[i];
    }
    return 0;
  }

  /** 
      \brief Compute the mean of the first \c n elements of a vector

      If \c n is zero, this will set \c avg to zero and return 
      \ref gsl_success.
   */
  template<class data_t, class vec_t>
    int vector_avg(const size_t n, vec_t &data, data_t &avg) {
    
    avg=0;
    for(size_t i=0;i<n;i++) {
      avg+=data[i];
    }
    avg/=((data_t)n);
    return 0;
  }

  /** \brief Compute the variance of the first \c n elements of a vector
      given the mean \c mean.

      If \c n is zero, this will set \c avg to zero and return 
      \ref gsl_success.
   */
  template<class data_t, class vec_t>
    int vector_variance(const size_t n, vec_t &data, data_t &mean,
			data_t &var) {
    
    var=0;
    for(size_t i=0;i<n;i++) {
      data_t delta=(data[i]-mean);
      var+=(delta*delta-var)/((double)(i+1));
    }
    return 0;
  }

  /** \brief Reverse a vector
   */
  template<class data_t, class vec_t>
    int vector_reverse(const size_t n, vec_t &data) {
    data_t tmp;

    for(size_t i=0;i<n/2;i++) {
      tmp=data[n-1-i];
      data[n-1-i]=data[i];
      data[i]=tmp;
    }
    return 0;
  }

  /** \brief Compute the standard deviation of the first \c n elements 
      of a vector
  */
  template<class data_t, class vec_t>
    int vector_stdev(const size_t n, vec_t &data, data_t &var) {
    
    data_t mean;
    vector_mean(n,data,mean);
    var=0;
    for(size_t i=0;i<n;i++) {
      data_t delta=(data[i]-mean);
      var+=(delta*delta-var)/((double)(i+1));
    }
    var=sqrt(var*((double)n)/((double)(n-1)));
    return 0;
  }

#ifndef DOXYGENP
}
#endif

#endif
