/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ERR_HND_H
#define O2SCL_ERR_HND_H

#include <iostream>
#include <string>
#include <gsl/gsl_errno.h>

/** \file err_hnd.h
    \brief File for definitions for err_class
*/

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief The error definitions from GSL

  */
  enum {
    /// Success
    gsl_success  = 0,
    /// Failure
    gsl_failure  = -1,
    /// iteration has not converged 
    gsl_continue = -2,  
    /// input domain error, e.g sqrt(-1) 
    gsl_edom     = 1,   
    /// output range error, e.g. exp(1e100) 
    gsl_erange   = 2,   
    /// invalid pointer 
    gsl_efault   = 3,   
    /// invalid argument supplied by user 
    gsl_einval   = 4,   
    /// generic failure 
    gsl_efailed  = 5,   
    /// factorization failed 
    gsl_efactor  = 6,   
    /// sanity check failed - shouldn't happen 
    gsl_esanity  = 7,   
    /// malloc failed 
    gsl_enomem   = 8,   
    /// problem with user-supplied function 
    gsl_ebadfunc = 9,   
    /// iterative process is out of control 
    gsl_erunaway = 10,  
    /// exceeded max number of iterations 
    gsl_emaxiter = 11,  
    /// tried to divide by zero 
    gsl_ezerodiv = 12,  
    /// user specified an invalid tolerance 
    gsl_ebadtol  = 13,  
    /// failed to reach the specified tolerance 
    gsl_etol     = 14,  
    /// underflow 
    gsl_eundrflw = 15,  
    /// overflow 
    gsl_eovrflw  = 16,  
    /// loss of accuracy 
    gsl_eloss    = 17,  
    /// failed because of roundoff error 
    gsl_eround   = 18,  
    /// matrix, vector lengths are not conformant 
    gsl_ebadlen  = 19,  
    /// matrix not square 
    gsl_enotsqr  = 20,  
    /// apparent singularity detected 
    gsl_esing    = 21,  
    /// integral or series is divergent 
    gsl_ediverge = 22,  
    /// requested feature is not supported by the hardware
    gsl_eunsup   = 23,  
    /// requested feature not (yet) implemented 
    gsl_eunimpl  = 24,  
    /// cache limit exceeded 
    gsl_ecache   = 25,  
    /// table limit exceeded 
    gsl_etable   = 26,  
    /// iteration is not making progress toward solution
    gsl_enoprog  = 27,  
    /// jacobian evaluations are not improving the solution 
    gsl_enoprogj = 28,  
    /// cannot reach the specified tolerance in f 
    gsl_etolf    = 29,  
    /// cannot reach the specified tolerance in x 
    gsl_etolx    = 30,  
    /// cannot reach the specified tolerance in gradient 
    gsl_etolg    = 31,  
    /// end of file 
    gsl_eof      = 32,  
    /// a blank method in a base class has been called
    gsl_nobase   = 33,
    /// Generic "not found" result
    gsl_notfound = 34,
    /// Incorrect type for memory object
    gsl_memtype = 35,
    /// File not found
    gsl_efilenotfound = 36,
    /// Invalid index for array or matrix
    gsl_index = 37
  };
  
  /** 
      \brief The error handler 
      
      An error handler for use in O2scl which replaces the GSL error handler
      
      Note that the string arguments to set() can refer to temporary
      storage, since they are copied when the function is called and
      an error is set.

  */
  class err_class {

  public:  

    err_class();

    /** 
	\brief Set an error 
    
	This is separate from set(), since the gsl error handler
	needs to be a static function.
     */
    static void gsl_hnd(const char *reason, const char *file, 
			int line, int lerrno);

    /// Set an error 
    void set(const char *reason, const char *file, 
	     int line, int lerrno);

    /// Add information to previous error
    void add(const char *reason, const char *file, 
	     int line, int lerrno);

    /// Get the last error
    void get(const char *&reason, const char *&file,
	     int &line, int &lerrno);

    /// Return the last error number
    int get_errno();

    /// Return the line number of the last error
    int get_line();

    /// Return the reason for the last error
    const char *get_reason();

    /// Return the file name of the last error
    const char *get_file();

    /// Return a string summarizing the last error
    const char *get_str();

    /// Remove last error information
    void reset();

    /// Force a hard exit if an error occurs
    void set_mode(int m) {
      if (m<0) mode=0;
      else if (m>2) mode=2;
      else mode=m;
    }

    /** \brief If true, call exit() when an array index error is set 
	(default true)
	
	This is ignored if \c O2SCL_ARRAY_ABORT is not defined.
    */
    bool array_abort;

  protected:

#ifndef DOXYGEN_INTERNAL

    /// A pointer to the default error handler
    static err_class *ptr;

    /// The maximum size of error explanations
    static const int rsize=300;
    /// The maximum size of error explanations with the line and file info
    static const int fsize=400;

    /// The error number
    int a_errno;
    /// The line number
    int a_line;
    /// The mode of error handling
    int mode;

    /// The filename
    char *a_file;
    /// The error explanation
    char a_reason[rsize];
    /// A full string with explanation and line and file info
    char fullstr[fsize];

#endif

  };

  /** \brief The global error handler pointer
  */      
  extern err_class *err_hnd;

  /** \brief The default error handler
  */      
  extern err_class def_err_hnd;

  /** \brief Set an error
  */
#define set_err(d,n) o2scl::set_err_fn(d,__FILE__,__LINE__,n);

  /** \brief Set an error and return the error value
  */
#define set_err_ret(d,n) \
do { o2scl::set_err_fn(d,__FILE__,__LINE__,n); return n; } while (0)
  
  /** \brief Set an error and add the information from the last error
  */
#define add_err(d,n) o2scl::add_err_fn(d,__FILE__,__LINE__,n);
  
  /** \brief Set an error, add the information from the last error,
      and return the error value
  */
#define add_err_ret(d,n) \
do { o2scl::add_err_fn(d,__FILE__,__LINE__,n); return n; } while(0)
  
  /** \brief Set an error
  */
  inline void set_err_fn(const char *desc, const char *file, int line,
			 int errnum) {
    err_hnd->set(desc,file,line,errnum);
    return;
  }
  
  /** \brief Set an error and add the information from the last error
  */
  inline void add_err_fn(const char *desc, const char *file, int line,
			 int errnum) {
    err_hnd->add(desc,file,line,errnum);
    return;
  }
  
  /// Desc
  inline void error_update(int &ret, int err) { if (ret==0) ret=err; }
  
  /** \brief Print out error information
   */
#define err_print(ev)							\
  do { if (ev!=0) std::cout << ev << " " << err_hnd->get_str() << std::endl; \
    else std::cout << "No error occured." << std::endl; } while (0)
  
  /** \brief Print out error information to \c cerr, do nothing 
      occured
  */
#define cerr_print(ev)							\
  do { if (ev!=0) std::cerr << ev << " " << err_hnd->get_str() << std::endl; \
  } while (0)
  
  /** \brief A version of \c assert, i.e. exit if the error value is non-zero
      and do nothing otherwise

      \future Make this consistent with assert() using NDEBUG?
  */
#define err_assert(ev)							\
  do { if (ev!=0) { std::cout << "O2scl: Macro err_assert() causing exit" \
			      << " from error " << ev << " at "		\
			      << __LINE__ << " in file:\n "		\
			      << __FILE__ << std::endl;			\
	 std::cout << "Error handler string:\n " << err_hnd->get_str()	\
		   << std::endl; exit(ev); } } while (0)
  
  /** \brief A version of \c assert for bool types. Exit if the argument
      is false
  */
#define bool_assert(ev,str)						\
  do { if (ev==false) {							\
      std::cout << "O2scl: Macro bool_assert() causing exit at line "	\
		<< __LINE__ << " in file:\n "				\
		<< __FILE__ << std::endl;				\
      std::cout << "Given string: " << str				\
		<< std::endl; exit(-1); } } while (0)
  
#ifndef DOXYGENP
}
#endif

#endif
