/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_MISC_H
#define O2SCL_MISC_H
/** \file misc.h
    \brief Miscellaneous functions
*/

#include <iostream>
#include <cmath>
#include <string>
#include <fstream>
#include <sstream>
#include <o2scl/err_hnd.h>
#include <o2scl/lib_settings.h>
#include <gsl/gsl_ieee_utils.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Calculate a Fermi-Dirac distribution function safely
      
      \f$ \left[1+\exp\left(E/T-\mu/T\right)\right]^{-1} \f$ 
      
      This calculates a Fermi-Dirac distribution function guaranteeing
      that numbers larger than \f$ \exp(\mathrm{limit}) \f$ and
      smaller than \f$ \exp(-\mathrm{limit}) \f$ will be avoided. The
      default value of \c limit ensures accuracy to within 1 part in
      \f$ 10^{17} \f$ compared to the maximum of the distribution
      (which is unity).
      
      Note that this function may return Inf or NAN if \c limit is too 
      large, depending on the machine precision.
  */
  double fermi_function(double E, double mu, double T, double limit=40.0);

  /** 
      \brief Reformat the columns for output of width \c size 

      Given a string array \c in_cols of size \c nin, screenify()
      reformats the array into columns creating a new string array \c
      outc with size \c nout.
      
      For example, for an array of 10 strings 
      \verbatim
      test1
      test_of_string2
      test_of_string3
      test_of_string4
      test5
      test_of_string6
      test_of_string7
      test_of_string8
      test_of_string9
      test_of_string10
      \endverbatim
      screenify() will create an array of 3 new strings:
      \verbatim
      test1            test_of_string4  test_of_string7  test_of_string10
      test_of_string2  test5            test_of_string8
      test_of_string3  test_of_string6  test_of_string9
      \endverbatim
      
      The string array given in outc must be deleted with 
      delete[] after usage.
      
      If the value of \c max_size is less than the 
      length of the longest input string (plus one for
      a space character), then the output strings may
      have a larger length than \c max_size.

      \todo Consider making this into a class so that the 
      memory for the output columns is automatically 
      handled.
  */
  void screenify(const std::string *in_cols, int nin, 
		 std::string *&outc, int &nout, int max_size=80);
  
  /** 
      \brief Count the number of words in the string \c str 
   
      Words are defined as groups of characters separated by
      whitespace, where whitespace is any combination of adjacent
      spaces, tabs, carriage returns, etc. On most systems, whitespace
      is usually defined as any character corresponding to the
      integers 9 (horizontal tab), 10 (line feed), 11 (vertical tab),
      12 (form feed), 13 (carriage return), and 32 (space bar).  The
      test program \c misc_ts enumerates the characters between 0 and
      255 (inclusive) that count as whitespace for this purpose.

      Note that this function is used in text_in_file::string_in 
      to perform string input.
  */
  int count_words(std::string str);
  
  /** 
      \brief Naive string comparison 
      
      This is used internally for the STL routines which require a way
      to compare strings in the class \ref table and in the I/O
      classes.
  */
  typedef struct {
    /// Return \c s1<s2
    bool operator()(const std::string s1, const std::string s2) const {
      return s1<s2;
    }
  } string_comp;
  
  /** \brief Take a string of binary quads and compress them to 
      hexadecimal digits

      This function proceeds from left to right, ignoring parts of the
      string that do not consist of squences of four '1's or '0's.
  */
  std::string binary_to_hex(std::string s);

  /// Create a new C-style 2-dimensional array
  template<class type> type **new_2d_array(size_t nr, size_t nc) {
    type **t;
    t=new type *[nr];
    for(size_t i=0;i<nr;i++) t[i]=new type[nc];
    return t;
  }

  /// Create a new C-style 2-dimensional array
  template<class type> int delete_2d_array(type **t, size_t nr) {
    for(size_t i=0;i<nr;i++) delete[] t[i];
    delete[] t;
    return 0;
  }

  /** 
      \brief Generate number sequence for testing
      
      A class which generates \c tot numbers from -1 to 1, making sure
      to include -1, 1, 0, and numbers near -1, 0 and 1 (so long as \c
      tot is sufficiently large). If gen() is called more than \c tot
      times, it just recycles through the list again.  The template
      argument \c tot should probably be greater than or equal to
      three.
      
      At present, this is used to generate combinations of
      coefficients for testing the polynomial solvers.
      
  */
  template<size_t tot> class gen_test_number {

#ifndef DOXYGEN_INTERNAL

  protected:

    /// Count number of numbers generated
    int n;

    /// A constant factor for the argument to \c tanh().
    double fact;

#endif

  public:

    gen_test_number() {
      n=0;
      fact=((double)tot)/4.0/5.0;
    }

    /// Return the next number in the sequence
    double gen() {
      double x, dtot=((double)tot), dn=((double)n);
      if (n==0) {
	x=-1.0;
      } else if (n==tot/2) {
	x=0.0;
      } else if (n==tot-1) {
	x=1.0;
      } else if (n==tot) {
	x=-1.0;
	n=0;
      } else if (n<((int)tot)/2) {
	x=(tanh((dn-dtot/4.0)/fact)-1.0)/2.0;
      } else {
	x=(tanh((dn-3.0*dtot/4.0)/fact)+1.0)/2.0;
      }
      n++;
      return x;
    }
  };

#ifdef O2SCL_NEVER_DEFINED
  class bm_base {
  };

  class benchmark {
  protected:
    size_t nt;
    bm_base *ptrs;
  public:
    int set_types(size_t ntypes, bm_base *types[]) {
      return 0;
    }
    
  };

#endif

#ifndef DOXYGENP
}
#endif

#endif

