/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_TEXT_FILE_H
#define O2SCL_TEXT_FILE_H

#include <iostream>
#include <stack>
#include <o2scl/err_hnd.h>
#include <o2scl/file_format.h>
#include <o2scl/misc.h>
#include <o2scl/lib_settings.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief An output text file

      A collection file is simply a text file containing a list of
      objects specially formatted for input and output. Each entry in
      the text file is of the form:
    
      object_type object_name object_version word_number word1 word2 word3 ...
    
      The type, name, and version are all strings that contain no
      whitespace.  "a_name" is a valid name but, "a name" is not.
    
      Parts of the object definition may be separated by any amount of
      whitespace, with the exception of 'strings'.
    
      The collection file may contain comments, which are lines that begin
      with the '#' character. Comments may last more than one line (so long as 
      every line begins with '#'), but they may not occur in the middle of
      an object definition.
      \verbatim
      # comment 1
      double a 4.0
      double[] c 4 
      4.5 3.4 2.3 4.2
      # comment 2
      double b 5.0
      \endverbatim
      is acceptable, but
      \verbatim
      # comment 1
      double a 4.0
      double[] c 4 
      4.5 3.4 2.3 4.2
      double b
      # comment 2
      5.0
      \endverbatim
      is not, since the second comment occurs in the middle of the definition
      for the object named 'b'.
    
      Normal variable:
      type name version word data1 ... data2
    
      Object containing pointer:
      type name version word data1 ... ptr_type ptr_name ... data2
    
      where ptr_type is the type of the object being pointed to
      and ptr_name is the name of the object. If it's not in the
      list, then the object is assigned a unique name of the
      form ptrX where 'X' is an integer >= 0. 
    
      Static objects are the same, except they are preceeded by 
      the keyword \c static  and do not have a name associated with them.
    
      This is useful for output to text files and to std::cout, i.e.
      text_out_file tf(&cout);

      <b>Note:</b> Text files are not entirely architecture-independent,
      For example, a larger integer will not be read correctly on small 
      integer systems.
   */
  class text_out_file : public out_file_format {
  public:

    /** \brief Use output stream \c out_file  for text output 

        This constructor assumes that the I/O properties of 
	\c out_file  have already been set.

	Note that the stream \c out_file  should not have 
	been opened in binary mode, and errors will likely occur
	if this is the case.

	\todo Ensure streams are not opened in binary mode
	for safety.
    */
    text_out_file(std::ostream *out_file, int width=80);
    
    /** \brief Create an output file with name \c file_name  
	
        If \c prop  is not null, then the I/O properties
	(precision, fill, flags, etc) for the newly created file are
	taken to be the same as \c prop .
	
    */
    text_out_file(std::string file_name, std::ostream *prop=NULL,
		  bool append=false, int width=80);

    virtual ~text_out_file();

    /** \brief Output a bool variable
     */
    virtual int bool_out(bool dat, std::string name="");

    /** \brief Output a char variable
     */
    virtual int char_out(char dat, std::string name="");

    /** \brief Output a char variable
     */
    virtual int char_out_internal(char dat, std::string name="");

    /** \brief Output a double variable
     */
    virtual int double_out(double dat, std::string name="");

    /** \brief Output a float variable
     */
    virtual int float_out(float dat, std::string name="");

    /** \brief Output an int variable
     */
    virtual int int_out(int dat, std::string name="");

    /** \brief Output an long variable
     */
    virtual int long_out(unsigned long int dat, std::string name="");

    /** \brief Output a string
     */
    virtual int string_out(std::string dat, std::string name="");

    /** \brief Output a word
     */
    virtual int word_out(std::string dat, std::string name="");
    
    /** \brief Start object output
     */
    virtual int start_object(std::string type, std::string name);

    /** \brief End object output
     */
    virtual int end_object();

    /** \brief End line
     */
    virtual int end_line();

    /** \brief Output initialization */
    virtual int init_file();

    virtual int clean_up() {
      return 0;
    }
    
    /** \brief Output a comment (only for text files) */
    int comment_out(std::string comment);

    /// Add brackets and replace carriage returns with spaces
    std::string reformat_string(std::string in);

#ifndef DOXYGEN_INTERNAL

  protected:

    friend class collection;

    /** \brief A list to indicate if the current object and subobjects 
	are "hard-coded"
    */
    std::stack<bool> hcs;

    /// True if the constructor was called with a string, false otherwise
    bool from_string;

    /// True if the file is to be compressed
    bool compressed;

    /// True if the file is to be compressed with gzip
    bool gzip;

    /// The width of the file
    int file_width;

    /// The output stream
    std::ostream *outs;

    /// A pointer to an output stream to define output properties
    std::ostream *props;

    /// The temporary buffer as a stringstream
    std::ostringstream *strout;

    /// The user-specified filename
    std::string user_filename;

    /// The temporary filename used
    std::string temp_filename;

    /// Flush the string buffer
    virtual int flush();

    /// If true, then \c type is a "hard-coded" type
    bool is_hc_type(std::string type);

#endif
			   
  };

  /** 
      \brief An input text file

      <b>Note:</b> Text files are not entirely architecture-independent,
      For example, a larger integer will not be read correctly on small 
      integer systems.
   */
  class text_in_file : public in_file_format {
  public:

    /** \brief Use input stream \c in_file for text input */
    text_in_file(std::istream *in_file);

    /** \brief Read an input file with name \c file_name  */
    text_in_file(std::string file_name);

    virtual ~text_in_file();
  
    virtual int bool_in(bool &dat, std::string name="");
    virtual int char_in(char &dat, std::string name="");
    virtual int double_in(double &dat, std::string name="");
    virtual int float_in(float &dat, std::string name="");
    virtual int int_in(int &dat, std::string name="");
    virtual int long_in(unsigned long int &dat, std::string name="");
    virtual int string_in(std::string &dat, std::string name="");
    virtual int word_in(std::string &dat, std::string name="");
    
    virtual int start_object(std::string &type, std::string &name);
    virtual int skip_object();
    virtual int end_object();

    /** \brief Initialize file input */
    virtual int init_file();
    
    /// Finish file input
    virtual int clean_up() {
      return 0;
    }

    /// Add brackets and replace carriage returns with spaces
    std::string reformat_string(std::string in);

#ifndef DOXYGEN_INTERNAL

  protected:

    /// If true, then \c type is a "hard-coded" type
    bool is_hc_type(std::string type);

    /** \brief A list to indicate if the current object and subobjects 
	are "hard-coded"
    */
    std::stack<bool> hcs;

    /// The input stream
    std::istream *ins;

    /// True if the string version of the constructor was called
    bool from_string;

    /// A version of word_in() which doesn't call the error handler
    virtual int word_in_noerr(std::string &dat, std::string name="") {
      std::string tmp;
      if (name.length()>0) (*ins) >> tmp;
      if ((*ins) >> dat) {
	return 0;
      }
      return -1;
    }
    
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
