/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_DERIV_H
#define O2SCL_DERIV_H

#include <iostream>
#include <cmath>
#include <o2scl/collection.h>
#include <o2scl/funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Numerical differentiation base
      
      This base class does not perform any actual differentiation.
      Use one of the children cern_deriv, gsl_deriv, or eqi_deriv
      instead.

      This base class contains some code to automatically apply
      the first derivative routines to compute second or third
      derivatives. The error estimates for these will likely
      be underestimated.
      
      \note Because this class template aims to automatically provide
      second and third derivatives, one must overload either both
      calc() and calc_int() or both calc_err() and calc_err_int().

      \future Improve the methods for second and third derivatives

  */
  template<class param_t, class func_t> class deriv {

#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    /// A structure for passing the function to second and third derivatives
    typedef struct {
      public:
      /// The pointer to the function
      func_t *func;
      /// The pointer to the user-specified parameters
      param_t *up;
    } dpars;
    
    /// Avoids infinite loops in case the user calls the base class version
    bool from_calc;

#endif

  public:

    deriv() {
      verbose=0;
      from_calc=false;
    }

    virtual ~deriv() {}

    /** 
	\brief Calculate the first derivative of \c func w.r.t. x
	
	After calling calc(), the error may be obtained from 
	\ref get_err().
    */
    virtual double calc(double x, param_t &pa, func_t &func) {
      double dx;
      from_calc=true;
      calc_err(x,pa,func,dx,derr);
      from_calc=false;
      return dx;
    }

    /** \brief Calculate the second derivative of \c func w.r.t. x
     */
    virtual double calc2(double x, param_t &pa, func_t &func) {
      double val;
      dpars dp={&func,&pa};
      funct_mfptr_noerr<deriv,dpars> mf(this,&deriv::derivfun);
      val=calc_int(x,dp,mf);
      return val;
    }
  
    /** \brief Calculate the third derivative of \c func w.r.t. x
     */
    virtual double calc3(double x, param_t &pa, func_t &func) {
      double val;
      dpars dp={&func,&pa};
      funct_mfptr_noerr<deriv,dpars> mf(this,&deriv::derivfun2);
      val=calc_int(x,dp,mf);
      return val;
    }

    /** \brief Get uncertainty of last calculation
     */
    virtual double get_err() {
      return derr;
    }

    /** \brief Output control
     */
    int verbose;
    
    /** \brief Calculate the first derivative of \c func w.r.t. x and the
	uncertainty
    */
    virtual int calc_err(double x, param_t &pa, func_t &func, double &dfdx, 
			 double &err) {
      if (from_calc==true) {
	set_err_ret("No method specified in calc_err().",gsl_nobase);
      }
      dfdx=calc(x,pa,func);
      err=0.0;
      return 0;
    }

    /** \brief Calculate the second derivative of \c func w.r.t. x and the
	uncertainty
    */
    virtual int calc2_err(double x, param_t &pa, func_t &func, 
			  double &d2fdx2, double &err) {
      int ret;
      dpars dp={&func,&pa};
      funct_mfptr_noerr<deriv,dpars> mf(this,&deriv::derivfun);
      ret=calc_err_int(x,dp,mf,d2fdx2,err);
      return 0;
    }
    
    /** \brief Calculate the third derivative of \c func w.r.t. x and the
	uncertainty
    */
    virtual int calc3_err(double x, param_t &pa, func_t &func, 
			  double &d3fdx3, double &err) {
      int ret;
      dpars dp={&func,&pa};
      funct_mfptr_noerr<deriv,dpars> mf(this,&deriv::derivfun2);
      ret=calc_err_int(x,dp,mf,d3fdx3,err);
      return 0;
    }
    
    /// Return string denoting type ("deriv")
    virtual const char *type() { return "deriv"; }
    
  protected:

#ifndef DOXYGEN_INTERNAL

    /** 
	\brief Calculate the first derivative of \c func w.r.t. x

	This is an internal version of calc() which is used in
	computing second and third derivatives
     */
    virtual double calc_int(double x, dpars &pa, o2scl::funct<dpars> &func) {
      double dx;
      from_calc=true;
      calc_err_int(x,pa,func,dx,derr);
      from_calc=false;
      return dx;
    }

    /** \brief Calculate the first derivative of \c func w.r.t. x and the
	uncertainty

	This is an internal version of calc_err() which is used in
	computing second and third derivatives
    */
    virtual int calc_err_int(double x, dpars &pa, o2scl::funct<dpars> &func, 
			     double &dfdx, double &err) {
      if (from_calc==true) {
	set_err_ret("No method specified in calc_err_int().",gsl_nobase);
      }
      dfdx=calc_int(x,pa,func);
      err=0.0;
      return 0;
    }
    
    /// The uncertainity in the most recent derivative computation
    double derr;

    /// The function for the second derivative
    double derivfun(double x, dpars &dp) {
      return calc(x,*(dp.up),*(dp.func));
    }

    /// The function for the third derivative
    double derivfun2(double x, dpars &dp) {
      funct_mfptr_noerr<deriv,dpars> mf(this,&deriv::derivfun);
      double val=calc_int(x,dp,mf);
      return val;
    }
    
#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif
