/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_DERIV_H
#define O2SCL_GSL_DERIV_H

#include <iostream>
#include <o2scl/deriv.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_errno.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Numerical differentiation (GSL)

      This class computes the numerical derivative of a function. The
      stepsize \ref h should be specified before use. If similar
      functions are being differentiated in succession, the user may
      be able to increase the speed of later derivatives by setting
      the new stepsize equal to the optimized stepsize from the
      previous differentiation, by setting \ref h to \ref h_opt.

      \note Second and third derivatives are computed by naive nested
      applications of the formula for the first derivative and the
      uncertainty for these will likely be underestimated.

      \future Include the forward and backward GSL derivatives

  */
  template<class param_t, class func_t>
    class gsl_deriv : public deriv<param_t,func_t> {
  public:

    gsl_deriv() {
      h=0.0;
      h_opt=0.0;
    }

    virtual ~gsl_deriv() {}
    
    /** 
	\brief Initial stepsize 
	
	This should be specified before a call to calc() or
	calc_err(). If it is zero, then \f$ x 10^{-4} \f$ will used, or
	if \c x is zero, then \f$ 10^{-4} \f$ will be used.
    */
    double h;

    /** 
	\brief The last value of the optimized stepsize

	This is initialized to zero in the constructor and 
	set by calc_err() to the most recent value of the 
	optimized stepsize. 
    */
    double h_opt;

    /** \brief Calculate the first derivative of \c func  w.r.t. x and 
	uncertainty
    */
    virtual int calc_err(double x, param_t &pa, func_t &func,
			 double &dfdx, double &err) {
      
      double hh;
      if (h==0.0) {
	if (x==0.0) hh=1.0e-4;
	else hh=1.0e-4*x;
      } else {
	hh=h;
      }

      double r_0, round, trunc, error;

      central_deriv(x,h,r_0,round,trunc,func,pa);
      error = round + trunc;
      
      if (round < trunc && (round > 0 && trunc > 0)) {
	double r_opt, round_opt, trunc_opt, error_opt;
	
	/* Compute an optimised stepsize to minimize the total error,
	   using the scaling of the truncation error (O(h^2)) and
	   rounding error (O(1/h)). */
	
	h_opt = h * pow (round / (2.0 * trunc), 1.0 / 3.0);
	central_deriv(x,h_opt,r_opt,round_opt,trunc_opt,func,pa);
	error_opt = round_opt + trunc_opt;
	
	/* Check that the new error is smaller, and that the new derivative
	   is consistent with the error bounds of the original estimate. */
	
	if (error_opt < error && fabs (r_opt - r_0) < 4.0 * error) {
	  r_0 = r_opt;
	  error = error_opt;
	}
      }
      
      dfdx=r_0;
      err=error;
      
      return 0;
    }

    /// Return string denoting type ("gsl_deriv")
    virtual const char *type() { return "gsl_deriv"; }

#ifndef DOXYGEN_INTERNAL

  protected:

    /** \brief Internal version of calc_err() for second
	and third derivatives
    */
    virtual int calc_err_int
      (double x, typename deriv<param_t,func_t>::dpars &pa, 
       funct<typename deriv<param_t,func_t>::dpars> &func, 
       double &dfdx, double &err) {
      
      err_hnd->reset();
      
      double hh;
      if (h==0.0) {
	if (x==0.0) hh=1.0e-4;
	else hh=1.0e-4*x;
      } else {
	hh=h;
      }
      
      double r_0, round, trunc, error;

      central_deriv(x,h,r_0,round,trunc,func,pa);
      error = round + trunc;
      
      if (round < trunc && (round > 0 && trunc > 0)) {
	double r_opt, round_opt, trunc_opt, error_opt;
	
	/* Compute an optimised stepsize to minimize the total error,
	   using the scaling of the truncation error (O(h^2)) and
	   rounding error (O(1/h)). 
	*/
	
	// This value 'hopt' is different from 'h_opt' above so that
	// the two derivative functions don't get confused about which
	// 'hopt' they are using.
	double hopt = h * pow (round / (2.0 * trunc), 1.0 / 3.0);
	central_deriv(x,hopt,r_opt,round_opt,trunc_opt,func,pa);
	error_opt = round_opt + trunc_opt;
	
	/* Check that the new error is smaller, and that the new derivative
	   is consistent with the error bounds of the original estimate. 
	*/
	
	if (error_opt < error && fabs (r_opt - r_0) < 4.0 * error) {
	  r_0 = r_opt;
	  error = error_opt;
	}
      }
      
      dfdx=r_0;
      err=error;
      
      return 0;
    }
    
    /** 
	\brief Compute derivative using 5-point rule
	
	Compute the derivative using the 5-point rule (x-h, x-h/2, x,
	x+h/2, x+h) and the error using the difference between the
	5-point and the 3-point rule (x-h,x,x+h). Note that the
	central point is not used for either.

	This must be a class template because it is used by
	both calc_err() and calc_err_int().
    */
    template<class func2_t, class param2_t> 
      int central_deriv(double x, double hh, double &result, 
			double &abserr_round, double &abserr_trunc, 
			func2_t &func, param2_t &pa) {
      
      double fm1, fp1, fmh, fph;
      
      func(x-hh,fm1,pa);
      func(x+hh,fp1,pa);

      func(x-hh/2,fmh,pa);
      func(x+hh/2,fph,pa);

      double r3 = 0.5 * (fp1 - fm1);
      double r5 = (4.0 / 3.0) * (fph - fmh) - (1.0 / 3.0) * r3;
      
      double e3 = (fabs (fp1) + fabs (fm1)) * GSL_DBL_EPSILON;
      double e5 = 2.0 * (fabs (fph) + fabs (fmh)) * GSL_DBL_EPSILON + e3;
      
      /* The next term is due to finite precision in x+h = O (eps * x) */
      
      double dy=GSL_MAX(fabs(r3/hh),fabs(r5/hh))*fabs(x/hh)*GSL_DBL_EPSILON;
      
      /* The truncation error in the r5 approximation itself is O(h^4).
	 However, for safety, we estimate the error from r5-r3, which is
	 O(h^2).  By scaling h we will minimise this estimated error, not
	 the actual truncation error in r5. 
      */
      
      result = r5 / hh;
      /* Estimated truncation error O(h^2) */
      abserr_trunc = fabs ((r5 - r3) / hh); 
      /* Rounding error (cancellations) */
      abserr_round = fabs (e5 / hh) + dy;   
      
      if (this->verbose>0) {
	std::cout << "gsl_deriv: " << std::endl;
	std::cout << "step: " << hh << std::endl;
	std::cout << "abcissas:  " << x-hh/2 << " " << x-hh << " " 
		  << x+hh/2 << " " << x+hh << std::endl;
	std::cout << "ordinates: " << fm1 << " " << fmh << " " << fph << " " 
		  << fp1 << std::endl;
	std::cout << "res: " << result << " trc: " << abserr_trunc 
		  << " rnd: " << abserr_round << std::endl;
      }
      
      return 0;
    }
    
#endif

  };

#ifndef DOXYGENP
}
#endif

#endif



