/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CFL6_EOS_H
#define O2SCL_CFL6_EOS_H

#include <iostream>
#include <o2scl/cfl_njl_eos.h>
#include <o2scl/columnify.h>
#include <o2scl/omatrix_cx_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief CFL NJL EOS with a color-superconducting 't Hooft interaction

      Beginning with the Lagrangian:
      \f[
      {\cal L} = {\cal L}_{Dirac} + {\cal L}_{NJL} + 
      {\cal L}_{'t Hooft} + {\cal L}_{SC} + {\cal L}_{SC6} 
      \f]
      \f[
      {\cal L}_{Dirac} = {\bar q} \left( i \partial -m - 
      \mu \gamma^0 \right) q
      \f]
      \f[
      {\cal L}_{NJL} = G_S \sum_{a=0}^8 
      \left[ \left( {\bar q} \lambda^a q \right)^2
      - \left( {\bar q} \lambda^a \gamma^5 q \right)^2 \right]
      \f]
      \f[
      {\cal L}_{'t Hooft} = G_D \left[
      \mathrm{det}_f {\bar q} \left(1-\gamma^5 \right) q
      +\mathrm{det}_f {\bar q} \left(1+\gamma^5 \right) q
      \right]
      \f]
      \f[
      {\cal L}_{SC} = G_{DIQ}
      \left( {\bar q}_{i \alpha} i \gamma^5 
      \varepsilon^{i j k} \varepsilon^{\alpha \beta \gamma} 
      q^C_{j \beta} \right)
      \left( {\bar q}_{\ell \delta} i \gamma^5 
      \epsilon^{\ell m k} 
      \epsilon^{\delta \varepsilon \gamma}
      q^C_{m \varepsilon} \right)
      \f]
      \f[
      {\cal L}_{SC6} = K_D
      \left( {\bar q}_{i \alpha} i \gamma^5 
      \varepsilon^{i j k} \varepsilon^{\alpha \beta \gamma} 
      q^C_{j \beta} \right)
      \left( {\bar q}_{\ell \delta} i \gamma^5 
      \epsilon^{\ell m n} 
      \epsilon^{\delta \varepsilon \eta}
      q^C_{m \varepsilon} \right)
      \left( {\bar q}_{k \gamma} q_{n \eta} \right)
      \f]

      We can simplify the relevant terms in \f${\cal L}_{NJL}\f$:
      \f[
      {\cal L}_{NJL} = G_S \left[ 
      \left({\bar u} u\right)^2+
      \left({\bar d} d\right)^2+
      \left({\bar s} s\right)^2
      \right]
      \f]
      and in \f${\cal L}_{'t Hooft}\f$:
      \f[
      {\cal L}_{NJL} = G_D \left( 
      {\bar u} u {\bar d} d {\bar s} s
      \right)
      \f]

      Using the definition:
      \f[
      \Delta^{k \gamma} =  \left< {\bar q} i \gamma^5 
      \epsilon \epsilon q^C_{} \right>
      \f]
      and the ansatzes:
      \f[
      ({\bar q}_1 q_2) ({\bar q}_3 q_4) \rightarrow
      {\bar q}_1 q_2 \left< {\bar q}_3 q_4 \right>
      +{\bar q}_3 q_4 \left< {\bar q}_1 q_2 \right>
      -\left< {\bar q}_1 q_2 \right> \left< {\bar q}_3 q_4 \right>
      \f]
      \f[
      ({\bar q}_1 q_2) ({\bar q}_3 q_4) ({\bar q}_5 q_6) \rightarrow
      {\bar q}_1 q_2 \left< {\bar q}_3 q_4 \right> 
      \left< {\bar q}_5 q_6 \right>
      +{\bar q}_3 q_4 \left< {\bar q}_1 q_2 \right>
      \left< {\bar q}_5 q_6 \right>
      +{\bar q}_5 q_6 \left< {\bar q}_1 q_2 \right>
      \left< {\bar q}_3 q_4 \right>
      -2\left< {\bar q}_1 q_2 \right> \left< {\bar q}_3 q_4 \right>
      \left< {\bar q}_5 q_6 \right>
      \f]
      for the mean field approximation, we can rewrite the Lagrangian
      \f[
      {\cal L}_{NJL} = 2 G_S \left[ 
      \left( {\bar u} u \right) \left< {\bar u} u \right>
      +\left( {\bar d} d \right) \left< {\bar d} d \right>
      +\left( {\bar s} s \right) \left< {\bar s} s \right>
      - \left< {\bar u} u \right>^2
      - \left< {\bar d} d \right>^2
      - \left< {\bar s} s \right>^2
      \right]
      \f]
      \f[
      {\cal L}_{'t Hooft} = - 2 G_D \left[
      \left( {\bar u} u \right) \left< {\bar u} u \right>
      \left< {\bar s} s \right>
      + \left( {\bar d} d \right) \left< {\bar u} u \right>
      \left< {\bar s} s \right>
      + \left( {\bar s} s \right) \left< {\bar u} u \right>
      \left< {\bar d} d \right>
      - 2 \left< {\bar u} u \right>\left< {\bar d} d \right>
      \left< {\bar s} s \right>
      \right]
      \f]
      \f[
      {\cal L}_{SC} = G_{DIQ} \left[
      \Delta^{k \gamma}
      \left( {\bar q}_{\ell \delta} i \gamma^5 
      \epsilon^{\ell m k} 
      \epsilon^{\delta \varepsilon \gamma}
      q^C_{m \varepsilon} \right)
      + \left( {\bar q}_{i \alpha} i \gamma^5 
      \varepsilon^{i j k} \varepsilon^{\alpha \beta \gamma} 
      q^C_{j \beta} \right)
      \Delta^{k \gamma \dagger}
      - \Delta^{k \gamma}
      \Delta^{k \gamma \dagger}
      \right]
      \f]
      \f[
      {\cal L}_{SC6} = K_D \left[
      \left( {\bar q}_{m \varepsilon} q_{n \eta} \right)
      \Delta^{k \gamma} \Delta^{m \varepsilon \dagger}
      + \left( {\bar q}_{i \alpha} i \gamma^5 
      \varepsilon^{i j k} \varepsilon^{\alpha \beta \gamma} 
      q^C_{j \beta} \right)
      \Delta^{m \varepsilon \dagger} 
      \left< {\bar q}_{m \varepsilon} q_{n \eta} \right>
      \right]
      \f]
      \f[
      + K_D \left[\Delta^{k \gamma}
      \left( {\bar q}_{\ell \delta} i \gamma^5 
      \epsilon^{\ell m n} 
      \epsilon^{\delta \varepsilon \eta}
      q^C_{m \varepsilon} \right)
      \left< {\bar q}_{m \varepsilon} q_{n \eta} \right>
      -2 
      \Delta^{k \gamma} \Delta^{m \varepsilon \dagger}
      \left< {\bar q}_{m \varepsilon} q_{n \eta} \right>
      \right]
      \f]
      
      If we make the definition \f$ {\tilde \Delta} =
      2 G_{DIQ} \Delta \f$

  */
  class cfl6_eos : public cfl_njl_eos {
  public:
  
    cfl6_eos();

    virtual ~cfl6_eos();
  
    /** 
	\brief Calculate the EOS
      
	Calculate the EOS from the quark condensates.  Return the mass
	gap equations in \c qq1, \c qq2, \c qq3, and the normal gap
	equations in \c gap1, \c gap2, and \c gap3.
	
	Using \c fromqq=true as in nambujl_eos and nambujl_temp_eos does
	not work here and will return an error. The quarks must be set
	through quark_eos::quark_set() before use.
	
	If all of the gaps are less than gap_limit, then the
	nambujl_temp_eos::calc_temp_p() is used, and \c gap1, \c gap2,
	and \c gap3 are set to equal \c u.del, \c d.del, and \c s.del,
	respectively.
      
    */
    virtual int calc_eq_temp_p(quark &u, quark &d, quark &s, 
			    double &qq1, double &qq2, double &qq3, 
			    double &gap1, double &gap2, double &gap3, 
			    double mu3, double mu8,
			    double &n3, double &n8, thermo &qb, 
			    const double temper);

    /** \brief Direct calculation of the thermodynamic potential
     */
    virtual double thd_potential(quark &u, quark &d, quark &s,
				 double mu3, double mu8, 
				 const double ltemper);


    /** 
	\brief Calculate the energy eigenvalues as a function of the momentum

	Given the momentum 'mom', and the chemical potentials
	associated with the third and eighth gluons ('mu3' and 'mu8'),
	the energy eigenvalues are computed in egv[0] ... egv[35].  No
	space is allocated for the array by the function.

	\todo This function may make some inappropriate assumptions 
	on the vector egv.
    */
    virtual int eigenvalues(double mom, ovector_view &egv, double mu3, 
			    double mu8);

    /// The color superconducting 't Hooft coupling (default 0)
    double KD;

    /// Return string denoting type ("cfl6_eos")
    virtual const char *type() { return "cfl6_eos"; };

    /** \brief The absolute value below which the CSC 't Hooft coupling 
	is ignored(default \f$ 10^{-6} \f$)
    */
    double kdlimit;

  protected:

#ifndef DOXYGEN_INTERNAL

    /// To clear all of the matrix entries
    gsl_complex zero;
    
    /// The size of the matrix to be diagonalized
    static const int size=36;
    
    /// Storage for the inverse propagator
    omatrix_cx m6;

    /// Storage for the eigenvalues
    ovector eval6;

    /// GSL workspace for the eigenvalue computation
    gsl_eigen_herm_workspace *w6;

    /** \brief The function used to take derivatives of the thermodynamic
	potential (used by calc_eq_temp_p() )
    */
    virtual double tpot(double var, void *&pa);
    
  private:

    cfl6_eos(const cfl6_eos &);
    cfl6_eos& operator=(const cfl6_eos&);

#endif

  };

#ifndef DOXYGENP
}
#endif

#endif
