/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/cfl_njl_eos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

cfl_njl_eos::cfl_njl_eos() {

  // Allocate storage for computation of the eigenvalues
  
  m.allocate(12,12);
  eval.allocate(12);
  w=gsl_eigen_herm_alloc(12);

  m2.allocate(6,6);
  eval2.allocate(6);
  w2=gsl_eigen_herm_alloc(6);

  gsl_complex zero;
  GSL_REAL(zero)=0.0;
  GSL_IMAG(zero)=0.0;
  m.set_all(zero);
  m2.set_all(zero);
  
  // Other initializations

  quartic_set=false;
  quartic=&def_quartic;

  gap_limit=0.0;
  zerot=false;
  fixed_mass=false;
  faster=false;
  fromqq=true;

  smu3=0.0;
  smu8=0.0;
  var=0;
  
  min1=&def_min1;
  min2=&def_min2;

  use_solver=true;
}

cfl_njl_eos::~cfl_njl_eos() {
  m.free();
  eval.free();
  gsl_eigen_herm_free(w);

  m2.free();
  eval2.free();
  gsl_eigen_herm_free(w2);
}

int cfl_njl_eos::set_parameters(double lambda, double fourferm, 
				double sixferm, double fourgap) {
  int ret;

  if (fixed_mass) {

    L=lambda;
    G=0.0;
    K=0.0;
    GD=fourgap;
    
  } else {

    // Run the parent method first
    ret=nambujl_eos::set_parameters(lambda,fourferm,sixferm);

    if (fourgap==0.0) {
      GD=3.0*G/4.0;
    } else {
      GD=fourgap;
    }
  }

  return 0;
}

int cfl_njl_eos::calc_temp_p(quark &u, quark &d, quark &s, 
			     const double lT, thermo &th) {
  void *vp=0;
  int ret=0;
  
  // Set particle and thermo objects
  up=&u;
  down=&d;
  strange=&s;
  eos_thermo=&th;

  // Decide whether or not to solve gap equations
  ugap_solve=false;
  dgap_solve=false;
  sgap_solve=false;
  if (u.del>gap_limit) ugap_solve=true;
  if (d.del>gap_limit) dgap_solve=true;
  if (s.del>gap_limit) sgap_solve=true;

  // If there's no work to be done, just call calc_eq_temp_p() and 
  // exit
  if (fixed_mass && !color_neut && !ugap_solve && !dgap_solve &&
      !sgap_solve) {

    // No solution required, just evaluate
    double lmu3=0.0, lmu8=0.0, qq1, qq2, qq3, gap1, gap2, gap3, n3, n8;
    ret=calc_eq_temp_p(u,d,s,qq1,qq2,qq3,gap1,gap2,gap3,lmu3,lmu8,
		       n3,n8,th,lT);
    solver_qual=0.0;
    
    return ret;
  }

  if (use_solver) {
    
    // Compute number of variables
    size_t nv=0;
    if (!fixed_mass) nv+=3;
    if (color_neut) nv+=2;
    if (ugap_solve) nv++;
    if (dgap_solve) nv++;
    if (sgap_solve) nv++;
    
    ovector x(nv), y(nv);
    
    // Provide initial guesses
    size_t ix=0;
    if (!fixed_mass) {
      x[ix]=u.qq; ix++;
      x[ix]=d.qq; ix++;
      x[ix]=s.qq; ix++;
    }
    if (color_neut) {
      x[ix]=smu3; ix++;
      x[ix]=smu8; ix++;
    }
    if (ugap_solve) { x[ix]=u.del; ix++; }
    if (dgap_solve) { x[ix]=d.del; ix++; }
    if (sgap_solve) { x[ix]=s.del; ix++; }
    
    // Solve
    mm_funct_mfptr<cfl_njl_eos,void *> mf(this,&cfl_njl_eos::solve_fun_p);
    ret=solver->msolve(nv,x,vp,mf);

    // Evaluate quality
    solve_fun_p(nv,x,y,vp);
    solver_qual=0.0;
    for(size_t iy=0;iy<nv;iy++) solver_qual+=y[iy]*y[iy];
    solver_qual=sqrt(solver_qual);

  } else {
    
    ovector x(2);
    double y;
    
    if (color_neut) {
      
      x[0]=0.0;
      x[1]=0.0;
      
      multi_funct_mfptr<cfl_njl_eos,void *> mfm(this,&cfl_njl_eos::min_fun2);
      min2->mmin(2,x,y,vp,mfm);
      
    } else {
      
      x[0]=0.0;
      x[1]=0.0;

      min_fun2(2,x,y,vp);

    }
    
    // After minimization is performed, evaluate the densities, etc.
    
    double lmu3=x[0], lmu8=x[1], qq1, qq2, qq3, gap1, gap2, gap3, n3, n8;
    ret=calc_eq_temp_p(u,d,s,qq1,qq2,qq3,gap1,gap2,gap3,lmu3,lmu8,
		       n3,n8,th,lT);
    solver_qual=0.0;

  }
  
  return ret;
} 


int cfl_njl_eos::calc_eq_temp_p(quark &u, quark &d, quark &s,
				double &qq1, double &qq2, double &qq3, 
				double &gap1, double &gap2, double &gap3, 
				double mu3, double mu8, 
				double &n3, double &n8, 
				thermo &qb, const double ltemper) {

  up=&u;
  down=&d;
  strange=&s;

  if (fromqq==false) {
    set_err("cfl_njl_eos does not work with fromqq=false.",1);
    return 1;
  }

  if (fixed_mass) {
    up->ms=up->m;
    down->ms=down->m;
    strange->ms=strange->m;
  } else {
    up->ms=up->m-4.0*G*up->qq+2.0*K*down->qq*strange->qq;
    down->ms=down->m-4.0*G*down->qq+2.0*K*up->qq*strange->qq;
    strange->ms=strange->m-4.0*G*strange->qq+2.0*K*down->qq*up->qq;
  }
  
  // static variables for communication with tpot
  temper=ltemper;
  smu3=mu3;
  smu8=mu8;

  // In the case that the gaps are too small, go ahead
  // and force them to zero by solving the old gapless equations
  if ((up->del)<gap_limit && (down->del)<gap_limit && 
      (strange->del)<gap_limit) {
    int ret;
    ret=nambujl_eos::calc_eq_temp_p(*up,*down,*strange,
				    qq1,qq2,qq3,qb,temper);
    gap1=up->del;
    gap2=down->del;
    gap3=strange->del;
    return ret;
  }

  // Quark densities
  up->n=-derivfn(&up->mu);
  down->n=-derivfn(&down->mu);
  strange->n=-derivfn(&strange->mu);

  // We add 1 here to cancel the extra factors added to tpot.
  // This seems to help the numerical derivative when n3 and
  // n8 are close to zero.
  n3=-derivfn(&smu3)+1.0;
  n8=-derivfn(&smu8)+1.0;

  if (fixed_mass) {
    qq1=up->qq;
    qq2=down->qq;
    qq3=strange->qq;
  } else {
    qq1=derivfn(&up->qq);
    qq2=derivfn(&down->qq);
    qq3=derivfn(&strange->qq);
  }

  // Gaps
  if (fabs(up->del)>gap_limit) {
    gap1=derivfn(&up->del);
  } else {
    gap1=up->del;
  }
  if (fabs(down->del)>gap_limit) gap2=derivfn(&down->del);
  else gap2=down->del;
  if (fabs(strange->del)>gap_limit) gap3=derivfn(&strange->del);
  else gap3=strange->del;

  // Calculate pressure and entropy after the calculation of the quark
  // condensates, which are needed for the final evaluation of the
  // pressure (9/30/04 - I think this comment is moot, since fromqq is
  // now always true)

  double dtemp;
  var=&dtemp;
  void *vpx=0;
  qb.pr=-tpot(temper,vpx);

  funct_mfptr_noerr<cfl_njl_eos,void *> 
    f(this,&cfl_njl_eos::entropy_integrand);
  qb.en=it->integ(f,0.0,L,vpx);
  
  up->ed=-up->pr+up->n*up->mu;
  down->ed=-down->pr+down->n*down->mu;
  strange->ed=-strange->pr+strange->n*strange->mu;
  qb.ed=qb.en*temper-qb.pr+up->mu*up->n+down->mu*down->n+
    strange->mu*strange->n;

  return 0;

}

double cfl_njl_eos::derivfn(double *dvar) {
  double temp, ret, t1, t2;

  var=dvar;
  temp=*var;

  funct_mfptr_noerr<cfl_njl_eos,void *> fmn(this,&cfl_njl_eos::tpot);
  
  de.h=fabs(temp/5.0);
  if (de.h<1.0e-4) de.h=1.0e-1;
  void *vpx=0;
  ret=de.calc(temp,vpx,fmn);
  
  // Return the variable back to the original just in case the
  // derivative routine changed it and return the matrix to its
  // original form
  *dvar=temp;

  return ret;
}

double cfl_njl_eos::tpot(double tvar, void *&pa) {
  double tp;
  
  if (var!=0) *var=tvar;
  
  if (fixed_mass) {
    up->ms=up->m;
    down->ms=down->m;
    strange->ms=strange->m;
  } else {
    up->ms=up->m-4.0*G*up->qq+2.0*K*down->qq*strange->qq;
    down->ms=down->m-4.0*G*down->qq+2.0*K*up->qq*strange->qq;
    strange->ms=strange->m-4.0*G*strange->qq+2.0*K*down->qq*up->qq;
  }

  funct_mfptr_noerr<cfl_njl_eos,void *> f(this,&cfl_njl_eos::tpot_integrand);
  void *vpx=0;
  tp=it->integ(f,0.0,L,vpx);
  
  if (fixed_mass) {
    tp+=B0;
  } else {
    tp+=up->qq*up->qq*2.0*G;
    tp+=down->qq*down->qq*2.0*G;
    tp+=strange->qq*strange->qq*2.0*G;
    tp-=up->qq*down->qq*strange->qq*4.0*K-B0;
  }

  tp+=up->del*up->del/4.0/GD;
  tp+=down->del*down->del/4.0/GD;
  tp+=strange->del*strange->del/4.0/GD;
  
  // I've had trouble with some derivative routines when the
  // derivatives are nearly zero. The next two lines force the
  // "color densities" to be non-zero to avoid this difficulty.
  if (var==&smu8) tp+=smu8;
  if (var==&smu3) tp+=smu3;

  return tp;
}

double cfl_njl_eos::entropy_integrand(double p, void *&pa) {
  double e, entropy, temp2;
  int k;
  
  entropy=0.0; 

  if (zerot==false) {
    
    ovector egv(36);
    
    eigenvalues(p,egv,smu3,smu8);
  
    temp2=0.0;

    for (k=0;k<36;k++) {
      e=egv[k];
      if (temper!=0.0 && e/temper<30.0) {
	temp2+=log(1+exp(-e/temper))+e/temper/(1.0+exp(e/temper));
      }
    }
    
    entropy+=temp2;
    entropy*=p*p*0.5/pi2;
  }
  
  return entropy;
}


double cfl_njl_eos::thd_potential(quark &u, quark &d, quark &s,
				  double mu3, double mu8, 
				  const double ltemper) {
  double tpotx;

  up=&u;
  down=&d;
  strange=&s;

  if (fromqq==false) {
    set_err("cfl_njl_eos does not work with fromqq=false.",1);
    return 1;
  }

  temper=ltemper;
  smu3=mu3;
  smu8=mu8;

  if (fixed_mass) {
    up->ms=up->m;
    down->ms=down->m;
    strange->ms=strange->m;
  } else {
    up->ms=up->m-4.0*G*up->qq+2.0*K*down->qq*strange->qq;
    down->ms=down->m-4.0*G*down->qq+2.0*K*up->qq*strange->qq;
    strange->ms=strange->m-4.0*G*strange->qq+2.0*K*down->qq*up->qq;
  }
  
  void *vpx=0;
  funct_mfptr_noerr<cfl_njl_eos,void *> f(this,&cfl_njl_eos::tpot_integrand);
  tpotx=it->integ(f,0.0,L,vpx);

  if (fixed_mass) {
    tpotx+=B0;
  } else {
    tpotx+=up->qq*up->qq*2.0*G;
    tpotx+=down->qq*down->qq*2.0*G;
    tpotx+=strange->qq*strange->qq*2.0*G;
    tpotx-=up->qq*down->qq*strange->qq*4.0*K-B0;
  }

  tpotx+=up->del*up->del/4.0/GD;
  tpotx+=down->del*down->del/4.0/GD;
  tpotx+=strange->del*strange->del/4.0/GD;

  return tpotx;

}

double cfl_njl_eos::tpot_integrand(double p, void *&pa) {
  double temp, e, ret;
  int k;
  ovector egv(36);
  
  eigenvalues(p,egv,smu3,smu8);
  
  // Calculate thermodynamic potential
  temp=0.0;
  for (k = 0; k < 36; k++) {
    e=egv[k];
    if (zerot) temp+=fabs(e/2.0);
    else {
      temp+=e/2.0;
      if ((temper==0.0 && e<0.0) || e/temper<-30.0) {
	temp-=e;
      } else if (temper!=0.0 && e/temper<30.0) {
	temp+=temper*log(1.0+exp(-e/temper));
      }
    }
  }
  
  // Multiply by the integration weights and the 
  // Jacobian (momentum^2)
  ret=-temp*p*p*0.5/pi2;

  return ret;
}

/// Treat the simply gapped quarks in all cases gracefully
int cfl_njl_eos::gapped_eigenvalues(quark *q1, quark *q2, double lmom,
				    double mu1, double mu2, double tdelta,
				    ovector_cx_view &ev) {
  gsl_complex gc;
  GSL_IMAG(gc)=0.0;

  if (q1->ms==0.0 && q2->ms==0.0) {

    if (tdelta<gap_limit) {

      GSL_REAL(gc)=mu1+lmom;
      ev[0]=gc;
      GSL_REAL(gc)=-mu1+lmom;
      ev[1]=gc;
      GSL_REAL(gc)=mu2+lmom;
      ev[2]=gc;
      GSL_REAL(gc)=-mu2+lmom;
      ev[3]=gc;
      
    } else {
      
      GSL_REAL(gc)=0.5*((mu2-mu1)-sqrt((mu1+mu2)*(mu1+mu2)-4.0*lmom*(mu1+mu2)+
				       4.0*lmom*lmom+4.0*tdelta*tdelta));
      ev[0]=gc;
      GSL_REAL(gc)=0.5*((mu2-mu1)+sqrt((mu1+mu2)*(mu1+mu2)-4.0*lmom*(mu1+mu2)+
				       4.0*lmom*lmom+4.0*tdelta*tdelta));
      ev[1]=gc;
      GSL_REAL(gc)=0.5*((mu2-mu1)-sqrt((mu1+mu2)*(mu1+mu2)+4.0*lmom*(mu1+mu2)+
				       4.0*lmom*lmom+4.0*tdelta*tdelta));
      ev[2]=gc;
      GSL_REAL(gc)=0.5*((mu2-mu1)+sqrt((mu1+mu2)*(mu1+mu2)+4.0*lmom*(mu1+mu2)+
				       4.0*lmom*lmom+4.0*tdelta*tdelta));
      ev[3]=gc;
    }

  } else if (tdelta<gap_limit) {
    
    GSL_REAL(gc)=-mu1-sqrt(q1->ms*q1->ms+lmom*lmom);
    ev[0]=gc;
    GSL_REAL(gc)=-mu1+sqrt(q1->ms*q1->ms+lmom*lmom);
    ev[1]=gc;
    GSL_REAL(gc)=-mu2-sqrt(q2->ms*q2->ms+lmom*lmom);
    ev[2]=gc;
    GSL_REAL(gc)=-mu2+sqrt(q2->ms*q2->ms+lmom*lmom);
    ev[3]=gc;
  
  } else {

    double coef1, coef2, coef3, coef4, coef5;

    coef1=1.0;
    coef2=2.0*(mu1-mu2);
    coef3=mu1*mu1-4.0*mu1*mu2+mu2*mu2-2.0*lmom*lmom-2.0*tdelta*tdelta-
      q1->ms*q1->ms-q2->ms*q2->ms;
    coef4=(mu1-mu2)*(-2.0*tdelta*tdelta-2.0*mu1*mu2-
		     2.0*lmom*lmom)+2.0*q1->ms*q1->ms*mu2-
      2.0*q2->ms*q2->ms*mu1;
    coef5=mu1*mu1*mu2*mu2+
      lmom*lmom*(q1->ms*q1->ms+q2->ms*q2->ms+2.0*tdelta*tdelta-
		 mu1*mu1-mu2*mu2)-mu1*mu1*q2->ms*q2->ms-
      mu2*mu2*q1->ms*q1->ms+pow(tdelta,4.0)+pow(lmom,4.0)+
      2.0*q1->ms*q2->ms*tdelta*tdelta+q1->ms*q1->ms*q2->ms*q2->ms+
      2.0*mu1*mu2*tdelta*tdelta;
    
    std::complex<double> a1, a2, a3, a4;
    quartic->solve_rc(coef1,coef2,coef3,coef4,coef5,a1,a2,a3,a4);
    ev.set_stl(0,a1);
    ev.set_stl(1,a2);
    ev.set_stl(2,a3);
    ev.set_stl(3,a4);
  }
  
  ev[4]=gsl_complex_negative(ev[0]);
  ev[5]=gsl_complex_negative(ev[1]);
  ev[6]=gsl_complex_negative(ev[2]);
  ev[7]=gsl_complex_negative(ev[3]);

  return 0;
}

int cfl_njl_eos::eigenvalues(double lmom, ovector_view &egv, double mu3, 
			     double mu8) {
  ovector_cx qev(36);
  gsl_vector *gvp;
  gsl_complex z;
  int k;
  double coef1, coef2, coef3, coef4, coef5, mur, mug, mub;
  quark *u=up, *d=down, *s=strange;
  
  mur=mu3+mu8/sqrt(3.0);
  mug=-mu3+mu8/sqrt(3.0);
  mub=-2.0*mu8/sqrt(3.0);

  // Here and in similar code below, we cast to a void * first to
  // avoid warnings. (This is obviously not the best strategy.)
  gsl_matrix_complex *gm2=(gsl_matrix_complex *)(&m2);
  
  if (faster) {

    double alpha, beta, gamma, delta;
    
    // Experimental faster eigenvalues

    double Delta=s->del;
    double d2=Delta*Delta;
    gamma=d->del/Delta-1.0;
    delta=u->del/Delta-1.0;
    double gd2=gamma*gamma*(d2);
    double dd2=delta*delta*(d2);

    // up-green--down-red 
    double Mu=u->mu+mug;
    double Mu2=Mu*Mu;
    alpha=(d->mu+mur)/Mu-1.0;
    beta=(s->mu+s->ms*s->ms/2.0/s->mu)/Mu-1.0;
    double t1p=lmom*lmom+(d2)+2.0*lmom*Mu+(Mu2);
    double t1m=lmom*lmom+(d2)-2.0*lmom*Mu+(Mu2);
    double st1p=sqrt(t1p);
    double st1m=sqrt(t1m);
    double pdmua=2.0*(lmom*lmom+Delta*Delta)+Mu*Mu*(2.0+alpha);
    
    // up-down
    egv[0]=(pdmua-st1m*alpha*Mu-lmom*(4.0+alpha)*Mu)/2.0/st1m;
    egv[1]=(pdmua-st1p*alpha*Mu+lmom*(4.0+alpha)*Mu)/2.0/st1p;
    egv[2]=(pdmua+st1m*alpha*Mu-lmom*(4.0+alpha)*Mu)/2.0/st1m;
    egv[3]=(pdmua+st1p*alpha*Mu+lmom*(4.0+alpha)*Mu)/2.0/st1p;
    egv[4]=-egv[0];
    egv[5]=-egv[1];
    egv[6]=-egv[2];
    egv[7]=-egv[3];

    // up-blue--strange-red 
    Mu=u->mu+mub;
    Mu2=Mu*Mu;
    alpha=(d->mu)/Mu-1.0;
    beta=(s->mu+mur+s->ms*s->ms/2.0/(s->mu+mur))/Mu-1.0;
    t1p=lmom*lmom+(d2)+2.0*lmom*Mu+(Mu2);
    t1m=lmom*lmom+(d2)-2.0*lmom*Mu+(Mu2);
    st1p=sqrt(t1p);
    st1m=sqrt(t1m);
    pdmua=2.0*(lmom*lmom+Delta*Delta)+Mu*Mu*(2.0+beta);
    
    // up-strange
    egv[8]=(pdmua+st1p*beta*Mu+lmom*(4.0+beta)*Mu)/2.0/st1p;
    egv[9]=(pdmua-st1p*beta*Mu+lmom*(4.0+beta)*Mu)/2.0/st1p;
    egv[10]=(pdmua+st1m*beta*Mu-lmom*(4.0+beta)*Mu)/2.0/st1m;
    egv[11]=(pdmua-st1m*beta*Mu-lmom*(4.0+beta)*Mu)/2.0/st1m;
    egv[12]=-egv[8];
    egv[13]=-egv[9];
    egv[14]=-egv[10];
    egv[15]=-egv[11];

    // down-blue--strange-green
    Mu=u->mu;
    Mu2=Mu*Mu;
    alpha=(d->mu+mub)/Mu-1.0;
    beta=(s->mu+mug+s->ms*s->ms/2.0/(s->mu+mug))/Mu-1.0;
    t1p=lmom*lmom+(d2)+2.0*lmom*Mu+(Mu2);
    t1m=lmom*lmom+(d2)-2.0*lmom*Mu+(Mu2);
    st1p=sqrt(t1p);
    st1m=sqrt(t1m);
    pdmua=2.0*(lmom*lmom+Delta*Delta)+Mu*Mu*(2.0+alpha+beta);
    double tds=(4.0+alpha+beta);

    // down-strange
    egv[16]=(pdmua+st1m*Mu*(alpha+beta)-lmom*tds*Mu)/2.0/st1m;
    egv[17]=(pdmua-st1m*Mu*(alpha+beta)-lmom*tds*Mu)/2.0/st1m;
    egv[18]=(pdmua+st1p*Mu*(alpha+beta)+lmom*tds*Mu)/2.0/st1p;
    egv[19]=(pdmua-st1p*Mu*(alpha+beta)+lmom*tds*Mu)/2.0/st1p;
    egv[20]=-egv[16];
    egv[21]=-egv[17];
    egv[22]=-egv[18];
    egv[23]=-egv[19];

    // up-red--down-green--strange-blue
    Mu=u->mu+mur;
    Mu2=Mu*Mu;
    alpha=(d->mu+mug)/Mu-1.0;
    beta=(s->mu+mub+s->ms*s->ms/2.0/(s->mu+mub))/Mu-1.0;
    t1p=lmom*lmom+(d2)+2.0*lmom*Mu+(Mu2);
    t1m=lmom*lmom+(d2)-2.0*lmom*Mu+(Mu2);
    st1p=sqrt(t1p);
    st1m=sqrt(t1m);

    double t6m=-3.0+gamma+delta;
    double t6p=3.0+gamma+delta;
    double t2=3.0+alpha+beta;
    double t3m=(alpha+beta)*fabs(lmom-Mu);
    double t3p=(alpha+beta)*fabs(lmom+Mu);
    double t4=6.0+alpha+beta;
    double t5=1.0+gamma+delta;
    double t7p=lmom*lmom+4.0*(d2)+2.0*lmom*Mu+(Mu2);
    double t7m=lmom*lmom+4.0*(d2)-2.0*lmom*Mu+(Mu2);
    double st7p=sqrt(t7p);
    double st7m=sqrt(t7m);
    
    egv[24]=(3.0*lmom*lmom+4.0*t6p*d2-lmom*t4*Mu+t2*Mu2)/3.0/st7m;
    egv[25]=(3.0*lmom*lmom-t6m*d2-lmom*t4*Mu+Mu*(t2*Mu-t3m))/3.0/st1m;
    egv[26]=(3.0*lmom*lmom+3.0*t5*d2-lmom*t4*Mu+Mu*(t2*Mu+t3m))/3.0/st1m;
    egv[27]=(3.0*lmom*lmom+4.0*t6p*d2+lmom*t4*Mu+t2*Mu2)/3.0/st7p;
    egv[28]=(3.0*lmom*lmom-t6m*d2+lmom*t4*Mu+Mu*(t2*Mu-t3p))/3.0/st1p;
    egv[29]=(3.0*lmom*lmom+3.0*t5*d2+lmom*t4*Mu+Mu*(t2*Mu+t3p))/3.0/st1p;
    egv[30]=-egv[24];
    egv[31]=-egv[25];
    egv[32]=-egv[26];
    egv[33]=-egv[27];
    egv[34]=-egv[28];
    egv[35]=-egv[29];

  } else {
  
    // The standard 2sc phase:
    if (u->del<gap_limit && d->del<gap_limit) {
      
      gapped_eigenvalues(u,d,lmom,u->mu+mur,d->mu+mug,s->del,qev);
      
      qev.set(8,-(s->mu+mub)+sqrt(lmom*lmom+s->ms*s->ms),0.0);
      qev.set(9,-(s->mu+mub)-sqrt(lmom*lmom+s->ms*s->ms),0.0);
      qev.set(10,(s->mu+mub)+sqrt(lmom*lmom+s->ms*s->ms),0.0);
      qev.set(11,(s->mu+mub)-sqrt(lmom*lmom+s->ms*s->ms),0.0);
      
      for(k=0;k<12;k++) egv[k]=qev.real(k);
    
      // The 2SCds phase:
    } else if (s->del<gap_limit && d->del<gap_limit) {

      gapped_eigenvalues(d,s,lmom,s->mu+mub,d->mu+mug,u->del,qev);

      qev.set(8,-(u->mu+mur)+sqrt(lmom*lmom+u->ms*u->ms),0.0);
      qev.set(9,-(u->mu+mur)-sqrt(lmom*lmom+u->ms*u->ms),0.0);
      qev.set(10,(u->mu+mur)+sqrt(lmom*lmom+u->ms*u->ms),0.0);
      qev.set(11,(u->mu+mur)-sqrt(lmom*lmom+u->ms*u->ms),0.0);
      
      for(k=0;k<12;k++) egv[k]=qev.real(k);

      // The 2SCus phase:
    } else if (u->del<gap_limit && s->del<gap_limit) {

      gapped_eigenvalues(u,s,lmom,u->mu+mur,s->mu+mub,d->del,qev);
    
      qev.set(8,-(d->mu+mug)+sqrt(lmom*lmom+d->ms*d->ms),0.0);
      qev.set(9,-(d->mu+mug)-sqrt(lmom*lmom+d->ms*d->ms),0.0);
      qev.set(10,(d->mu+mug)+sqrt(lmom*lmom+d->ms*d->ms),0.0);
      qev.set(11,(d->mu+mug)-sqrt(lmom*lmom+d->ms*d->ms),0.0);
      
      for(k=0;k<12;k++) egv[k]=qev.real(k);
    
      // The us/ud phase:
    } else if (u->del<gap_limit) {
      
      gsl_matrix_complex_set_zero(gm2);
      
      GSL_IMAG(z)=0.0;
      
      GSL_REAL(z)=-(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm2,0,0,z);
      GSL_REAL(z)=-(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm2,1,1,z);
      GSL_REAL(z)=(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm2,2,2,z);
      GSL_REAL(z)=(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm2,3,3,z);
      GSL_REAL(z)=(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm2,4,4,z);
      GSL_REAL(z)=(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm2,5,5,z);
      
      GSL_REAL(z)=-lmom;
      gsl_matrix_complex_set(gm2,1,0,z);
      gsl_matrix_complex_set(gm2,3,2,z);
      gsl_matrix_complex_set(gm2,5,4,z);
      
      GSL_REAL(z)=0.0;
      GSL_IMAG(z)=s->del;
      gsl_matrix_complex_set(gm2,2,0,z);
      GSL_IMAG(z)=-s->del;
      gsl_matrix_complex_set(gm2,3,1,z);
      GSL_IMAG(z)=d->del;
      gsl_matrix_complex_set(gm2,4,0,z);
      GSL_IMAG(z)=-d->del;
      gsl_matrix_complex_set(gm2,5,1,z);

      gsl_eigen_herm(gm2,(gsl_vector *)(&eval2),w2);
      
      for(k=0;k<6;k++) egv[k]=eval2[k];
    
      gsl_matrix_complex_set_zero(gm2);
      
      GSL_IMAG(z)=0.0;
      
      GSL_REAL(z)=(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm2,0,0,z);
      GSL_REAL(z)=(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm2,1,1,z);
      GSL_REAL(z)=-(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm2,2,2,z);
      GSL_REAL(z)=-(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm2,3,3,z);
      GSL_REAL(z)=-(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm2,4,4,z);
      GSL_REAL(z)=-(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm2,5,5,z);
      
      GSL_REAL(z)=-lmom;
      gsl_matrix_complex_set(gm2,1,0,z);
      gsl_matrix_complex_set(gm2,3,2,z);
      gsl_matrix_complex_set(gm2,5,4,z);
      
      GSL_REAL(z)=0.0;
      GSL_IMAG(z)=s->del;
      gsl_matrix_complex_set(gm2,2,0,z);
      GSL_IMAG(z)=-s->del;
      gsl_matrix_complex_set(gm2,3,1,z);
      GSL_IMAG(z)=d->del;
      gsl_matrix_complex_set(gm2,4,0,z);
      GSL_IMAG(z)=-d->del;
      gsl_matrix_complex_set(gm2,5,1,z);

      gsl_eigen_herm(gm2,(gsl_vector *)(&eval2),w2);
      
      for(k=0;k<6;k++) egv[k+6]=eval2[k];
    
      // The ds/ud phase:
    } else if (d->del<gap_limit) {
      
      gsl_matrix_complex_set_zero(gm2);
      
      GSL_IMAG(z)=0.0;
      
      GSL_REAL(z)=-(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm2,0,0,z);
      GSL_REAL(z)=-(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm2,1,1,z);
      GSL_REAL(z)=(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm2,2,2,z);
      GSL_REAL(z)=(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm2,3,3,z);
      GSL_REAL(z)=-(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm2,4,4,z);
      GSL_REAL(z)=-(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm2,5,5,z);
      
      GSL_REAL(z)=-lmom;
      gsl_matrix_complex_set(gm2,1,0,z);
      gsl_matrix_complex_set(gm2,3,2,z);
      gsl_matrix_complex_set(gm2,5,4,z);
      
      GSL_REAL(z)=0.0;
      GSL_IMAG(z)=s->del;
      gsl_matrix_complex_set(gm2,2,0,z);
      GSL_IMAG(z)=-s->del;
      gsl_matrix_complex_set(gm2,3,1,z);
      GSL_IMAG(z)=-u->del;
      gsl_matrix_complex_set(gm2,4,2,z);
      GSL_IMAG(z)=u->del;
      gsl_matrix_complex_set(gm2,5,3,z);

      gsl_eigen_herm(gm2,(gsl_vector *)(&eval2),w2);

      for(k=0;k<6;k++) egv[k]=eval2[k];

      gsl_matrix_complex_set_zero(gm2);
      
      GSL_IMAG(z)=0.0;
      
      GSL_REAL(z)=(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm2,0,0,z);
      GSL_REAL(z)=(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm2,1,1,z);
      GSL_REAL(z)=-(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm2,2,2,z);
      GSL_REAL(z)=-(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm2,3,3,z);
      GSL_REAL(z)=(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm2,4,4,z);
      GSL_REAL(z)=(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm2,5,5,z);
      
      GSL_REAL(z)=-lmom;
      gsl_matrix_complex_set(gm2,1,0,z);
      gsl_matrix_complex_set(gm2,3,2,z);
      gsl_matrix_complex_set(gm2,5,4,z);
      
      GSL_REAL(z)=0.0;
      GSL_IMAG(z)=-s->del;
      gsl_matrix_complex_set(gm2,2,0,z);
      GSL_IMAG(z)=s->del;
      gsl_matrix_complex_set(gm2,3,1,z);
      GSL_IMAG(z)=u->del;
      gsl_matrix_complex_set(gm2,4,2,z);
      GSL_IMAG(z)=-u->del;
      gsl_matrix_complex_set(gm2,5,3,z);

      gsl_eigen_herm(gm2,(gsl_vector *)(&eval2),w2);

      for(k=0;k<6;k++) egv[k+6]=eval2[k];

      // The us/ds phase:
    } else if (s->del<gap_limit) {
      
      gsl_matrix_complex_set_zero(gm2);
      
      GSL_IMAG(z)=0.0;
      
      GSL_REAL(z)=-(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm2,0,0,z);
      GSL_REAL(z)=-(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm2,1,1,z);
      GSL_REAL(z)=-(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm2,2,2,z);
      GSL_REAL(z)=-(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm2,3,3,z);
      GSL_REAL(z)=(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm2,4,4,z);
      GSL_REAL(z)=(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm2,5,5,z);
      
      GSL_REAL(z)=-lmom;
      gsl_matrix_complex_set(gm2,1,0,z);
      gsl_matrix_complex_set(gm2,3,2,z);
      gsl_matrix_complex_set(gm2,5,4,z);
      
      GSL_REAL(z)=0.0;
      GSL_IMAG(z)=u->del;
      gsl_matrix_complex_set(gm2,4,2,z);
      GSL_IMAG(z)=-u->del;
      gsl_matrix_complex_set(gm2,5,3,z);
      GSL_IMAG(z)=d->del;
      gsl_matrix_complex_set(gm2,4,0,z);
      GSL_IMAG(z)=-d->del;
      gsl_matrix_complex_set(gm2,5,1,z);

      gsl_eigen_herm(gm2,(gsl_vector *)(&eval2),w2);

      for(k=0;k<6;k++) egv[k]=eval2[k];
    
      gsl_matrix_complex_set_zero(gm2);
      
      GSL_IMAG(z)=0.0;
      
      GSL_REAL(z)=(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm2,0,0,z);
      GSL_REAL(z)=(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm2,1,1,z);
      GSL_REAL(z)=(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm2,2,2,z);
      GSL_REAL(z)=(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm2,3,3,z);
      GSL_REAL(z)=-(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm2,4,4,z);
      GSL_REAL(z)=-(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm2,5,5,z);
      
      GSL_REAL(z)=-lmom;
      gsl_matrix_complex_set(gm2,1,0,z);
      gsl_matrix_complex_set(gm2,3,2,z);
      gsl_matrix_complex_set(gm2,5,4,z);
      
      GSL_REAL(z)=0.0;
      GSL_IMAG(z)=u->del;
      gsl_matrix_complex_set(gm2,4,2,z);
      GSL_IMAG(z)=-u->del;
      gsl_matrix_complex_set(gm2,5,3,z);
      GSL_IMAG(z)=d->del;
      gsl_matrix_complex_set(gm2,4,0,z);
      GSL_IMAG(z)=-d->del;
      gsl_matrix_complex_set(gm2,5,1,z);

      gsl_eigen_herm(gm2,(gsl_vector *)(&eval2),w2);

      for(k=0;k<6;k++) egv[k+6]=eval2[k];

    } else {
      
      // Since the matrix is destroyed, we have to fill the entries 
      // every time. Some of the entries are commented out, since 
      // they represent the part of the matrix that isn't needed.
      gsl_matrix_complex *gm=(gsl_matrix_complex *)(&m);

      gsl_matrix_complex_set_zero(gm);

      GSL_IMAG(z)=0.0;

      GSL_REAL(z)=(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm,0,0,z);
      GSL_REAL(z)=-(u->mu+mur)-u->ms;
      gsl_matrix_complex_set(gm,10,10,z);
      GSL_REAL(z)=(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm,1,1,z);
      GSL_REAL(z)=-(u->mu+mur)+u->ms;
      gsl_matrix_complex_set(gm,11,11,z);


      GSL_REAL(z)=(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm,6,6,z);
      GSL_REAL(z)=-(d->mu+mug)-d->ms;
      gsl_matrix_complex_set(gm,4,4,z);
      GSL_REAL(z)=-(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm,2,2,z);
      GSL_REAL(z)=(d->mu+mug)+d->ms;
      gsl_matrix_complex_set(gm,7,7,z);

      GSL_REAL(z)=-(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm,5,5,z);
      GSL_REAL(z)=(s->mu+mub)-s->ms;
      gsl_matrix_complex_set(gm,8,8,z);
      GSL_REAL(z)=(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm,9,9,z);
      GSL_REAL(z)=-(s->mu+mub)+s->ms;
      gsl_matrix_complex_set(gm,3,3,z);

      GSL_REAL(z)=-lmom;
      // gsl_matrix_complex_set(gm,0,1,z);
      gsl_matrix_complex_set(gm,1,0,z);
      // gsl_matrix_complex_set(gm,2,4,z);
      // gsl_matrix_complex_set(gm,3,5,z);
      gsl_matrix_complex_set(gm,4,2,z);
      gsl_matrix_complex_set(gm,5,3,z);
      // gsl_matrix_complex_set(gm,6,7,z);
      gsl_matrix_complex_set(gm,7,6,z);
      // gsl_matrix_complex_set(gm,8,9,z);
      gsl_matrix_complex_set(gm,9,8,z);
      // gsl_matrix_complex_set(gm,10,11,z);
      gsl_matrix_complex_set(gm,11,10,z);

      GSL_REAL(z)=0.0;

      GSL_IMAG(z)=u->del;
      // gsl_matrix_complex_set(gm,2,8,z);
      // gsl_matrix_complex_set(gm,3,6,z);
      gsl_matrix_complex_set(gm,7,5,z);
      gsl_matrix_complex_set(gm,9,4,z);
      GSL_IMAG(z)=-u->del;
      // gsl_matrix_complex_set(gm,4,9,z);
      // gsl_matrix_complex_set(gm,5,7,z);
      gsl_matrix_complex_set(gm,6,3,z);
      gsl_matrix_complex_set(gm,8,2,z);

      GSL_IMAG(z)=d->del;
      // gsl_matrix_complex_set(gm,1,5,z);
      gsl_matrix_complex_set(gm,3,0,z);
      // gsl_matrix_complex_set(gm,9,10,z);
      gsl_matrix_complex_set(gm,11,8,z);
      GSL_IMAG(z)=-d->del;
      // gsl_matrix_complex_set(gm,0,3,z);
      gsl_matrix_complex_set(gm,5,1,z);
      // gsl_matrix_complex_set(gm,8,11,z);
      gsl_matrix_complex_set(gm,10,9,z);

      GSL_IMAG(z)=s->del;
      // gsl_matrix_complex_set(gm,1,4,z);
      gsl_matrix_complex_set(gm,2,0,z);
      // gsl_matrix_complex_set(gm,7,10,z);
      gsl_matrix_complex_set(gm,11,6,z);
      GSL_IMAG(z)=-s->del;
      // gsl_matrix_complex_set(gm,0,2,z);
      gsl_matrix_complex_set(gm,4,1,z);
      // gsl_matrix_complex_set(gm,6,11,z);
      gsl_matrix_complex_set(gm,10,7,z);

      gsl_eigen_herm(gm,(gsl_vector *)(&eval),w);
      
      for(k=0;k<12;k++) {
	egv[k]=eval[k];
      }
    }

    ovector_cx_subvector q1(qev,12,8);
    ovector_cx_subvector q2(qev,20,8);
    ovector_cx_subvector q3(qev,28,8);
    gapped_eigenvalues(u,d,lmom,u->mu+mug,d->mu+mur,s->del,q1);
    gapped_eigenvalues(d,s,lmom,d->mu+mub,s->mu+mug,u->del,q2);
    gapped_eigenvalues(u,s,lmom,u->mu+mub,s->mu+mur,d->del,q3);

    for(k=12;k<36;k++) {
      egv[k]=qev.real(k);
    }
  }

  return 0;
}

int cfl_njl_eos::solve_fun_p(size_t nv, const ovector_view &x, 
			     ovector_view &y, void *&vp) {
  size_t ix=0;
  double lmu3=0.0, lmu8=0.0;
  if (!fixed_mass) {
    up->qq=x[ix]; ix++;
    down->qq=x[ix]; ix++;
    strange->qq=x[ix]; ix++;
    if (up->qq>0.0 || down->qq>0.0 || strange->qq>0.0) return 1;
  }
  if (color_neut) {
    lmu3=x[ix]; ix++;
    lmu8=x[ix]; ix++;
  }
  if (ugap_solve) { 
    up->del=x[ix]; 
    ix++;
    if (up->del<0.0) return 2;
  }
  if (dgap_solve) { 
    down->del=x[ix]; 
    ix++;
    if (down->del<0.0) return 3;
  }
  if (sgap_solve) { 
    strange->del=x[ix]; 
    ix++;
    if (strange->del<0.0) return 4;
  }

  double qq1, qq2, qq3, gap1, gap2, gap3, n3, n8;
  calc_eq_temp_p(*up,*down,*strange,qq1,qq2,qq3,
		 gap1,gap2,gap3,lmu3,lmu8,n3,n8,*eos_thermo,temper);
  
  ix=0;
  if (!fixed_mass) {
    y[ix]=qq1; ix++;
    y[ix]=qq2; ix++;
    y[ix]=qq3; ix++;
  }
  if (color_neut) {
    y[ix]=n3; ix++;
    y[ix]=n8; ix++;
  }
  if (ugap_solve) { y[ix]=gap1; ix++; }
  if (dgap_solve) { y[ix]=gap2; ix++; }
  if (sgap_solve) { y[ix]=gap3; ix++; }

  return 0;
}

int cfl_njl_eos::min_fun2(size_t nv, const ovector_view &x, double &y,
			  void *&vp) {

  smu3=x[0];
  smu8=x[1];
  
  size_t nv2=0;
  if (!fixed_mass) nv2+=3;
  if (ugap_solve) nv2++;
  if (dgap_solve) nv2++;
  if (sgap_solve) nv2++;

  ovector x2(nv2);
  size_t ix=0;
  if (!fixed_mass) {
    x2[ix]=up->qq; ix++;
    x2[ix]=down->qq; ix++;
    x2[ix]=strange->qq; ix++;
  }
  if (ugap_solve) {
    x2[ix]=up->del; ix++;
  }
  if (dgap_solve) {
    x2[ix]=down->del; ix++;
  }
  if (sgap_solve) {
    x2[ix]=strange->del; ix++;
  }

  multi_funct_mfptr<cfl_njl_eos,void *> mfm(this,&cfl_njl_eos::min_fun);
  min1->mmin(nv2,x2,y,vp,mfm);
  
  // We want to maximize with respect to the chemical potentials
  y*=-1.0;

  return 0;
}

int cfl_njl_eos::min_fun(size_t nv, const ovector_view &x, double &y,
			 void *&vp) {
  
  size_t ix=0;
  double lmu3=smu3, lmu8=smu8;
  if (!fixed_mass) {
    up->qq=x[ix]; ix++;
    down->qq=x[ix]; ix++;
    strange->qq=x[ix]; ix++;
  }
  if (ugap_solve) { 
    up->del=x[ix]; 
    ix++;
  }
  if (dgap_solve) { 
    down->del=x[ix]; 
    ix++;
  }
  if (sgap_solve) { 
    strange->del=x[ix]; 
    ix++;
  }

  double qq1, qq2, qq3, gap1, gap2, gap3, n3, n8;
  y=thd_potential(*up,*down,*strange,lmu3,lmu8,temper);

  return 0;
}
      
template<> int io_tlate<cfl_njl_eos>::input
(cinput *co, in_file_format *ins, cfl_njl_eos *nj) {
  std::string stemp;
  /*
    co->object_in("nambujl_eos",ins,nj,stemp);
    ins->double_in(nj->GD,"GD");
    ins->double_in(nj->gap_limit,"gap_limit");
    ins->bool_in(nj->zerot,"zerot");
    string tmp;
    bool sw;
    ins->bool_in(sw,"quartic");
    if (sw) pointer_in(co,ins,((void **)(&nj->quartic)),tmp);
    ins->bool_in(sw,"it");
    if (sw) pointer_in(co,ins,((void **)(&nj->it)),tmp);
  */
  return 0;
}

template<> int io_tlate<cfl_njl_eos>::output
(coutput *co, out_file_format *outs, cfl_njl_eos *nj) {
  /*
    std::string t;
    co->object_out("nambujl_eos",outs,nj,0,0,t);
    outs->double_out(nj->GD,"GD");
    outs->double_out(nj->gap_limit,"gap_limit");
    outs->bool_out(nj->zerot,"zerot");
    
    if (nj->quartic_set==true) {
    string tmp;
    outs->bool_out(true,"quartic");
    pointer_out(co,outs,nj->quartic,tmp);
    } else {
    outs->bool_out(false,"quartic");
    }
    if (nj->it!=&nj->def_it) {
    outs->bool_out(true,"it");
    pointer_out(co,outs,nj->it,nj->it->type());
    } else {
    outs->bool_out(false,"it");
    }
  */
  return 0;
}

template<> const char *io_tlate<cfl_njl_eos>::type()
{ return "cfl_njl_eos"; }

