/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_CFL_NJL_EOS_H
#define O2SCL_CFL_NJL_EOS_H

#include <iostream>
#include <complex>

#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_sort_vector.h>
#include <gsl/gsl_poly.h>

#include <o2scl/constants.h> 
#include <o2scl/err_hnd.h>
#include <o2scl/ovector_cx_tlate.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/omatrix_cx_tlate.h>
#include <o2scl/multi_funct.h>
#include <o2scl/mm_funct.h>

#include <o2scl/gsl_deriv.h>
#include <o2scl/cern_deriv.h>
#include <o2scl/gsl_mmin_conf.h>
#include <o2scl/poly.h>

#include <o2scl/part.h>
#include <o2scl/nambujl_eos.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Nambu Jona-Lasinio model with a schematic CFL 
      di-quark interaction at finite temperature
      
      The variable B0 must be set before use.
      
      The original Lagrangian is
      
      \f[
      {\cal L} =
      {\cal L}_{\mathrm{Dirac}} +
      {\cal L}_{\mathrm{4-fermion}} +
      {\cal L}_{\mathrm{6-fermion}} +
      {\cal L}_{CSC1} +
      {\cal L}_{CSC2} 
      \f]
      
      \f[
      {\cal L}_{\mathrm{Dirac}} = \bar{q}_{i \alpha}
      \left( i {\partial} \delta_{i j} \delta_{\alpha \beta} - 
      m_{i j} \delta_{\alpha \beta} - \mu_{i j,~\alpha \beta} \gamma^0
      \right) q_{j \beta}  
      \f]
      
      \f[
      {\cal L}_{\mathrm{4-fermion}} =
      G_S \sum_{a=0}^8 \left[ 
      \left( \bar{q} \lambda^a_f q \right)^2 +
      \left( \bar{q} i \gamma_5 \lambda^a_f q \right)^2 
      \right] 
      \f]
      
      \f[
      {\cal L}_{\mathrm{6-fermion}} =
      - G_{D} \left[ 
      {\mathrm det}_{i j} \, \bar{q}_{i \alpha} \left( 1 + i \gamma_5 \right) 
      q_{j \beta} +
      {\mathrm det}_{i j} \, \bar{q}_{i \alpha} \left( 1 - i \gamma_5 \right) 
      q_{j \beta} \right] \delta_{\alpha \beta} 
      \f]
      
      \f[
      {\cal L}_{CSC1} = 
      G_{DIQ} \sum_k \sum_{\gamma} \left[
      \left(\bar{q}_{i \alpha} \epsilon_{i j k} 
      \epsilon_{\alpha \beta \gamma} q^C_{j \beta}\right)
      \left(\bar{q}_{i^{\prime} \alpha^{\prime}}^C 
      \epsilon_{i^{\prime} j^{\prime} k} \epsilon_{\alpha^{\prime} 
      \beta^{\prime} \gamma} q_{j^{\prime} \beta^{\prime}}\right)\right]
      \f]
      
      \f[
      {\cal L}_{CSC2} = 
      G_{DIQ} \sum_k \sum_{\gamma} \left[
      \left(\bar{q}_{i \alpha} i \gamma_5 \epsilon_{i j k} 
      \epsilon_{\alpha \beta \gamma} q^C_{j \beta}\right)
      \left(\bar{q}_{i^{\prime} \alpha^{\prime}}^C i \gamma_5 
      \epsilon_{i^{\prime} j^{\prime} k} \epsilon_{\alpha^{\prime} 
      \beta^{\prime} \gamma} q_{j^{\prime} \beta^{\prime}}\right) \right] \,, 
      \f]
      
      where \f$ \mu \f$ is the quark number chemical potential.  
      couplings \f$ G_S \f$, \f$ G_D \f$, and \f$ G_{DIQ} \f$
      ultra-violet three-momentum cutoff, \f$ \Lambda \f$
      
      The thermodynamic potential is
      \f[
      \Omega(\mu_i,\left<\bar{q} q\right>_i,\left< q q\right>_i,T)
      =  \Omega_{\mathrm{vac}}+\Omega_{\mathrm{stat}} + \Omega_0
      \f]
      where \f$ i \f$ runs over all nine (three colors times three
      flavors) quarks. We assume that the condensates are independent
      of color and 
      that the quark chemical potentials
      are of the form
      \f$ \mu_Q=\mu_{\mathrm{Flavor(Q)}}+\mu_{\mathrm{Color(Q)}} \f$
      with
      \f[
      \mu_{\mathrm{red}} = \mu_3 + \mu_8/\sqrt{3}
      \quad
      \mu_{\mathrm{green}} = -\mu_3 + \mu_8/\sqrt{3}
      \quad
      \mu_{\mathrm{blue}} = -2 \mu_8 /\sqrt{3}
      \f]
      With these assumptions, the thermodynamic potential as given
      by the function thd_potential(), is a function of 12 variables
      \f[
      \Omega(\mu_u, \mu_d, \mu_s, \mu_3, \mu_8, \left<\bar{u} u\right>,
      \left<\bar{d} d\right>, \left<\bar{s} s\right>,
      \left< u d\right>, \left< u s\right>, \left< d s\right>,
      T)
      \f]
      
      The individual terms are
      \f[
      \Omega_{\mathrm{stat}} = - \frac{1}{2} \int \frac{d^3
      p}{\left(2 \pi\right)^3} \, \sum_{i=1}^{72} \left[ \frac{\lambda_i}{2} +
      T \ln{\left(1 + e^{-\lambda_i/T} \right)} \right] 
      \f]
      
      \f[
      \Omega_{\mathrm{vac}} =  
      - 2 G_S \sum_{i=u,d,s} \langle {\bar q_i} q_i \rangle^2  
      +4 G_D \left<{\bar u} u\right> \left<{\bar d} d \right> 
      \left<{\bar s} s\right>
      + \sum_k \sum_{\gamma} \frac{\left|\Delta^{k \gamma}\right|^2}{4 G_{DIQ}}
      \f]
      
      where \f$ \lambda_i \f$ are the eigenvalues of the (72 by 72) matrix
      (calculated by the function eigenvalues())
      \f[
      D = \left[
      \begin{array}{cc}
      - \gamma^0 \vec{\gamma} \cdot \vec{p} - M_{i} \gamma^0 + \mu_{i \alpha}  
      & \Delta i \gamma^0 \gamma_5 C \\
      i \Delta \gamma^0 C \gamma_5 
      & - \gamma^0 \vec{\gamma}^T \cdot \vec{p} + M_{i} \gamma^0 
      - \mu_{i \alpha}\end{array}
      \right] 
      \f]
      and \f$ C \f$ is the charge conjugation matrix (in the Dirac
      representation).
      
      The values of the various condensates are usually determined by
      the condition
      \f[
      \frac{\partial \Omega}{\left<\bar{q} q\right>_i} = 0
      \quad
      \frac{\partial \Omega}{\left<q q\right>_i} = 0
      \f]
      
      Note that setting fixed_mass to \c true and setting all of the
      gaps to zero when \c gap_limit is less than zero will reproduce
      an analog of the bag model with a momentum cutoff.
      
      The variable nambujl_eos::fromqq is automatically set to true in
      the constructor, as computations with \c fromqq=false are not
      implemented.

      \future This class internally mixes ovector, omatrix, gsl_vector
      and gsl_matrix objects in a confusing and non-optimal way. Fix this. 

      \future Allow user to change derivative object? This isn't possible
      right now because the stepsize parameter of the derivative
      object is used.
      
  */
  class cfl_njl_eos : public nambujl_eos {
  public:
    
    cfl_njl_eos();
    
    virtual ~cfl_njl_eos();
    
    /** 
	\brief Set the parameters and the bag constant 'B0'
	
	This function allows the user to specify the momentum cutoff,
	\c lambda, the four-fermion coupling \c fourferm, the
	six-fermion coupling from the 't Hooft interaction \c sixferm,
	and the color-superconducting coupling, \c fourgap. If 0.0 is
	given for any of the values, then the default is used (\f$
	\Lambda=602.3/(\hbar c), G=1.835/\Lambda^2, K=12.36/\Lambda^5
	\f$).
	
	If the four-fermion coupling that produces a gap is not
	specified, it is automatically set to 3/4 G, which is the
	value obtained from the Fierz transformation.
	
	The value of the shift in the bag constant nambujl_eos::B0 is
	automatically calculated to ensure that the vacuum has zero
	energy density and zero pressure. The functions set_quarks()
	and set_thermo() must be used before hand to specify the \ref
	quark and \ref thermo objects.
	
    */
    virtual int set_parameters(double lambda=0.0, double fourferm=0.0, 
			       double sixferm=0.0, double fourgap=0.0);
    
    /** \brief Calculate equation of state as a function of chemical 
	potentials

	If any of the gaps given in quark::del are greater than
	gap_limit, then the respective gap equations are 
	solved. If fixed_mass is false, then the quark condensates are
	solved for. If color_neut is true, then color neutrality is
	enforced. 
    */
    virtual int calc_temp_p(quark &u, quark &d, quark &s, 
			    const double temper, thermo &th);
    
    /** 
	\brief Calculate the EOS
	
	Calculate the EOS from the quark condensates in \c u.qq, \c
	d.qq and \c s.qq. Return the mass gap equations in \c qq1, \c
	qq2, \c qq3, and the normal gap equations in \c gap1, \c gap2,
	and \c gap3.
	
	Using \c fromqq=false as in nambujl_eos and nambujl_eos does not
	work here and will return an error. Also, the quarks must be
	set through quark_eos::quark_set() before use.
	
	If all of the gaps are less than gap_limit, then the
	nambujl_eos::calc_temp_p() is used, and \c gap1, \c gap2, and
	\c gap3 are set to equal \c u.del, \c d.del, and \c s.del,
	respectively.
	
    */
    virtual int calc_eq_temp_p(quark &u, quark &d, quark &s, double &qq1, 
			       double &qq2, double &qq3, double &gap1, 
			       double &gap2, double &gap3, double mu3, 
			       double mu8,
			       double &n3, double &n8, thermo &qb, 
			       const double temper);
    
    /** \brief Calculate the energy eigenvalues as a function of the
	momentum
	
	Given the momentum \c mom, and the chemical potentials
	associated with the third and eighth gluons (\c mu3 and \c
	mu8), the energy eigenvalues are computed in egv[0]
	... egv[35]. No space is allocated for the array by the
	function.
    */
    virtual int eigenvalues(double mom, ovector_view &egv, double mu3, 
			    double mu8);
    
    /** \brief Direct calculation of the thermodynamic potential
     */
    virtual double thd_potential(quark &u, quark &d, quark &s,
				 double mu3, double mu8, 
				 const double ltemper);

    /// \name Numerical methods
    //@{
    /// Set the minimizers for the thermodynamic potential
    int set_minimizers(multi_min<void *, multi_funct<void *> > &mi1, 
		       multi_min<void *, multi_funct<void *> > &mi2) {
      min1=&mi1;
      min2=&mi2;
      return 0;
    }
    
    /// Set the routine for solving quartics
    int set_quartic(quartic_real_coeff &q) { 
      quartic=&q; 
      quartic_set=true; 
      return 0;
    }

    /// The first default minimizer
    gsl_mmin_conf<void *, multi_funct<void *> > def_min1;
    /// The second default minimizers
    gsl_mmin_conf<void *, multi_funct<void *> > def_min2;

    /** 
	\brief The default quartic routine

	Slightly better accuracy (with slower execution times) can be
	achieved using \ref gsl_poly_real_coeff which polishes the
	roots of the quartics. For example

	\code
	cfl_njl_eos cfl;
	gsl_poly_real_coeff gp;
	cfl.set_quartic(gp);
	\endcode
     */
    cern_quartic_real_coeff def_quartic;
    //@}
    
    /** 
	\brief Smallest allowable gap (default 0.0)
	
	If any of the gaps are below this value, then it is assumed
	that they are zero and the equation of state is simplified
	accordingly. If all of the gaps are less than gap_limit, then
	the results from nambujl_eos are used in
	calc_eq_temp_p(), calc_temp_p() and thd_potential().
    */
    double gap_limit;
    
    /** \brief If this is true, then finite temperature
	corrections are ignored (default false)
	
	This implements some simplifications in the 
	momentum integration that are not possible at finite temperature.
    */
    bool zerot;
    
    /** \brief Use a fixed quark mass and ignore the quark
	condensates
    */
    bool fixed_mass;

    /// If true, then ensure color neutrality
    bool color_neut;
    
    /** 
	\brief Diquark coupling constant (default 3 G/4)
	
	The default value is the one derived from a Fierz
	transformation. (\ref Buballa04)
    */
    double GD;
    
    /** 
	\brief Faster eigenvalues
	
	This assumes that the strange quark mass can be used
	as an effective chemical potential
	\f$ \mu_s^{\prime}=\mu_s-m_s^2/2/\mu_s\f$
	and calculates
	the eigenvalues to first order in the four variables
	\f[
	\alpha=\frac{\mu_d-\mu_u}{\mu_u}
	\f]
	\f[
	\beta=\frac{\mu_s-\mu_u}{\mu_u}
	\f]
	\f[
	\gamma=\frac{\Delta_{us}-\Delta_{ud}}{\Delta_{ud}}
	\f]
	\f[
	\delta=\frac{\Delta_{ds}-\Delta_{ud}}{\Delta_{ud}}
	\f]
	
    */
    bool faster;

    /// The quality of the most recently obtained solution
    double solver_qual;

    /// If true, use the solver, otherwise, use the minimizers
    bool use_solver;
    
    /// Return string denoting type ("cfl_njl_eos")
    virtual const char *type() { return "cfl_njl_eos"; };
    
    /** \brief The function used to take derivatives of the thermodynamic
	potential (used by calc_temp_p() )

	To use this function just to compute the thermodynamic potential,
	make sure \ref var is 0 (or \c NULL) and use 
	\code
	tpot(0.0,0);
	\endcode
	since the arguments are both ignored.
    */
    virtual double tpot(double var, void *&pa);
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    friend class io_tlate<cfl_njl_eos>;

    /// Function specifying gap equations to solve
    int solve_fun_p(size_t nv, const ovector_view &x, ovector_view &y,
		    void *&vp);
    
    /// Outer minimization function for color neutrality
    int min_fun2(size_t nv, const ovector_view &x, double &y,
		 void *&vp);
    
    /// Inner minimization function for mass and gap equations
    int min_fun(size_t nv, const ovector_view &x, double &y,
		void *&vp);
    
    /** \brief A function that takes the derivative of the thermodynamic
	potential with respect to the variable whose address is 'dvar'.
    */
    double derivfn(double *dvar);
    
    /// The integrand of the thermodynamic potential
    double tpot_integrand(double p, void *&pa);
    
    /// The integrand of the entropy 
    double entropy_integrand(double p, void *&pa);
    
    /** 
	\brief Treat the simply gapped quarks in all cases gracefully

	This function uses the quarks \c q1 and \c q2 to construct the
	eigenvalues of the inverse propagator, properly handling the
	either zero or finite quark mass and either zero or finite
	quark gaps. In the case of finite quark mass and finite quark
	gaps, the quartic solver is used.
	
	The chemical potentials are separated so we can add the 
	color chemical potentials to the quark chemical potentials
	if necessary.

	This function is used by eigenvalues(). It does not work for
	the "ur-dg-sb" set of quarks which are paired in a non-trivial
	way.

	\todo Only the "ms" part of the quarks is referenced, so
	we should rewrite to use only double's as function arguments,
	and avoid passing pointers to quark objects.
    */
    int gapped_eigenvalues(quark *q1, quark *q2, double mom,
			   double mu1, double mu2, double delta,
			   ovector_cx_view &ev);
    
    /// Temperature
    double temper;
    
    /// 3rd gluon chemical potential
    double smu3;
    
    /// 8th gluon chemical potential
    double smu8;

    /// \name Numerical methods
    //@{
    /// The routine to solve quartics
    quartic_real_coeff *quartic;
    
    /** 
	\brief true if the quartic routine has been set

	This is used for the I/O.
    */
    bool quartic_set;

    /** \brief The derivative routine
     */
    gsl_deriv<void *,funct<void *> > de;

    /// Pointer to inner minimizer
    multi_min<void *, multi_funct<void *> > *min1;

    /// Pointer to outer minimizer
    multi_min<void *, multi_funct<void *> > *min2;
    //@}
    
    /// Pointer to the variable to be differentiated by
    double *var;  

    /// \name For computing eigenvalues
    //@{
    /// Inverse propagator matrix
    omatrix_cx m;

    /// Inverse propagator matrix
    omatrix_cx m2;

    /// Storage for eigenvalues
    ovector eval;

    /// Storage for eigenvalues
    ovector eval2;

    /// Workspace for eigenvalue computation
    gsl_eigen_herm_workspace *w;
    
    /// Workspace for eigenvalue computation
    gsl_eigen_herm_workspace *w2;
    //@}

    /// Switches for solving the gap equations
    //@{
    bool ugap_solve;
    bool dgap_solve;
    bool sgap_solve;
    //@}
    
  private:
    
    cfl_njl_eos(const cfl_njl_eos &);
    cfl_njl_eos& operator=(const cfl_njl_eos&);
    
#endif
    
  };
  
  template<> int io_tlate<cfl_njl_eos>::input
    (cinput *co, in_file_format *ins, cfl_njl_eos *nj);
  template<> int io_tlate<cfl_njl_eos>::output
    (coutput *co, out_file_format *outs, cfl_njl_eos *nj);
  template<> const char *io_tlate<cfl_njl_eos>::type();
  
  typedef io_tlate<cfl_njl_eos> cfl_njl_eos_io_type;
   
#ifndef DOXYGENP
}
#endif

#endif
