/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/nambujl_eos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

nambujl_eos::nambujl_eos() {

  def_up.non_interacting=true;
  def_down.non_interacting=true;
  def_strange.non_interacting=true;
  def_up.init(up_default_mass,6.0);
  def_down.init(down_default_mass,6.0);
  def_strange.init(strange_default_mass,6.0);
  
  up=&def_up;
  down=&def_down;
  strange=&def_strange;

  B0=0.0;

  up_default_mass=5.5/197.3269602;
  down_default_mass=5.5/197.3269602;
  strange_default_mass=140.7/197.3269602;

  L=602.3/hc_mev_fm;
  G=1.835/L/L;
  K=12.36/pow(L,5.0);
  
  solver=&def_solver;
  it=&def_it;

  // We don't call set_parameters() here, because we can't call a
  // virtual member function from the constructor

  limit=20.0;

  fromqq=true;

}

int nambujl_eos::set_quarks(quark &u, quark &d, quark &s) {
  up=&u;
  down=&d;
  strange=&s;

  return 0;
}

int nambujl_eos::set_parameters(double lambda, double fourferm, 
				double sixferm) {
  
  if (lambda!=0.0) L=lambda;
  else L=602.3/hc_mev_fm;
  if (fourferm!=0.0) G=fourferm;
  else G=1.835/L/L;
  if (sixferm!=0.0) K=sixferm;
  else K=12.36/pow(L,5.0);

  ovector bx(3);
  bool fromqqold=fromqq;
  fromqq=false;

  // Solve zero-density gap-equations:
  bx[0]=1.0;
  bx[1]=1.0;
  bx[2]=2.0;

  mm_funct_mfptr<nambujl_eos,void *> fmf(this,&nambujl_eos::B0fun);
  void *vpx=0;
  solver->msolve(3,bx,vpx,fmf);
  
  // Make the appropriate correction
  B0+=eos_thermo->pr;

  // Return fromqq to its original value
  fromqq=fromqqold;
  return 0;
}

int nambujl_eos::calc_p(quark &u, quark &d, quark &s, thermo &th) {
  ovector x(3);
  void *vp=0;
  int ret;

  up=&u;
  down=&d;
  strange=&s;
  eos_thermo=&th;
  
  if (fromqq==true) {

    x[0]=u.qq;
    x[1]=d.qq;
    x[2]=s.qq;
    mm_funct_mfptr<nambujl_eos,void *> fmf(this,&nambujl_eos::gapfunqq);
    ret=solver->msolve(3,x,vp,fmf);
    
  } else {

    x[0]=u.ms;
    x[1]=d.ms;
    x[2]=s.ms;
    mm_funct_mfptr<nambujl_eos,void *> fmf(this,&nambujl_eos::gapfunms);
    ret=solver->msolve(3,x,vp,fmf);

  }

  double gap1, gap2, gap3;
  
  return ret;
}

int nambujl_eos::calc_temp_p(quark &u, quark &d, quark &s, 
			     const double T, thermo &th) {
  ovector x(3);
  void *vp=0;
  int ret;

  up=&u;
  down=&d;
  strange=&s;
  eos_thermo=&th;
  cp_temp=T;
  
  if (fromqq==true) {

    x[0]=u.qq;
    x[1]=d.qq;
    x[2]=s.qq;
    mm_funct_mfptr<nambujl_eos,void *> fmf(this,&nambujl_eos::gapfunqqT);
    ret=solver->msolve(3,x,vp,fmf);
    
  } else {

    x[0]=u.ms;
    x[1]=d.ms;
    x[2]=s.ms;
    mm_funct_mfptr<nambujl_eos,void *> fmf(this,&nambujl_eos::gapfunmsT);
    ret=solver->msolve(3,x,vp,fmf);

  }

  double gap1, gap2, gap3;
  
  return ret;
}

int nambujl_eos::calc_eq_p(quark &u, quark &d, quark &s, double &gap1,
			   double &gap2, double &gap3, thermo &th) {
  if (fromqq==true) {
    
    //--------------------------------------------
    // Calculate everything from the quark 
    // condensates and the chemical potentials

    u.ms=u.m-4.0*G*u.qq+2.0*K*d.qq*s.qq;
    d.ms=d.m-4.0*G*d.qq+2.0*K*u.qq*s.qq;
    s.ms=s.m-4.0*G*s.qq+2.0*K*d.qq*u.qq;
    
    if (u.mu>u.ms) u.n=1.0/pi2*pow(u.mu*u.mu-u.ms*u.ms,1.5);
    else u.n=0.0;
    if (d.mu>d.ms) d.n=1.0/pi2*pow(d.mu*d.mu-d.ms*d.ms,1.5);
    else d.n=0.0;
    if (s.mu>s.ms) s.n=1.0/pi2*pow(s.mu*s.mu-s.ms*s.ms,1.5);
    else s.n=0.0;

    u.kffromden();
    d.kffromden();
    s.kffromden();
  
    gap1=3.0/pi2*(-u.ms/2.0*L*sqrt(u.ms*u.ms+L*L)+pow(u.ms,3.0)/2.0*
		  log(L+sqrt(L*L+u.ms*u.ms)));
    gap1+=3.0/pi2*(u.ms/2.0*u.kf*sqrt(u.ms*u.ms+u.kf*u.kf)-
		   pow(u.ms,3.0)/2.0*
		   log(u.kf+sqrt(u.kf*u.kf+u.ms*u.ms)));
    gap1-=u.qq;
    gap2=3.0/pi2*(-d.ms/2.0*L*sqrt(d.ms*d.ms+L*L)+pow(d.ms,3.0)/2.0*
		  log(L+sqrt(L*L+d.ms*d.ms)));
    gap2+=3.0/pi2*(d.ms/2.0*d.kf*sqrt(d.ms*d.ms+d.kf*d.kf)-
		   pow(d.ms,3.0)/2.0*
		   log(d.kf+sqrt(d.kf*d.kf+d.ms*d.ms)));
    gap2-=d.qq;
    gap3=3.0/pi2*(-s.ms/2.0*L*sqrt(s.ms*s.ms+L*L)+pow(s.ms,3.0)/2.0*
		  log(L+sqrt(L*L+s.ms*s.ms)));
    gap3+=3.0/pi2*(s.ms/2.0*s.kf*sqrt(s.ms*s.ms+s.kf*s.kf)-
		   pow(s.ms,3.0)/2.0*
		   log(s.kf+sqrt(s.kf*s.kf+s.ms*s.ms)));
    gap3-=s.qq;

  } else {

    //--------------------------------------------
    // Calculate everything from the dynamical 
    // masses and the chemical potentials

    if (u.mu>u.ms) u.kf=sqrt(u.mu*u.mu-u.ms*u.ms);
    else u.kf=0.0;
    if (d.mu>d.ms) d.kf=sqrt(d.mu*d.mu-d.ms*d.ms);
    else d.kf=0.0;
    if (s.mu>s.ms) s.kf=sqrt(s.mu*s.mu-s.ms*s.ms);
    else s.kf=0.0;

    u.qq=3.0/pi2*(-u.ms/2.0*L*sqrt(u.ms*u.ms+L*L)+pow(u.ms,3.0)/2.0*
		  log(L+sqrt(L*L+u.ms*u.ms)));
    u.qq+=3.0/pi2*(u.ms/2.0*u.kf*sqrt(u.ms*u.ms+u.kf*u.kf)-
		   pow(u.ms,3.0)/2.0*
		   log(u.kf+sqrt(u.kf*u.kf+u.ms*u.ms)));
    d.qq=3.0/pi2*(-d.ms/2.0*L*sqrt(d.ms*d.ms+L*L)+pow(d.ms,3.0)/2.0*
		  log(L+sqrt(L*L+d.ms*d.ms)));
    d.qq+=3.0/pi2*(d.ms/2.0*d.kf*sqrt(d.ms*d.ms+d.kf*d.kf)-
		   pow(d.ms,3.0)/2.0*
		   log(d.kf+sqrt(d.kf*d.kf+d.ms*d.ms)));
    s.qq=3.0/pi2*(-s.ms/2.0*L*sqrt(s.ms*s.ms+L*L)+pow(s.ms,3.0)/2.0*
		  log(L+sqrt(L*L+s.ms*s.ms)));
    s.qq+=3.0/pi2*(s.ms/2.0*s.kf*sqrt(s.ms*s.ms+s.kf*s.kf)-
		   pow(s.ms,3.0)/2.0*
		   log(s.kf+sqrt(s.kf*s.kf+s.ms*s.ms)));

    if (u.mu>u.ms) u.n=1.0/pi2*pow(u.mu*u.mu-u.ms*u.ms,1.5);
    else u.n=0.0;
    if (d.mu>d.ms) d.n=1.0/pi2*pow(d.mu*d.mu-d.ms*d.ms,1.5);
    else d.n=0.0;
    if (s.mu>s.ms) s.n=1.0/pi2*pow(s.mu*s.mu-s.ms*s.ms,1.5);
    else s.n=0.0;

    gap1=-u.ms+u.m-4.0*G*u.qq+2.0*K*d.qq*s.qq;
    gap2=-d.ms+d.m-4.0*G*d.qq+2.0*K*u.qq*s.qq;
    gap3=-s.ms+s.m-4.0*G*s.qq+2.0*K*d.qq*u.qq;

  }    
  
  u.eden();
  d.eden();
  s.eden();
  
  njbag(u);
  njbag(d);
  njbag(s);
  
  u.ed-=u.B;
  d.ed-=d.B;
  s.ed-=s.B;
  
  u.pr=-u.ed+u.n*u.mu;
  d.pr=-d.ed+d.n*d.mu;
  s.pr=-s.ed+s.n*s.mu;
  
  th.ed=u.ed+d.ed+s.ed+
    2.0*G*(u.qq*u.qq+d.qq*d.qq+s.qq*s.qq)+B0-4.0*K*u.qq*d.qq*s.qq;
  th.pr=-th.ed+u.n*u.mu+d.n*d.mu+s.n*s.mu;
  th.en=0.0;

  return 0;
}

int nambujl_eos::calc_eq_e(quark &u, quark &d, quark &s, double &gap1,
			   double &gap2, double &gap3, thermo &th) {

  if (fromqq==true) {

    //--------------------------------------------
    // Calculate everything from the quark 
    // condensates and the chemical potentials

    u.ms=u.m-4.0*G*u.qq+2.0*K*d.qq*s.qq;
    d.ms=d.m-4.0*G*d.qq+2.0*K*u.qq*s.qq;
    s.ms=s.m-4.0*G*s.qq+2.0*K*d.qq*u.qq;

    u.kffromden();
    d.kffromden();
    s.kffromden();
  
    u.mu=sqrt(u.kf*u.kf+u.ms*u.ms);
    d.mu=sqrt(d.kf*d.kf+d.ms*d.ms);
    s.mu=sqrt(s.kf*s.kf+s.ms*s.ms);
    
    gap1=3.0/pi2*(-u.ms/2.0*L*sqrt(u.ms*u.ms+L*L)+pow(u.ms,3.0)/2.0*
		  log(L+sqrt(L*L+u.ms*u.ms)));
    gap1+=3.0/pi2*(u.ms/2.0*u.kf*sqrt(u.ms*u.ms+u.kf*u.kf)-
		   pow(u.ms,3.0)/2.0*
		   log(u.kf+sqrt(u.kf*u.kf+u.ms*u.ms)));
    gap1-=u.qq;
    gap2=3.0/pi2*(-d.ms/2.0*L*sqrt(d.ms*d.ms+L*L)+pow(d.ms,3.0)/2.0*
		  log(L+sqrt(L*L+d.ms*d.ms)));
    gap2+=3.0/pi2*(d.ms/2.0*d.kf*sqrt(d.ms*d.ms+d.kf*d.kf)-
		   pow(d.ms,3.0)/2.0*
		   log(d.kf+sqrt(d.kf*d.kf+d.ms*d.ms)));
    gap2-=d.qq;
    gap3=3.0/pi2*(-s.ms/2.0*L*sqrt(s.ms*s.ms+L*L)+pow(s.ms,3.0)/2.0*
		  log(L+sqrt(L*L+s.ms*s.ms)));
    gap3+=3.0/pi2*(s.ms/2.0*s.kf*sqrt(s.ms*s.ms+s.kf*s.kf)-
		   pow(s.ms,3.0)/2.0*
		   log(s.kf+sqrt(s.kf*s.kf+s.ms*s.ms)));
    gap3-=s.qq;

  } else {

    //--------------------------------------------
    // Calculate everything from the dynamical 
    // masses and the chemical potentials

    u.kffromden();
    d.kffromden();
    s.kffromden();
  
    u.mu=sqrt(u.kf*u.kf+u.ms*u.ms);
    d.mu=sqrt(d.kf*d.kf+d.ms*d.ms);
    s.mu=sqrt(s.kf*s.kf+s.ms*s.ms);

    u.qq=3.0/pi2*(-u.ms/2.0*L*sqrt(u.ms*u.ms+L*L)+pow(u.ms,3.0)/2.0*
		  log(L+sqrt(L*L+u.ms*u.ms)));
    u.qq+=3.0/pi2*(u.ms/2.0*u.kf*sqrt(u.ms*u.ms+u.kf*u.kf)-
		   pow(u.ms,3.0)/2.0*
		   log(u.kf+sqrt(u.kf*u.kf+u.ms*u.ms)));
    d.qq=3.0/pi2*(-d.ms/2.0*L*sqrt(d.ms*d.ms+L*L)+pow(d.ms,3.0)/2.0*
		  log(L+sqrt(L*L+d.ms*d.ms)));
    d.qq+=3.0/pi2*(d.ms/2.0*d.kf*sqrt(d.ms*d.ms+d.kf*d.kf)-
		   pow(d.ms,3.0)/2.0*
		   log(d.kf+sqrt(d.kf*d.kf+d.ms*d.ms)));
    s.qq=3.0/pi2*(-s.ms/2.0*L*sqrt(s.ms*s.ms+L*L)+pow(s.ms,3.0)/2.0*
		  log(L+sqrt(L*L+s.ms*s.ms)));
    s.qq+=3.0/pi2*(s.ms/2.0*s.kf*sqrt(s.ms*s.ms+s.kf*s.kf)-
		   pow(s.ms,3.0)/2.0*
		   log(s.kf+sqrt(s.kf*s.kf+s.ms*s.ms)));

    gap1=-u.ms+u.m-4.0*G*u.qq+2.0*K*d.qq*s.qq;
    gap2=-d.ms+d.m-4.0*G*d.qq+2.0*K*u.qq*s.qq;
    gap3=-s.ms+s.m-4.0*G*s.qq+2.0*K*d.qq*u.qq;

  }    

  u.eden();
  d.eden();
  s.eden();

  njbag(u);
  njbag(d);
  njbag(s);

  u.ed-=u.B;
  d.ed-=d.B;
  s.ed-=s.B;
  
  u.pr=-u.ed+u.n*u.mu;
  d.pr=-d.ed+d.n*d.mu;
  s.pr=-s.ed+s.n*s.mu;
  
  th.ed=u.ed+d.ed+s.ed+
    2.0*G*(u.qq*u.qq+d.qq*d.qq+s.qq*s.qq)+B0-4.0*K*u.qq*d.qq*s.qq;
  th.pr=-th.ed+u.n*u.mu+d.n*d.mu+s.n*s.mu;
  th.en=0.0;

  return 0;
}

void nambujl_eos::njbag(quark &pp) {

  if (fabs(pp.ms)<1.0e-6) {
    pp.B=0.0;
  } else {
    if (fabs(pp.m)<1.0e-6) {
      pp.B=3.0/pi2*(1.0/4.0*L*pow(pp.ms*pp.ms+L*L,1.5)
		    -1.0/8.0*pp.ms*pp.ms*L*sqrt(pp.ms*pp.ms+L*L)
		    -1.0/8.0*pow(pp.ms,4.0)*log(L+sqrt(pp.ms*pp.ms+L*L))
		    +1.0/16.0*pow(pp.ms,4.0)*log(pp.ms*pp.ms));
    } else {
      pp.B=3.0/pi2*(1.0/4.0*L*pow(pp.ms*pp.ms+L*L,1.5)
		    -1.0/8.0*pp.ms*pp.ms*L*sqrt(pp.ms*pp.ms+L*L)
		    -1.0/8.0*pow(pp.ms,4.0)*log(L+sqrt(pp.ms*pp.ms+L*L))
		    +1.0/16.0*pow(pp.ms,4.0)*log(pp.ms*pp.ms));
    }
  }

  return;
}

int nambujl_eos::gapfunqq(size_t nv, const ovector_view &x, ovector_view &y, 
			  void *&pa) {

  double gap1,gap2,gap3;
  
  up->qq=x[0];
  down->qq=x[1];
  strange->qq=x[2];
  
  if (x[0]>0.0 || x[1]>0.0 || x[2]>0.0) return 1;

  calc_eq_p(*up,*down,*strange,gap1,gap2,gap3,*eos_thermo);
  
  y[0]=gap1;
  y[1]=gap2;
  y[2]=gap3;

  if (!finite(y[0]) || !finite(y[1]) || !finite(y[2])) return 2;
  
  return 0;
}

int nambujl_eos::gapfunms(size_t nv, const ovector_view &x, ovector_view &y, 
			  void *&pa) {

  double gap1,gap2,gap3;

  up->ms=x[0];
  down->ms=x[1];
  strange->ms=x[2];

  if (x[0]<0.0 || x[1]<0.0 || x[2]<0.0) return 1;

  calc_eq_p(*up,*down,*strange,gap1,gap2,gap3,*eos_thermo);

  y[0]=gap1;
  y[1]=gap2;
  y[2]=gap3;

  if (!finite(y[0]) || !finite(y[1]) || !finite(y[2])) return 2;

  return 0;
}
int nambujl_eos::gapfunqqT(size_t nv, const ovector_view &x, ovector_view &y, 
			  void *&pa) {

  double gap1,gap2,gap3;
  
  up->qq=x[0];
  down->qq=x[1];
  strange->qq=x[2];
  
  if (x[0]>0.0 || x[1]>0.0 || x[2]>0.0) return 1;

  calc_eq_temp_p(*up,*down,*strange,gap1,gap2,gap3,*eos_thermo,cp_temp);
  
  y[0]=gap1;
  y[1]=gap2;
  y[2]=gap3;

  if (!finite(y[0]) || !finite(y[1]) || !finite(y[2])) return 2;
  
  return 0;
}

int nambujl_eos::gapfunmsT(size_t nv, const ovector_view &x, ovector_view &y, 
			  void *&pa) {

  double gap1,gap2,gap3;

  up->ms=x[0];
  down->ms=x[1];
  strange->ms=x[2];

  if (x[0]<0.0 || x[1]<0.0 || x[2]<0.0) return 1;

  calc_eq_temp_p(*up,*down,*strange,gap1,gap2,gap3,*eos_thermo,cp_temp);

  y[0]=gap1;
  y[1]=gap2;
  y[2]=gap3;

  if (!finite(y[0]) || !finite(y[1]) || !finite(y[2])) return 2;

  return 0;
}

int nambujl_eos::B0fun(size_t nv, const ovector_view &x, ovector_view &y, 
		       void *&pa) {
  double gap1,gap2,gap3;
  
  up->ms=x[0];
  down->ms=x[1];
  strange->ms=x[2];
  
  if (x[0]<0.0 || x[1]<0.0 || x[2]<0.0) return 1;

  up->mu=up->ms;
  down->mu=down->ms;
  strange->mu=strange->ms;
  
  calc_eq_p(*up,*down,*strange,gap1,gap2,gap3,*eos_thermo);
  
  y[0]=gap1;
  y[1]=gap2;
  y[2]=gap3;

  if (!finite(y[0]) || !finite(y[1]) || !finite(y[2])) return 2;

  return 0;
}

int nambujl_eos::calc_eq_temp_p(quark &u, quark &d, quark &s, 
				double &gap1, double &gap2, double &gap3, 
				thermo &qb, const double temper) {
  double ierr;
  int iret1, iret2, iret3, iret4;

  funct_mfptr_noerr<nambujl_eos,void *> fqq(this,&nambujl_eos::iqq);
  funct_mfptr_noerr<nambujl_eos,void *> fde(this,&nambujl_eos::ide);
  funct_mfptr_noerr<nambujl_eos,void *> fed(this,&nambujl_eos::ied);
  funct_mfptr_noerr<nambujl_eos,void *> fpr(this,&nambujl_eos::ipr);

  njtp pa;
  void *vpx=(void *)&pa;

  if (temper<=0.0) {
    calc_eq_p(u,d,s,gap1,gap2,gap3,qb);
    return 0;
  }

  if (u.ms<=0.0) u.ms=1.e-9;
  if (d.ms<=0.0) d.ms=1.e-9;
  if (s.ms<=0.0) s.ms=1.e-9;

  pa.temper=temper;
  pa.limit=limit;

  // -----------------------------------------------------------------
  // Some of these integrals (iqq, and ide) converge better when they
  // are non-zero, so we add 1 to the integrand and subtract off the
  // contribution after the fact. This may cause inaccuracies when the
  // densities are small.

  if (fromqq==true) {
    u.ms=u.m-4.0*G*u.qq+2.0*K*d.qq*s.qq;
    pa.ms=u.ms;
    pa.mu=u.mu;
    pa.m=u.m;
    iret1=0;
  } else {
    pa.ms=u.ms;
    pa.mu=u.mu;
    pa.m=u.m;
    iret1=it->integ_err(fqq,0.0,L,vpx,u.qq,ierr);
    u.qq-=L;
  }
  iret2=it->integ_err(fde,0.0,L,vpx,u.n,ierr);
  u.n-=L;
  u.kffromden();
  iret3=it->integ_err(fpr,0.0,L,vpx,u.pr,ierr);
  iret4=it->integ_err(fed,0.0,L,vpx,u.ed,ierr);
  if (iret1!=0 || iret2!=0 || iret3!=0 || iret4!=0) {
    add_err("Up quark failed in nambujl_eos::calc_eq_temp_p().",
	    gsl_efailed);
  }
  
  if (fromqq==true) {
    d.ms=d.m-4.0*G*d.qq+2.0*K*u.qq*s.qq;
    pa.ms=d.ms;
    pa.mu=d.mu;
    pa.m=d.m;
    iret1=0;
  } else {
    pa.ms=d.ms;
    pa.mu=d.mu;
    pa.m=d.m;
    iret1=it->integ_err(fqq,0.0,L,vpx,d.qq,ierr);
    d.qq-=L;
  }
  iret2=it->integ_err(fde,0.0,L,vpx,d.n,ierr);
  d.n-=L;
  d.kffromden();
  iret3=it->integ_err(fpr,0.0,L,vpx,d.pr,ierr);
  iret4=it->integ_err(fed,0.0,L,vpx,d.ed,ierr);
  if (iret1!=0 || iret2!=0 || iret3!=0 || iret4!=0) {
    add_err("Down quark failed in nambujl_eos::calc_eq_temp_p().",
	    gsl_efailed);
  }
  
  if (fromqq==true) {
    s.ms=s.m-4.0*G*s.qq+2.0*K*d.qq*u.qq;
    pa.ms=s.ms;
    pa.mu=s.mu;
    pa.m=s.m;
    iret1=0;
  } else {
    pa.ms=s.ms;
    pa.mu=s.mu;
    pa.m=s.m;
    iret1=it->integ_err(fqq,0.0,L,vpx,s.qq,ierr);
    s.qq-=L;
  }
  iret2=it->integ_err(fde,0.0,L,vpx,s.n,ierr);
  s.n-=L;
  s.kffromden();
  iret3=it->integ_err(fpr,0.0,L,vpx,s.pr,ierr);
  iret4=it->integ_err(fed,0.0,L,vpx,s.ed,ierr);
  if (iret1!=0 || iret2!=0 || iret3!=0 || iret4!=0) {
    add_err("Strange quark failed in nambujl_eos::calc_eq_temp_p().",
	    gsl_efailed);
  }

  qb.ed=u.ed+d.ed+s.ed+2.0*G*(u.qq*u.qq+d.qq*d.qq+s.qq*s.qq)+
    B0-4.0*K*u.qq*d.qq*s.qq;
  qb.pr=u.pr+d.pr+s.pr-2.0*G*(u.qq*u.qq+d.qq*d.qq+s.qq*s.qq)-
    B0+4.0*K*u.qq*d.qq*s.qq;
  qb.en=(qb.ed+qb.pr-u.n*u.mu-d.n*d.mu-s.n*s.mu)/
    temper;
  
  if (fromqq==true) {
    double qqt;

    pa.ms=u.ms;
    pa.mu=u.mu;
    pa.m=u.m;
    iret1=it->integ_err(fqq,0.0,L,vpx,qqt,ierr);
    gap1=-u.qq+qqt-L;
    
    pa.ms=d.ms;
    pa.mu=d.mu;
    pa.m=d.m;
    iret2=it->integ_err(fqq,0.0,L,vpx,qqt,ierr);
    gap2=-d.qq+qqt-L;
    
    pa.ms=s.ms;
    pa.mu=s.mu;
    pa.m=s.m;
    iret3=it->integ_err(fqq,0.0,L,vpx,qqt,ierr);
    gap3=-s.qq+qqt-L;

    if (iret1!=0 || iret2!=0 || iret3!=0) {
      add_err("Strange quark failed in nambujl_eos::calc_eq_temp_p().",
	      gsl_efailed);
    }
  } else {
    gap1=-u.ms+u.m-4.0*G*u.qq+2.0*K*d.qq*s.qq;
    gap2=-d.ms+d.m-4.0*G*d.qq+2.0*K*u.qq*s.qq;
    gap3=-s.ms+s.m-4.0*G*s.qq+2.0*K*d.qq*u.qq;
  }

  return 0;
}

double nambujl_eos::iqq(double x, void *&pa) {
  double en, ret;
  njtp *p=(njtp *)pa;

  en=sqrt(x*x+p->ms*p->ms);
  ret=1.0;
  ret-=fermi_function(en,p->mu,p->temper,p->limit)+
    fermi_function(en,-p->mu,p->temper,p->limit);
  ret*=-3.0*x*x/en/pi2*p->ms;
  ret+=1.0;
  return ret;
}

double nambujl_eos::ide(double x, void *&pa) {
  double en, ret;
  njtp *p=(njtp *)pa;

  en=sqrt(x*x+p->ms*p->ms);
  ret=fermi_function(en,p->mu,p->temper,p->limit)-
    fermi_function(en,-p->mu,p->temper,p->limit);
  ret*=3.0*x*x/pi2;
  ret+=1.0;
  return ret;
}

double nambujl_eos::ied(double x, void *&pa) {
  double en, ret;
  njtp *p=(njtp *)pa;
  en=sqrt(x*x+p->ms*p->ms);
  ret=-en;
  ret+=en*fermi_function(en,p->mu,p->temper,p->limit)+
    en*fermi_function(en,-p->mu,p->temper,p->limit);
  ret*=3.0/pi2*x*x;
  return ret;
}

double nambujl_eos::ipr(double x, void *&pa) {
  double en, ret;
  njtp *p=(njtp *)pa;
  en=sqrt(x*x+p->ms*p->ms);
  ret=en;
  if ((p->mu-en)/p->temper>p->limit) {
    ret+=p->mu-en;
  } else if ((p->mu-en)/p->temper<-p->limit) {
    ret+=0.0;
  } else {
    ret+=p->temper*log(1+exp((p->mu-en)/p->temper));
  }
  if ((-en-p->mu)/p->temper>p->limit) {
    ret+=-en-p->mu;
  } else if ((-en-p->mu)/p->temper<-p->limit) {
    ret+=0.0;
  } else {
    ret+=p->temper*log(1+exp((-en-p->mu)/p->temper));
  }
  ret*=3.0/pi2*x*x;
  return ret;
}

template<> int io_tlate<nambujl_eos>::input
(cinput *co, in_file_format *ins, nambujl_eos *ne) {
  std::string t;
  co->object_in("quark_eos",ins,ne,t);
  ins->double_in(ne->L,"L");
  ins->double_in(ne->G,"G");
  ins->double_in(ne->K,"K");
  ins->double_in(ne->B0,"B0");
  ins->bool_in(ne->fromqq,"fromqq");
  ins->double_in(ne->limit);
  return 0;
}

template<> int io_tlate<nambujl_eos>::output
(coutput *co, out_file_format *outs, nambujl_eos *nj) {
  std::string t;
  co->object_out("quark_eos",outs,nj,0,0,t);
  outs->double_out(nj->L,"L");
  outs->double_out(nj->G,"G");
  outs->double_out(nj->K,"K");
  outs->double_out(nj->B0,"B0");
  outs->bool_out(nj->fromqq,"fromqq");
  outs->double_out(nj->limit);
  return 0;
}

template<> const char *io_tlate<nambujl_eos>::type()
{ return "nambujl_eos"; }
  
