/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/rmf_eos.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

rmf_eos::rmf_eos() {
  mnuc=939.0/hc_mev_fm;
  ms=500.0/hc_mev_fm;
  mr=770.0/hc_mev_fm;
  mw=783.0/hc_mev_fm;
  cs=3.815826;
  cw=3.088793;
  cr=2.041844;
  b=2.479608e-03;
  c=-3.315842e-03;
  zeta=0.0;
  xi=0.0;
  a1=0.0;
  a2=0.0;
  a3=0.0;
  a4=0.0;
  a5=0.0;
  a6=0.0;
  b1=0.0;
  b2=0.0;
  b3=0.0;
  zm_mode=false;
  guess_set=false;
  
  provides=provides_calc_p | provides_calc_e | provides_calc_temp_p |
    provides_calc_temp_e;
  
  sat_mroot=&def_sat_mroot;
}

int rmf_eos::calc_eq_temp_p(fermion &ne, fermion &pr, double temper,
			    double sig, double ome, double lrho, double &f1, 
			    double &f2, double &f3, thermo &lth) {
  
  double gs2, gs, gr2, gr, gw2, gw, duds;
  double us, fun, dfdw, fnn=0.0, fnp=0.0, sig2, sig4, ome2, ome4, rho2;
  
  if (temper<=0.0) {
    calc_eq_p(ne,pr,sig,ome,rho,f1,f2,f3,lth);
  }
  
  gs=ms*cs;
  gw=mw*cw;
  gr=mr*cr;
  gs2=gs*gs;
  gw2=gw*gw;
  gr2=gr*gr;
  
  ne.nu=ne.mu-gw*ome+0.5*gr*lrho;
  pr.nu=pr.mu-gw*ome-0.5*gr*lrho;
  
  if (zm_mode) {
    fnn=(1.0+gs*sig/ne.m);
    fnp=(1.0+gs*sig/pr.m);
    ne.ms=ne.m/fnn;
    pr.ms=pr.m/fnp;
  } else {
    ne.ms=ne.m-gs*sig;
    pr.ms=pr.m-gs*sig;
  }

  if (ne.ms<=0.0) ne.ms=1.e-9;
  if (pr.ms<=0.0) pr.ms=1.e-9;
  
  ne.non_interacting=false;
  pr.non_interacting=false;
  int ret1=ne.pair_mu(temper);
  if (ret1!=0) {
    add_err("Neutrons failed in rmf_eos::calc_temp_p(). Continuing.",
	    gsl_efailed);
  }
  int ret2=pr.pair_mu(temper);
  if (ret2!=0) {
    add_err("Protons failed in rmf_eos::calc_temp_p(). Continuing.",
	    gsl_efailed);
  }
  
  sig2=sig*sig;
  sig4=sig2*sig2;
  ome2=ome*ome;
  ome4=ome2*ome2;
  rho2=lrho*lrho;

  duds=b*ne.m*gs2*gs*sig2+c*gs2*gs2*sig2*sig;

  fun=a1*sig+a2*sig2+a3*sig2*sig+a4*sig4+
    a5*sig4*sig+a6*sig4*sig2+b1*ome2+b2*ome4+b3*ome4*ome2;
  dfdw=2.0*b1*ome+4.0*b2*ome2*ome+6.0*b3*ome4*ome;
  double nsn=1.0/ne.ms*(ne.ed-3.0*ne.pr);
  double nsp=1.0/pr.ms*(pr.ed-3.0*pr.pr);


  if (zm_mode==true) {
    f1=ms*ms*sig-gs*(nsn/fnn/fnn+nsp/fnp/fnp)+duds-gr2*rho2*
      (a1+2.0*a2*sig+3.0*a3*sig2+4.0*a4*sig2*sig+
       5.0*a5*sig4+6.0*a6*sig4*sig);
  } else {
    f1=ms*ms*sig-gs*(nsn+nsp)+duds-gr2*rho2*
      (a1+2.0*a2*sig+3.0*a3*sig2+4.0*a4*sig2*sig+
       5.0*a5*sig4+6.0*a6*sig4*sig);
  }
  f2=mw*mw*ome-gw*(ne.n+pr.n)+zeta*gw2*gw2*ome2*ome/6.0+gr2*rho2*
    (2.0*b1*ome+4.0*b2*ome2*ome+6.0*b3*ome4*ome);
  f3=mr*mr*lrho-gr*(pr.n-ne.n)*0.5+xi*gr2*gr2*rho2*lrho/6.0+
    2.0*gr2*lrho*fun;

  us=b/3.0*ne.m*gs2*gs*sig2*sig+c/4.0*gs2*gs2*sig4;
  
  lth.pr=-us-0.5*ms*ms*sig2+0.5*mw*mw*ome*ome+0.5*mr*mr*lrho*lrho+
    zeta/24.0*gw2*gw2*ome4+xi/24.0*gr2*gr2*rho2*rho2+
    ne.pr+pr.pr+gr2*rho2*fun;
  lth.ed=us+0.5*ms*ms*sig2+0.5*mw*mw*ome*ome+0.5*mr*mr*lrho*lrho+
    ne.ed+pr.ed+zeta/8.0*gw2*gw2*ome4+xi/8.0*gr2*gr2*rho2*rho2+
    gr2*rho2*(fun+ome*dfdw);
  lth.en=(lth.ed+lth.pr-ne.n*ne.mu-pr.n*pr.mu)/temper;

  if (ret1!=0 || ret2!=0) return gsl_efailed;

  return gsl_success;
}

int rmf_eos::field_eqs(size_t nv, const ovector_view &x, ovector_view &y, 
		       void *&vp) {
  calc_eq_p(*neutron,*proton,x[0],x[1],x[2],y[0],y[1],y[2],
	 *eos_thermo);
  return 0;
}

int rmf_eos::field_eqsT(size_t nv, const ovector_view &x, ovector_view &y, 
			void *&vp) {
  calc_eq_temp_p(*neutron,*proton,fe_temp,x[0],x[1],x[2],y[0],y[1],y[2],
	      *eos_thermo);
  return 0;
}

int rmf_eos::calc_p(fermion &ne, fermion &pr, thermo &lth) {
  int ret;
  ovector x(3);

  ne.non_interacting=false;
  pr.non_interacting=false;

  set_thermo(lth);
  set_n_and_p(ne,pr);
  
  if (guess_set) {
    x[0]=sigma;
    x[1]=omega;
    x[2]=rho;
    guess_set=false;
  } else {
    x[0]=0.3;
    x[1]=0.3;
    x[2]=-0.05;
  }
  
  mm_funct_mfptr<rmf_eos,void *> fmf(this,&rmf_eos::field_eqs);
  void *vpx=0;
  ret=eos_mroot->msolve(3,x,vpx,fmf);

  sigma=x[0];
  omega=x[1];
  rho=x[2];

  if (ret!=0) {
    add_err("msolve failed in rmf_eos::calc_p().",2);
    return 2;
  }

  return 0;
}

int rmf_eos::calc_temp_p(fermion &ne, fermion &pr, const double T,
			 thermo &lth) {
  int ret;
  ovector x(3);

  ne.non_interacting=false;
  pr.non_interacting=false;

  set_thermo(lth);
  set_n_and_p(ne,pr);
  fe_temp=T;
  
  if (guess_set) {
    x[0]=sigma;
    x[1]=omega;
    x[2]=rho;
    guess_set=false;
  } else {
    x[0]=0.3;
    x[1]=0.3;
    x[2]=-0.05;
  }
  
  mm_funct_mfptr<rmf_eos,void *> fmf(this,&rmf_eos::field_eqsT);
  void *vpx=0;
  ret=eos_mroot->msolve(3,x,vpx,fmf);

  sigma=x[0];
  omega=x[1];
  rho=x[2];

  if (ret!=0) {
    add_err("msolve failed in rmf_eos::calc_p().",2);
    return 2;
  }

  return 0;
}

int rmf_eos::calc_e(fermion &ne, fermion &pr, thermo &lth) {
  ovector x(5), y(5);
  int ret;
  
  ne.non_interacting=false;
  pr.non_interacting=false;

  set_thermo(lth);
  set_n_and_p(ne,pr);

  // If zero-density, then just return rest mass energy
  // Otherwise, set whether we are in neutron or proton matter mode
  if (ne.n<=0.0 && pr.n<=0.0) {
    ne.mu=ne.m;
    ne.ms=ne.m;
    ne.nu=ne.m;
    ne.ed=0.0;
    ne.pr=0.0;
    ne.en=0.0;
    pr.mu=pr.m;
    pr.ms=pr.m;
    pr.nu=pr.m;
    pr.ed=0.0;
    pr.pr=0.0;
    pr.en=0.0;
    return 0;
  } else if (ne.n<=0.0) {
    ne.n=0.0;
    ce_prot_matter=true;
  } else if (pr.n<=0.0) {
    pr.n=0.0;
    ce_neut_matter=true;
  } else {
    ce_neut_matter=false;
    ce_prot_matter=false;
  }
  
  n_baryon=ne.n+pr.n;
  n_charge=pr.n;
  
  mm_funct_mfptr<rmf_eos,void *> fmf(this,&rmf_eos::calc_e_solve_fun);
  void *vpx=0;
  
  if (guess_set) {
    
    // If an initial guess is given, then use it to directly compute
    // the EOS

    x[0]=ne.mu;
    x[1]=pr.mu;
    x[2]=sigma;
    x[3]=omega;
    x[4]=rho;
    guess_set=false;

    ret=eos_mroot->msolve(5,x,vpx,fmf);
    calc_e_solve_fun(5,x,y,vpx);
    
  } else {

    // If no initial guess is given, then create one by beginning
    // at saturated nuclear matter and proceeding incrementally.

    double nn=ne.n;
    double np=pr.n;

    if (ce_neut_matter) {
      x[0]=4.9;
      x[1]=4.7;
      x[2]=0.06;
      x[3]=0.04;
      x[4]=-0.04;
    } else if (ce_prot_matter) {
      x[0]=4.7;
      x[1]=4.9;
      x[2]=0.1;
      x[3]=0.07;
      x[4]=0.07;
    } else {
      x[0]=4.8;
      x[1]=4.8;
      x[2]=0.1;
      x[3]=0.07;
      x[4]=-0.001;
    }
    
    for(double alpha=0.0;alpha<=1.001;alpha+=0.1) {
      if (ce_prot_matter) {
	n_baryon=0.12*(1.0-alpha)+np*alpha;
	n_charge=n_baryon;
      } else if (ce_neut_matter) {
	n_baryon=0.12*(1.0-alpha)+nn*alpha;
	n_charge=0.0;
      } else {
	n_baryon=0.16*(1.0-alpha)+(nn+np)*alpha;
	n_charge=0.08*(1.0-alpha)+np*alpha;
      }
      
      ret=eos_mroot->msolve(5,x,vpx,fmf);
    }
    calc_e_solve_fun(5,x,y,vpx);
    
  }
  
  sigma=x[2];
  omega=x[3];
  rho=x[4];
  
  if (ret!=0) {
    add_err("msolve failed in rmf_eos::calc_e(fermion,fermion,thermo).",3);
    return 3;
  }

  return 0;
}

int rmf_eos::calc_temp_e(fermion &ne, fermion &pr, const double T, 
			 thermo &lth) {
  ovector x(5), y(5);
  int ret;
  
  ce_temp=T;

  ne.non_interacting=false;
  pr.non_interacting=false;

  set_thermo(lth);
  set_n_and_p(ne,pr);

  // If zero-density, then just return rest mass energy
  // Otherwise, set whether we are in neutron or proton matter mode
  if (ne.n<=0.0 && pr.n<=0.0) {
    ne.mu=ne.m;
    ne.ms=ne.m;
    ne.nu=ne.m;
    ne.ed=0.0;
    ne.pr=0.0;
    ne.en=0.0;
    pr.mu=pr.m;
    pr.ms=pr.m;
    pr.nu=pr.m;
    pr.ed=0.0;
    pr.pr=0.0;
    pr.en=0.0;
    return 0;
  }
  
  n_baryon=ne.n+pr.n;
  n_charge=pr.n;
  
  mm_funct_mfptr<rmf_eos,void *> fmf(this,&rmf_eos::calc_temp_e_solve_fun);
  void *vpx=0;
  
  if (guess_set) {
    
    // If an initial guess is given, then use it to directly compute
    // the EOS

    x[0]=ne.mu;
    x[1]=pr.mu;
    x[2]=sigma;
    x[3]=omega;
    x[4]=rho;
    guess_set=false;

    ret=eos_mroot->msolve(5,x,vpx,fmf);
    calc_e_solve_fun(5,x,y,vpx);
    
  } else {

    // If no initial guess is given, then create one by beginning
    // at saturated nuclear matter and proceeding incrementally.

    double nn=ne.n;
    double np=pr.n;

    x[0]=4.8;
    x[1]=4.8;
    x[2]=0.1;
    x[3]=0.07;
    x[4]=-0.001;
    
    for(double alpha=0.0;alpha<=1.001;alpha+=0.1) {
      n_baryon=0.16*(1.0-alpha)+(nn+np)*alpha;
      n_charge=0.08*(1.0-alpha)+np*alpha;
      
      ret=eos_mroot->msolve(5,x,vpx,fmf);
    }
    calc_temp_e_solve_fun(5,x,y,vpx);
    
  }
  
  sigma=x[2];
  omega=x[3];
  rho=x[4];
  
  if (ret!=0) {
    add_err
      ("msolve failed in rmf_eos::calc_temp_e(fermion,fermion,double,thermo).",
       3);
    return 3;
  }

  return 0;
}

int rmf_eos::calc_e(fermion &ne, fermion &pr, thermo &lth,
		    double &sig, double &ome, double &lrho) {

  ne.non_interacting=false;
  pr.non_interacting=false;

  ovector x(5);
  int ret;
  
  set_thermo(lth);
  set_n_and_p(ne,pr);

  if (ne.n<=0.0 && pr.n<=0.0) {
    ne.mu=ne.m;
    ne.ms=ne.m;
    ne.nu=ne.m;
    ne.ed=0.0;
    ne.pr=0.0;
    ne.en=0.0;
    pr.mu=pr.m;
    pr.ms=pr.m;
    pr.nu=pr.m;
    pr.ed=0.0;
    pr.pr=0.0;
    pr.en=0.0;
    return 0;
  } else if (ne.n<=0.0) {
    ne.n=0.0;
    ce_prot_matter=true;
  } else if (pr.n<=0.0) {
    pr.n=0.0;
    ce_neut_matter=true;
  } else {
    ce_neut_matter=false;
    ce_prot_matter=false;
  }
  
  n_baryon=ne.n+pr.n;
  n_charge=pr.n;
  
  x[0]=ne.nu;
  x[1]=pr.nu;
  x[2]=sig;
  x[3]=ome;
  x[4]=lrho;

  mm_funct_mfptr<rmf_eos,void *> fmf(this,&rmf_eos::calc_e_solve_fun);
  void *vpx=0;
  ret=eos_mroot->msolve(5,x,vpx,fmf);

  if (ret!=0) {
    add_err("msolve failed in rmf_eos::calc_e() (version with fields).",3);
    return 3;
  }

  sigma=x[2];
  omega=x[3];
  rho=x[4];

  return 0;
}

int rmf_eos::calc_eq_p(fermion &ne, fermion &pr, double sig, double ome, 
		       double lrho, double &f1, double &f2, double &f3, 
		       thermo &lth) {

  ne.non_interacting=false;
  pr.non_interacting=false;
  
  double duds,us,fun,gs2,gr2,gw2,gs,gr,gw,sig2,ome2;
  double rho2,sig4,ome4,dfdw,fnn=0.0,fnp=0.0;

  gs=ms*cs;
  gw=mw*cw;
  gr=mr*cr;
  gs2=gs*gs;
  gw2=gw*gw;
  gr2=gr*gr;

  if (zm_mode) {
    fnn=(1.0+gs*sig/ne.m);
    fnp=(1.0+gs*sig/pr.m);
    ne.ms=ne.m/fnn;
    pr.ms=pr.m/fnp;
  } else {
    ne.ms=ne.m-gs*sig;
    pr.ms=pr.m-gs*sig;
  }
  
  ne.nu=ne.mu-gw*ome+0.5*gr*lrho;
  pr.nu=pr.mu-gw*ome-0.5*gr*lrho;

  if (ne.nu<ne.ms || ne.ms<0.0) ne.kf=0.0;
  else ne.kf=sqrt(ne.nu*ne.nu-ne.ms*ne.ms);
  if (pr.nu<pr.ms || pr.ms<0.0) pr.kf=0.0;
  else pr.kf=sqrt(pr.nu*pr.nu-pr.ms*pr.ms);
  
  // We don't record error values, since these functions usually
  // always succeed
  ne.calc_mu_zerot();
  pr.calc_mu_zerot();

  sig2=sig*sig;
  sig4=sig2*sig2;
  ome2=ome*ome;
  ome4=ome2*ome2;
  rho2=lrho*lrho;

  duds=b*ne.m*gs2*gs*sig2+c*gs2*gs2*sig2*sig;
  
  fun=a1*sig+a2*sig2+a3*sig2*sig+a4*sig4+
    a5*sig4*sig+a6*sig4*sig2+b1*ome2+b2*ome4+b3*ome4*ome2;
  dfdw=2.0*b1*ome+4.0*b2*ome2*ome+6.0*b3*ome4*ome;
  double nsn=1.0/ne.ms*(ne.ed-3.0*ne.pr);
  double nsp=1.0/pr.ms*(pr.ed-3.0*pr.pr);

  if (zm_mode==true) {
    f1=ms*ms*sig-gs*(nsn/fnn/fnn+nsp/fnp/fnp)+duds-gr2*rho2*
      (a1+2.0*a2*sig+3.0*a3*sig2+4.0*a4*sig2*sig+
       5.0*a5*sig4+6.0*a6*sig4*sig);
  } else {
    f1=ms*ms*sig-gs*(nsn+nsp)+duds-gr2*rho2*
      (a1+2.0*a2*sig+3.0*a3*sig2+4.0*a4*sig2*sig+
       5.0*a5*sig4+6.0*a6*sig4*sig);
  }
  f2=mw*mw*ome-gw*(ne.n+pr.n)+zeta*gw2*gw2*ome2*ome/6.0+gr2*rho2*
    (2.0*b1*ome+4.0*b2*ome2*ome+6.0*b3*ome4*ome);
  f3=mr*mr*lrho-gr*(pr.n-ne.n)*0.5+xi*gr2*gr2*rho2*lrho/6.0+
    2.0*gr2*lrho*fun;

  us=b/3.0*ne.m*gs2*gs*sig2*sig+c/4.0*gs2*gs2*sig4;
  
  // The thermodynamic identity could be used instead of 
  // these explicit expressions, but it's nice to have them
  // available here.

  lth.pr=-us-0.5*ms*ms*sig2+0.5*mw*mw*ome*ome+0.5*mr*mr*lrho*lrho+
    zeta/24.0*gw2*gw2*ome4+xi/24.0*gr2*gr2*rho2*rho2+
    ne.pr+pr.pr+gr2*rho2*fun;

  lth.ed=us+0.5*ms*ms*sig2-0.5*mw*mw*ome*ome-0.5*mr*mr*lrho*lrho+
    ne.ed+pr.ed-zeta/24.0*gw2*gw2*ome4-xi/24.0*gr2*gr2*rho2*rho2-
    fun*gr2*rho2+gw*ome*(ne.n+pr.n)-0.5*gr*lrho*(ne.n-pr.n);

  return gsl_success;
}

int rmf_eos::fix_saturation_fun(size_t nv, const ovector_view &x, 
				ovector_view &y, void *&pa) {
  
  double phi,power,ome,dome,one,two,tri;
  double cq,lr,lr13,sqt,fir,sec,kf,kf2,pr,ed,lfac;
  double efs,rhos,kbar,lbar,us,duds,uv,eps,denr,aknm,mns;
  
  cs=x[0];
  cw=x[1]; 
  b=x[2];
  c=x[3];

  // b and c can be negative but if cw or cs are negative
  // then omega or sigma are negative respectively
  if (cs<0.0 || cw<0.0) return 1;

  // Calculate phi from effective mass
  phi=mnuc*(1.0-msom);
  mns=mnuc*msom;

  // Calculate omega and domega We don't have to worry about the
  // non-linear rho interactions, since rho is zero for nuclear
  // matter.

  // This section is just the solution of the vector meson field
  // equation
  power=1.0/3.0;
  if (zeta>0.0) {
    cq=8.0/(9.0*pow(cw*cw,3.0)*n0*n0*zeta);
    lr=3*n0/zeta;
    lr13=pow(lr,power);
    sqt=sqrt(1.0+cq);
    fir=pow((1.0+sqt),power);
    sec=pow(fabs(1.0-sqt),power);
    ome=lr13*(fir-sec);
    dome=1.0/(1.0/(cw*cw)+zeta/2.0*ome*ome);
  } else {
    ome=cw*cw*n0;
    dome=cw*cw;
  }

  kf=pow(1.5*pi2*n0,power);
  kf2=kf*kf;
  efs=sqrt(kf*kf+mns*mns);

  // Scalar density
  rhos=1.0/pi2*mns*(kf*efs-mns*mns*log((kf+efs)/mns));

  kbar=2.0*b*mnuc;
  lbar=6.0*c;
  us=kbar/6.0*pow(phi,3.0)+lbar/24.0*pow(phi,4.0);
  duds=kbar/2.0*phi*phi+lbar/6.0*pow(phi,3.0);
  uv=-zeta/24.0*pow(ome,4.0);

  // Energy density and pressure
  lfac=(kf+efs)/mns;
  pr=1.0/24.0/pi2*(2.0*efs*pow(kf,3.0)-3.0*kf*efs*mns*mns
                   +3.0*pow(mns,4.0)*log(lfac));
  ed=1.0/8.0/pi2*(2.0*kf*pow(efs,3.0)-kf*efs*mns*mns
                  -pow(mns,4.0)*log(lfac));
  eps=phi*phi/(2.0*cs*cs)+ome*n0-ome*ome/(2.0*cw*cw)+
    us+uv+ed*2.0;

  // Compressibility - This expression for the compressibility
  // only works for nuclear matter density. 
  one=1.0/cs/cs+kbar*phi+lbar/2.0*phi*phi;
  two=3.0/mns*rhos;
  tri=3.0*n0/efs;
  denr=one+two-tri;
  aknm=9.0*n0*(dome+kf2/(3.0*efs*n0)-pow(mns/efs,2.0)/denr);

  // Scalar field equation
  y[0]=phi/cs/cs+duds-rhos;

  // Binding energy of nuclear matter
  y[1]=(eps/n0-mnuc)-eoa;

  // Pressure
  y[2]=-phi*phi/(2.0*cs*cs)+ome*ome/(2.0*cw*cw)-us-uv+pr*2.0;

  // Compressibilty
  y[3]=aknm-comp;

  if (!finite(y[1]) || !finite(y[2]) || !finite(y[3]) || !finite(y[0])) {
    set_err_ret("Equation not finite in fix_saturation_fun().",gsl_efailed);
  }
  return 0;
}

int rmf_eos::fix_saturation(double gcs, double gcw, double gb, double gc) {
  int nvar, test;
  ovector x(4);
  double power,kf,kf2,kf3,efs;
  double mns,gw,ome,lr,lr13,cq,sqt,fir,sec,sig,gs;

  if (zm_mode==true) {
    set_err("Fix_Saturation() does not work for zm_mode.",gsl_eunimpl);
    return gsl_eunimpl;
  }

  nvar=4;

  x[0]=gcs;
  x[1]=gcw;
  x[2]=gb;
  x[3]=gc;

  mm_funct_mfptr<rmf_eos,void *> fmf(this,&rmf_eos::fix_saturation_fun);
  void *vpx=0;
  test=sat_mroot->msolve(4,x,vpx,fmf);
  if (test!=0) {
    add_err("Solve failed in fix_saturation().",test);
    return test;
  }

  cs=x[0];
  cw=x[1];
  b=x[2];
  c=x[3];

  power=1.0/3.0;
  kf=pow(1.5*pi2*n0,power);
  kf2=kf*kf;
  kf3=kf2*kf;
  mns=msom*mnuc;
  efs=sqrt(kf*kf+mns*mns);

  gw=mw*cw;
  if (zeta>0.0) {
    cq=8.0/(9.0*pow(cw*cw,3.0)*n0*n0*zeta);
    lr=3*n0/zeta;
    lr13=pow(lr,power);
    sqt=sqrt(1.0+cq);
    fir=pow((1.0+sqt),power);
    sec=pow(fabs(1.0-sqt),power);
    ome=lr13*(fir-sec);
  } else {
    ome=(cw*cw)*n0;
  }
  ome/=gw;
  
  gs=ms*cs;
  sig=mnuc*(1.0-msom)/gs;
  if (zm_mode) {
    set_err_ret("Function fix_saturation() does not work with zm_mode=true.",
		gsl_efailed);
  }

  if (calc_cr(sig,ome,n0)==0) {
    return 0;
  } 
  
  add_err("calc_cr() failed in fix_saturation().",gsl_efailed);
  return gsl_efailed;
}

int rmf_eos::saturation() {
  ovector x(5), y(5);
  int test;
  mm_funct_mfptr<rmf_eos,void *> fmf(this,&rmf_eos::zero_pressure);
  void *vpx=0;

  if (guess_set) {
    x[0]=neutron->mu;
    x[1]=proton->mu;
    x[2]=sigma;
    x[3]=omega;
    x[4]=rho;
    guess_set=false;
  } else {
    x[0]=4.8;
    x[1]=4.8;
    x[2]=0.1;
    x[3]=0.07;
    x[4]=0.0;
  }

  // Ensure that the initial guess has a finite initial density
  int it=0;
  if (x[0]<0.0) x[0]=0.0;
  if (x[1]<0.0) x[1]=0.0;
  zero_pressure(5,x,y,vpx);
  while (it<30 && (neutron->n<=0.0 || proton->n<=0.0)) {
    x[0]+=neutron->m*0.05;
    x[1]+=proton->m*0.05;
    zero_pressure(5,x,y,vpx);
    it++;
  } 
  if (it==20) {
    set_err_ret("Failed to make initial density finite in saturation()",
		gsl_efailed);
  }
  
  test=sat_mroot->msolve(5,x,vpx,fmf);
  
  sigma=x[2];
  omega=x[3];
  rho=x[4];

  if (test!=0) {
    add_err("Solver failed in rmf_eos::saturation().",gsl_efailed);
  }
  
  n0=neutron->n+proton->n;
  msom=neutron->ms/neutron->m;
  eoa=(eos_thermo->ed/n0-neutron->m);
  fkprime_fields(x[2],x[3],n0,comp,kprime);
  esym=fesym_fields(x[2],x[3],n0);
  
  return 0;
}

int rmf_eos::calc_e_solve_fun(size_t nv, const ovector_view &ex, 
			      ovector_view &ey, void *&pa) {
  double f1,f2,f3,sig,ome,lrho;

  neutron->mu=ex[0];
  proton->mu=ex[1];
  sig=ex[2];
  ome=ex[3];
  lrho=ex[4];

  calc_eq_p(*neutron,*proton,sig,ome,lrho,f1,f2,f3,*eos_thermo);
  
  if (!ce_prot_matter && neutron->nu<neutron->ms) {
    set_err_ret("n.nu<n.ms in rmf_eos::calc_e_solve_fun().",gsl_efailed);
  }
  if (!ce_neut_matter && proton->nu<proton->ms) {
    set_err_ret("p.nu<p.ms in rmf_eos::calc_e_solve_fun().",gsl_efailed);
  }

  if (ce_neut_matter) {
    ey[0]=proton->nu-proton->ms;
    ey[1]=neutron->n-n_baryon;
  } else if (ce_prot_matter) {
    ey[0]=neutron->nu-neutron->ms;
    ey[1]=proton->n-n_baryon;
  } else {
    ey[0]=proton->n+neutron->n-n_baryon;
    ey[1]=proton->n-n_charge;
  }
  ey[2]=f1;
  ey[3]=f2;
  ey[4]=f3;

  for(int i=0;i<5;i++) {
    if (!finite(ex[i]) || !finite(ey[i])) {
      set_err_ret((((string)"Eq. ")+itos(i)+
		   " not finite in rmf_eos::calc_e_solve_fun().").c_str(),
		  gsl_efailed);
    }
  }

  return 0;
}

int rmf_eos::calc_temp_e_solve_fun(size_t nv, const ovector_view &ex, 
				   ovector_view &ey, void *&pa) {
  double f1,f2,f3,sig,ome,lrho;

  neutron->mu=ex[0];
  proton->mu=ex[1];
  sig=ex[2];
  ome=ex[3];
  lrho=ex[4];
  
  calc_eq_temp_p(*neutron,*proton,ce_temp,sig,ome,lrho,f1,f2,f3,*eos_thermo);
  
  ey[0]=proton->n+neutron->n-n_baryon;
  ey[1]=proton->n-n_charge;
  ey[2]=f1;
  ey[3]=f2;
  ey[4]=f3;

  for(int i=0;i<5;i++) {
    if (!finite(ex[i]) || !finite(ey[i])) {
      set_err_ret((((string)"Eq. ")+itos(i)+
		   " not finite in rmf_eos::calc_temp_e_solve_fun().").c_str(),
		  gsl_efailed);
    }
  }

  return 0;
}

int rmf_eos::zero_pressure(size_t nv, const ovector_view &ex, 
			   ovector_view &ey, void *&pa) {
  double f1,f2,f3,sig,ome,lrho;
  fermion *n=neutron, *p=proton;
  int i;
  
  n->mu=ex[0];
  p->mu=ex[1];
  sig=ex[2];
  ome=ex[3];
  lrho=ex[4];

  for(i=0;i<5;i++) {
    if (!finite(ex[i])) {
      set_err_ret("Variable not finite in zero_pressure()",gsl_efailed);
    }
  }
  
  calc_eq_p(*n,*p,sig,ome,lrho,f1,f2,f3,*eos_thermo);
  
  if (n->n<=0.0 || p->n<=0.0) {
    set_err_ret("Densities less than zero in zero_pressure().",gsl_efailed);
  }

  ey[0]=eos_thermo->pr;
  ey[1]=p->n/(n->n+p->n)-0.5;
  ey[2]=f1;
  ey[3]=f2;
  ey[4]=f3;

  for(i=0;i<5;i++) {
    if (!finite(ex[i]) || !finite(ey[i])) {
      set_err_ret((((string)"Eq. ")+itos(i)+
		   " not finite in zero_pressure()").c_str(),gsl_efailed);
    }
  }
  return 0;
}

double rmf_eos::fesym_fields(double sig, double ome, double l_nb) {
  double kf, efs, mstar, ret, fun;

  kf=pow(1.5*l_nb*pi2,1.0/3.0);
  mstar=mnuc-cs*ms*sig;
  efs=sqrt(mstar*mstar+kf*kf);

  fun=a1*sig+a2*sig*sig+a3*pow(sig,3.0)+a4*pow(sig,4.0)+
    a5*pow(sig,5.0)+a6*pow(sig,6.0)+b1*ome*ome+b2*pow(ome,4.0)+
    b3*pow(ome,6.0);
  
  ret=kf*kf/6.0/efs+l_nb/8.0/(1.0/cr/cr+2.0*fun);
  
  return ret;
}

int rmf_eos::calc_cr(double sig, double ome, double l_nb) {
  double kf, efs, mstar, up, dn, fun;

  kf=pow(1.5*l_nb*pi2,1.0/3.0);
  mstar=mnuc-cs*ms*sig;
  efs=sqrt(mstar*mstar+kf*kf);

  fun=a1*sig+a2*sig*sig+a3*pow(sig,3.0)+a4*pow(sig,4.0)+
    a5*pow(sig,5.0)+a6*pow(sig,6.0)+b1*ome*ome+b2*pow(ome,4.0)+
    b3*pow(ome,6.0);

  up=4.0*kf*kf-24.0*esym*efs;
  dn=fun*(48.0*esym*efs-8.0*kf*kf)-3.0*l_nb*efs;
  
  cr=sqrt(up/dn);

  if (!finite(cr)) {
    set_err_ret("cr not finite in rmf_eos::calc_cr()",gsl_efailed);
  }

  return 0;
}

double rmf_eos::fcomp_fields(double sig, double ome, double l_nb) {
  double gs, gw, mstar, efs, d2u;
  double alpha, dsdn, dwdn, ret, rhos, kf;

  gs=ms*cs;
  gw=mw*cw;
  mstar=mnuc-gs*sig;
  kf=pow(1.5*pi2*l_nb,1.0/3.0);
  efs=sqrt(kf*kf+mstar*mstar);
  rhos=(mstar*kf*efs-pow(mstar,3.0)*log((kf+efs)/mstar))/pi2;

  d2u=pow(gs,3.0)*(2.0*b*mnuc*sig+3.0*c*gs*sig*sig);

  alpha=0.5/pi2/efs*(2.0*pow(kf,3.0)+6.0*mstar*mstar*kf-
		     6.0*mstar*mstar*efs*log((kf+efs)/mstar));
  dsdn=gs*mstar/efs/(ms*ms+d2u+gs*gs*alpha);
  dwdn=gw/(mw*mw+zeta/2.0*pow(gw*gw*ome,2.0));

  ret=dwdn*(mw*mw*ome+zeta/6.0*pow(gw*ome,3.0)*gw);
  ret+=kf*kf/3.0/efs;
  ret+=dsdn*(gs*mstar*alpha/3.0-gs*rhos);
  ret*=9.0;

  return ret;
}
  
int rmf_eos::fkprime_fields(double sig, double ome, double l_nb,
			    double &k, double &l_kprime) {
  double gs, gw, mstar, efs, d2u;
  double alpha, dsdn, dwdn, rhos, kf, lterm;
  double dpdn, d3u, dalphadms, d2sdn2, d2pdn2, d2wdn2;

  gs=ms*cs;
  gw=mw*cw;
  mstar=mnuc-gs*sig;
  kf=pow(1.5*pi2*l_nb,1.0/3.0);
  efs=sqrt(kf*kf+mstar*mstar);
  lterm=log((kf+efs)/mstar);
  rhos=(mstar*kf*efs-pow(mstar,3.0)*lterm)/pi2;

  // d2u is d^2(U(sigma))/d(sigma)
  d2u=pow(gs,3.0)*(2.0*b*mnuc*sig+3.0*c*gs*sig*sig);

  // alpha is d(scalar density)/d(mstar)
  alpha=1.0/pi2/efs*(pow(kf,3.0)+3.0*mstar*mstar*kf-
		     3.0*mstar*mstar*efs*lterm);
  // d(sigma)/dn
  dsdn=gs*mstar/efs/(ms*ms+d2u+gs*gs*alpha);
  // d(omega)/dn
  dwdn=gw/(mw*mw+zeta/2.0*pow(gw*gw*ome,2.0));

  // d(Pressure)/dn
  dpdn=dwdn*(mw*mw*ome+zeta/6.0*pow(gw*ome,3.0)*gw);
  dpdn+=kf*kf/3.0/efs;
  dpdn+=dsdn*(gs*mstar*alpha/3.0-gs*rhos);

  // d3u is d^3(U(sigma))/d(sigma)
  d3u=pow(gs,3.0)*(2.0*b*mnuc+6.0*c*gs*sig);

  dalphadms=-(8.0*pow(kf,3.0)+6.0*kf*mstar*mstar-6.0*pow(efs,3.0)*
	      lterm)*mstar/pi2/pow(efs,3.0);

  // d^2(sigma)/dn^2
  d2sdn2=-pi2/2.0/kf/efs/efs*dsdn+pow(dsdn,3.0)*
    (gs*efs/mstar*dalphadms-efs/mstar/gs*d3u);

  // d^2(omega)/dn^2
  d2wdn2=-pow(gw*dwdn,3.0)*zeta*ome;

  // d^2(Pressure)/dn^2
  d2pdn2=dsdn*(2.0*gs*kf*kf*mstar/3.0/pow(efs,3.0)-gs*mstar/efs)+
    dsdn*dsdn*(2*gs*gs*alpha/3.0-gs*gs*mstar/3.0*dalphadms)+
    d2sdn2*(gs*mstar*alpha/3.0-gs*rhos)+
    dwdn*dwdn*(mw*mw+zeta/2.0*pow(gw*gw*ome,2.0))+
    d2wdn2*(mw*mw*ome+zeta/6.0*gw*pow(gw*ome,3.0))+
    pi2/3.0/efs*(1.0/kf-kf/2.0/efs/efs);

  k=9.0*dpdn;
  l_kprime=27.0*l_nb*(d2pdn2-4.0*dpdn/l_nb);
  
  return 0;
}

int rmf_eos::load(string model, bool external) {

  file_detect fd;

  bool_io_type bool_io;
  double_io_type double_io;
  string_io_type string_io;

  collection co;
  double gs,gw,gr;
  bool oakstyle, tokistyle, standardstyle;
  string fname;
  if (external) {
    fname=model;
  } else {
    fname=lib_settings.get_data_dir()+"rmfdata/"+model;
  }

  in_file_format *iff=fd.open(fname.c_str());
  if (co.fin(iff)!=0) {
    delete iff;
    set_err((((string)"Couldn't open file '")+fname+"' in load().").c_str(),
	    gsl_efailed);
    return gsl_efailed;
  }
  fd.close();
  
  if (bool_io.get_def(co,"oakstyle",oakstyle)!=0) oakstyle=false;
  if (bool_io.get_def(co,"tokistyle",tokistyle)!=0) tokistyle=false;
  if (bool_io.get_def(co,"standardstyle",standardstyle)!=0) 
    standardstyle=false;
  
  double_io.get_def(co,"ms",ms);
  double_io.get_def(co,"mw",mw);
  double_io.get_def(co,"mr",mr);
  double_io.get_def(co,"mnuc",mnuc);
  ms/=hc_mev_fm; mw/=hc_mev_fm; mr/=hc_mev_fm; mnuc/=hc_mev_fm;
  
  if (standardstyle==true) {
    double_io.get_def(co,"Cs2",cs);
    double_io.get_def(co,"Cw2",cw);
    double_io.get_def(co,"Cr2",cr);
    double_io.get_def(co,"b",b);
    double_io.get_def(co,"c",c);
    cs=sqrt(cs)/mnuc;
    cw=sqrt(cw)/mnuc;
    cr=sqrt(cr)/mnuc;
  } else if (oakstyle==true || tokistyle==true) {
    double_io.get_def(co,"gs",gs);
    double_io.get_def(co,"gw",gw);
    double_io.get_def(co,"gr",gr);
    double_io.get_def(co,"g2",b);
    double_io.get_def(co,"g3",c);
    b/=-mnuc*pow(fabs(gs),3.0);
    c/=pow(gs,4.0);
    gr*=2.0;
    cs=gs/ms;
    cw=gw/mw;
    cr=gr/mr;
  } else {
    double_io.get_def(co,"cs",cs);
    double_io.get_def(co,"cw",cw);
    double_io.get_def(co,"cr",cr);
    double_io.get_def(co,"b",b);
    double_io.get_def(co,"c",c);
  }

  if (tokistyle==true) {
    double_io.get_def(co,"zeta",zeta,0.0);
    zeta=zeta/gw/gw/gw/gw*6.0;
  } else {
    double_io.get_def(co,"zeta",zeta,0.0);
  }
  
  double_io.get_def(co,"xi",xi,0.0);
  double_io.get_def(co,"a1",a1,0.0);
  double_io.get_def(co,"a2",a2,0.0);
  double_io.get_def(co,"a3",a3,0.0);
  double_io.get_def(co,"a4",a4,0.0);
  double_io.get_def(co,"a5",a5,0.0);
  double_io.get_def(co,"a6",a6,0.0);
  double_io.get_def(co,"b1",b1,0.0);
  double_io.get_def(co,"b2",b2,0.0);
  double_io.get_def(co,"b3",b3,0.0);

  return 0;
}

template<> int io_tlate<rmf_eos>::input
(cinput *co, in_file_format *ins, rmf_eos *sk) {
  std::string t;
  co->object_in("hadronic_eos",ins,sk,t);
  ins->double_in(sk->mnuc,"mnuc");
  ins->double_in(sk->ms,"ms");
  ins->double_in(sk->mw,"mw");
  ins->double_in(sk->mr,"mr");
  ins->double_in(sk->cs,"cs");
  ins->double_in(sk->cw,"cw");
  ins->double_in(sk->cr,"cr");
  ins->double_in(sk->b,"b");
  ins->double_in(sk->c,"c");
  ins->double_in(sk->zeta,"zeta");
  ins->double_in(sk->xi,"xi");
  ins->double_in(sk->a1,"a1");
  ins->double_in(sk->a2,"a2");
  ins->double_in(sk->a3,"a3");
  ins->double_in(sk->a4,"a4");
  ins->double_in(sk->a5,"a5");
  ins->double_in(sk->a6,"a6");
  ins->double_in(sk->b1,"b1");
  ins->double_in(sk->b2,"b2");
  ins->double_in(sk->b2,"b2");
  return 0;
}

template<> int io_tlate<rmf_eos>::output
(coutput *co, out_file_format *outs, rmf_eos *sk) {
  std::string t;
  co->object_out("hadronic_eos",outs,sk,0,0,t);
  outs->double_out(sk->mnuc,"mnuc");
  outs->double_out(sk->ms,"ms");
  outs->double_out(sk->mw,"mw");
  outs->double_out(sk->mr,"mr");
  outs->double_out(sk->cs,"cs");
  outs->double_out(sk->cw,"cw");
  outs->double_out(sk->cr,"cr");
  outs->double_out(sk->b,"b");
  outs->double_out(sk->c,"c");
  outs->double_out(sk->zeta,"zeta");
  outs->double_out(sk->xi,"xi");
  outs->double_out(sk->a1,"a1");
  outs->double_out(sk->a2,"a2");
  outs->double_out(sk->a3,"a3");
  outs->double_out(sk->a4,"a4");
  outs->double_out(sk->a5,"a5");
  outs->double_out(sk->a6,"a6");
  outs->double_out(sk->b1,"b1");
  outs->double_out(sk->b2,"b2");
  outs->double_out(sk->b3,"b3");
  return 0;
}

template<> const char *io_tlate<rmf_eos>::type()
{ return "rmf_eos"; }


