/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_RMF_EOS_H
#define O2SCL_RMF_EOS_H

#include <string>
#include <cmath>
#include <o2scl/lib_settings.h>
#include <o2scl/constants.h>
#include <o2scl/collection.h>
#include <o2scl/part.h>
#include <o2scl/hadronic_eos.h>
#include <o2scl/fermion.h>
#include <o2scl/mm_funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Relativistic mean field theory EOS
      
      Before sending neutrons and protons to these member functions,
      the masses should be set to their vacuum values and the
      degeneracy factor should be 2. If an internal model is used
      (using load()), then the neutron and proton masses should be set
      to \ref mnuc.
    
      It is important to point out that expressions for the energy
      densities are often simplified in the literature using the field
      equations. These expressions are not used in this code since
      they are only applicable in infinite matter where the field
      equations hold, and are not suitable for use in applications
      (such as to finite nuclei) where the spatial derivatives of the
      fields are non-zero. Notice that in the proper expressions for
      the energy density the similarity between terms in the pressure
      up to a sign. This procedure allows one to verify the
      thermodynamic identity even if the field equations are not
      solved and allows the user to add gradient terms to the energy
      density and pressure.

      \note Since this EOS uses the effective masses and chemical
      potentials in the fermion class, the values of
      part::non_interacting for neutrons and protons are set to false
      in many of the functions.

      \todo 
      - The number of couplings is getting unmanagable, maybe new
      organization is required.
      - Check the formulas in the "Background" section
      - Overload hadronic_eos::fcomp() with an exact version
      - Fix calc_p() to be better at guessing
      - There are two calc_e() functions that solve. One is
      specially designed to work without a good initial guess.
      Possible the other calc_e() function should be 
      similarly designed?
      - Make sure that this class properly handles particles for which 
      inc_rest_mass is true/false
      - It might be nice to remove explicit reference to the
      meson masses in functions which only compute nuclear matter
      since they are unnecessary. This might, however, demand 
      redefining some of the couplings. 

      \hline

      \b Background

      The full Lagragian is:

      \f[
      {\cal L} = {\cal L}_{Dirac} + {\cal L}_{\sigma} + 
      {\cal L}_{\omega} + {\cal L}_{\rho} 
      \f]
    
      \f{eqnarray*}
      {\cal L}_{Dirac} &=& 
      \bar{\Psi} \left[ i {{\partial}\!\!\!{\slash}} - 
      g_{\omega} {{\omega}\!\!\!{\slash}} - \frac{g_{\rho}}{2} 
      {{\vec{\rho}}\!\!\!{\slash}}
      \vec{\tau} - M + g_{\sigma} \sigma - \frac{e}{2} 
      \left( 1 + \tau_3 \right) A_{\mu} \right] \Psi \nonumber \\
      {\cal L}_{\sigma} &=& 
      {\textstyle \frac{1}{2}} \left( \partial_{\mu} \sigma \right)^2 
      - {\textstyle \frac{1}{2}} m^2_{\sigma} \sigma^2 
      - \frac{b M}{3} \left( g_{\sigma} \sigma\right)^3 
      - \frac{c}{4} \left( g_{\sigma} \sigma\right)^4  \nonumber \\
      {\cal L}_{\omega} &=& 
      - {\textstyle \frac{1}{4}} f_{\mu \nu} f^{\mu \nu} 
      + {\textstyle \frac{1}{2}} m^2_{\omega}\omega^{\mu}\omega_{\mu} 
      + \frac{\zeta}{24} g_{\omega}^4 \left(\omega^\mu \omega_\mu\right)^2
      \nonumber \\
      {\cal L}_{\rho} &=& 
      - {\textstyle \frac{1}{4}} \vec{B}_{\mu \nu} \cdot \vec{B}^{\mu \nu}
      + {\textstyle \frac{1}{2}} m^2_{\rho} \vec{\rho}^{~\mu} \cdot 
      \vec{\rho}_{~\mu} 
      + \frac{\xi}{24} g_{\rho}^4 \left(\vec{\rho}^{~\mu}\right) \cdot 
      \vec{\rho}_{~\mu} 
      + g_{\rho}^2 f (\sigma, \omega) \vec{\rho}^{~\mu} \cdot 
      \vec{\rho}_{~\mu} \nonumber \\
      \f}
    
      The coefficients \f$ b \f$ and \f$ c \f$ are related to the somewhat
      standard \f$ \kappa \f$ and \f$ \lambda \f$ by:
    
      \f[
      \kappa=2 M b \quad \lambda=6 c;
      \f]
    
      The function \f$ f \f$ is the coefficient of \f$ g_r^2 \rho^2 \f$ 
    
      \f$ f(\sigma,\omega) = b_1 \omega^2 + b_2 \omega^4 + b_3 \omega^6 +
      a_1 \sigma + a_2 \sigma^2 + a_3 \sigma^3 + a_4 \sigma^4 +
      a_5 \sigma^5 + a_6 \sigma^6 \f$ 
    
      where the notation from \ref Horowitz01 is:
    
     \f$ f(\sigma,\omega) = \lambda_4 g_s^2 \sigma^2 + 
      \lambda_v g_w^2 \omega^2 \f$ 
    
      This implies \f$ b_1=\lambda_v g_w^2 \f$ and 
     \f$ a_2=\lambda_4 g_s^2 \f$ 
    
      The field equations are:
    
      \f[
      0 = m_{\sigma}^2 \sigma - g_{\sigma} \left( n_{s n} + n_{s p} \right)
      + b M g_{\sigma}^3 \sigma^2 + c g_{\sigma}^4 \sigma^3 -
      g_{\rho}^2 \rho^2 \frac{\partial f}{\partial \sigma}
      \f]

      \f[
      0 = m_{\omega}^2 \omega - g_{\omega} \left(n_n+n_p\right)
      + \frac{\zeta}{6} g_{\omega}^4 \omega^3 g_{\rho}^2 \rho^2 
      \frac{\partial f}{\partial \omega}
      \f]

      \f[
      0 = m_{\rho}^2 \rho + \frac{1}{2} g_{\rho} \left(n_n-n_p\right)
      + 2 g_{\rho}^2 \rho f + \frac{\xi}{6} g_{\rho}^4 \rho^3
      \f]

      When the variable zm_mode is true, the effective mass is fixed using
      the approach of \ref Zimanyi90 .
    
      Defining
    
      \f[
      U(\sigma)=\frac{1}{2} m_\sigma^2\sigma^2+\frac{b M}{3}(g_\sigma\sigma)^3
      +\frac{c}{4}(g_\sigma\sigma)^4\;, 
      \f]
    
      the binding energy per particle in symmetric matter at equilibrium
      is given by
    
      \f[
      \frac{E}{A} = \frac{1}{n_0} \left[U(\sigma_0)+
      \frac{1}{2} m_\omega\omega_0^2+
      \frac{\zeta}{8}(g_\omega\omega_0)^4+\frac{2}{\pi^2}
      \int\limits_0^{k_F} dk k^2\sqrt{k^2+M^{*2}} \right]\;.
      \f]
    
      where the Dirac
      effective mass is  \f$ M^{*} = M - g_{\sigma}\sigma_0 \f$ .
      The compressibility is given by
    
      \f[
      K=9\frac{g_\omega^2}{m_\omega^2}n_0+3\frac{k_F^2}{E_F^*}
      -9n_0\frac{M^{*2}}{E_F^{*2}}\left[\left(\frac{1}{g_\sigma^2}
      \frac{\partial^2}{\partial\sigma_0^2}+\frac{3}{g_\sigma M^*}
      \frac{\partial}{\partial\sigma_0}\right)
      U(\sigma_0)-3\frac{n_0}{E_F^*}\right]^{-1}\;.
      \f]
    
      The symmetry energy of bulk matter is given by
    
      \f[
      E_{sym} = \frac{k_F^2}{6 E_F^{*}} + \frac{ n }
      {8 \left(g_{\rho}^2/m_{\rho}^2 + 2 f (\sigma_0, \omega_0) \right)} 
      \f]
    
      In the above equations, the subscipt `` \f$ 0 \f$ '' denotes the mean
      field values of \f$ \sigma \f$ and \f$ \omega \f$ .  For the case
      \f$ f=0 \f$ , the symmetry energy varies linearly with the density at
      large densities. The function \f$ f \f$ permits variations in the
      density dependence of the symmetry energy above nuclear matter
      density.
    
      See also \ref Muller96, \ref Zimanyi90.
      
  */
  class rmf_eos : public hadronic_eos {
  public:

    /// Modifies method of calculating effective masses
    bool zm_mode;

    /// \name Masses
    //@{

    /** 
	\brief nucleon mass

	This need not be exactly equal to the neutron or proton mass, 
	but provides the scale for the coupling \c b.
    */
    double mnuc;

    /// \f$ \sigma \f$ mass (in \f$ \mathrm{fm}^{-1} \f$ )
    double ms;

    /// \f$ \omega \f$ mass (in \f$ \mathrm{fm}^{-1} \f$ )
    double mw;

    /// \f$ \rho \f$ mass (in \f$ \mathrm{fm}^{-1} \f$ )
    double mr;

    //@}

    /// \name Standard couplings (including nonlinear sigma terms)
    //@{
    double cs, cw, cr, b, c;
    //@}

    /// \name Non-linear terms for omega and rho.
    //@{
    double zeta, xi;
    //@}
  
    /// \name Additional isovector couplings
    //@{
    double a1, a2, a3, a4, a5, a6, b1, b2, b3;
    //@}

    rmf_eos();

    /** 
	\brief Load parameters for model named 'model'
	
	Presently accepted values from file rmfdata/model_list:
	\include rmfdata/model_list
	
	In these files, the nucleon and meson masses are by default
	specified in MeV, and cs, cw, and cr are given in fm. The
	parameters b and c are both unitless. If the bool 'oakstyle' is
	true, then load() assumes that gs, gw, and gr have been given
	where gs and gw are as usual, but gr is a factor of two smaller
	than usual, and g2 and g3 have been given where g2 = -b M gs^3
	and g3 = c gs^4. If tokistyle is true, then it is additionally
	assumed that c3 is given where c3=zeta/6*gw^4.
	
	If \c external is true, then model is the filename (relative
	to the current directory) of the file containing the model
	parameters. Otherwise, the model is assumed to be present in
	the O2scl library data directory.
    */
    int load(std::string model, bool external=false);

    /// \name Compute EOS from densities at zero temperature
    //@{
    /** 
	\brief Equation of state as a function of density

	Initial guesses for the chemical potentials are taken
	from the user-given values. Initial guesses for the fields
	can be set by set_fields(), or default values will be used.
	After the call to calc_e(), the final values of the fields
	can be accessed through get_fields(). 

	This is a little more robust than the standard version
	in the parent \ref hadronic_eos.
	
	\todo This doesn't always work well when the proton density
	is zero.
       
    */
    virtual int calc_e(fermion &ne, fermion &pr, thermo &lth);
  
    /** 
	\brief Equation of state as a function of density
	
	\todo This doesn't always work well when the proton density
	is zero.
	\todo Rename this function to distinguish between calc_e()'s
       
    */
    virtual int calc_e(fermion &ne, fermion &pr, thermo &lth,
		       double &sig, double &ome, double &rho);
    //@}
  
    /// \name Compute EOS from chemical at zero temperature
    //@{
    /**  
	 \brief Equation of state as a function of chemical potential
	 
	 Solves for the field equations automatically.
	 
	 \note This may not be too robust. Fix?
    */
    virtual int calc_p(fermion &ne, fermion &pr, thermo &lth);

    /** 
	\brief Equation of state and meson field equations 
	as a function of chemical potentials
      
	This calculates the pressure and energy density as a function of
	\f$ \mu_n,\mu_p,\sigma,\omega,rho \f$ . When the field equations
	have been solved, \c f1, \c f2, and \c f3 are all zero. 
	
	The thermodynamic identity is satisfied even when the field
	equations are not solved.
    */
    virtual int calc_eq_p(fermion& neu, fermion& p, double sig, double ome, 
			  double rho, double& f1, double& f2, double& f3, 
			  thermo& th);
    //@}

    /// \name Compute EOS from chemical potentials at finite temperature
    //@{
    /** \brief Equation of state and meson field equations as a 
	function of chemical potentials
    */
    virtual int calc_eq_temp_p(fermion &ne, fermion &pr, const double temper, 
			       double sig, double ome, double rho, double &f1, 
			       double &f2, double &f3, thermo &th);

    /**  
	 \brief Equation of state as a function of chemical potential
	 
	 Solves for the field equations automatically.
    */
    virtual int calc_temp_p(fermion &ne, fermion &pr, const double T,
			    thermo &lth);
    //@}

    int calc_temp_e(fermion &ne, fermion &pr, const double T, 
		    thermo &lth);
      
    /** 
	\brief Calculate cs, cw, cr, b, and c from the saturation 
	properties

	Note that the meson masses and \ref mnuc must be specified
	before calling this function.

	This function does not give correct results when bool zm_mode 
	is true. 
	
	\c guess_cs, \c guess_cw, \c guess_b, and \c guess_c are
	initial guesses for \c cs, \c cw, \c b, and \c c respectively.

	\todo 
	- Fix this for zm_mode=true
	- Ensure solver is more robust
	
    */
    int fix_saturation(double guess_cs=4.0, double guess_cw=3.0, 
		       double guess_b=0.001, double guess_c=-0.001);
    
    /** \brief Calculate properties for nuclear matter at the
	saturation density
	
	This requires initial guesses to the chemical 
	potentials, etc.
    */
    virtual int saturation();
  
    /** \brief Calculate symmetry energy assuming the field
	equations have already been solved
	
	This may only work at saturation density. Used by saturation().
    */
    double fesym_fields(double sig, double ome, double nb);

    /** \brief Calculate the compressibility assuming the field
	equations have already been solved
	
	This may only work at saturation density.
    */
    double fcomp_fields(double sig, double ome, double nb);

    /** \brief Calculate compressibilty and \c kprime assuming the field
	equations have already been solved

	This may only work at saturation density. Used by
	saturation(). 

	\todo Does this work? Fix fkprime_fields() if it does not.
    */
    int fkprime_fields(double sig, double ome, double nb,
		       double &k, double &kprime);
    
    /** 
	\brief A function for solving the field equations

	x[0], x[1], and x[2] should be set to \f$ \sigma, \omega \f$ , 
	and \f$ \rho \f$ on input (in \f$ \mathrm{fm}^{-1} \f$ ) and
	on exit, y[0], y[1] and y[2] contain the field equations 
	and are zero when the field equations have been solved.
	The \c pa parameter is ignored.
    */
    int field_eqs(size_t nv, const ovector_view &x, ovector_view &y, 
		  void *&pa);

    /** 
	\brief A function for solving the field equations at finite 
	temperature
	
	x[0], x[1], and x[2] should be set to \f$ \sigma, \omega \f$ , 
	and \f$ \rho \f$ on input (in \f$ \mathrm{fm}^{-1} \f$ ) and
	on exit, y[0], y[1] and y[2] contain the field equations 
	and are zero when the field equations have been solved.
	The \c pa parameter is ignored.
    */
    int field_eqsT(size_t nv, const ovector_view &x, ovector_view &y, 
		   void *&pa);

    /** \brief Set a guess for the fields for the next call to calc_e(), 
	calc_p(), or saturation()
    */
    virtual int set_fields(double sig, double ome, double lrho) {
      sigma=sig;
      omega=ome;
      rho=lrho;
      guess_set=true;
      return 0;
    }

    /** 
	\brief Return the most recent values of the meson fields 
	
	This returns the most recent values of the meson fields set by
	a call to saturation(), calc_e(), or calc_p(fermion &, fermion
	&, thermo &).
    */
    int get_fields(double &sig, double &ome, double &lrho) {
      sig=sigma;
      ome=omega;
      lrho=rho;
      return 0;
    }

    /// Return string denoting type ("rmf_eos")
    virtual const char *type() { return "rmf_eos"; }

    // Removed cause we can use sat_mroot
    // The default solver for calculating the saturation properties
    // gsl_mroot_hybrids<void *,mm_funct<void *> > def_cp_mroot;

    /** \brief Set the coefficients of a rmf_eos object to their 
	limits from naturalness

	As given in muller and Serot, npa 606, 508

	The definition of the vector-isovector field and coupling
	matches what is done here. Compare the Lagrangian above
	with Eq. 10 from the reference.

	The following couplings should all be of the same
	size:
	\f[
	\frac{1}{2 c_s^2 M^2}, \frac{1}{2 c_v^2 M^2} 
	\frac{1}{8 c_{\rho}^2 M^2},~\mathrm{and}~\frac{
	\bar{a}_{ijk} M^{i+2 j+2 k-4}}{2^{2 k}}
	\f]
	which are equivalent to 
	\f[
	\frac{m_s^2}{2 g_s^2 M^2}, \frac{m_v^2}{2 g_v^2 M^2} 
	\frac{m_{\rho}^2}{8 g_{\rho}^2 M^2},~\mathrm{and}~\frac{
	a_{ijk} M^{i+2 j+2 k-4}}{g_s^i g_v^{2 j} 
	g_{\rho}^{2 k} 2^{2 k}}
	\f]
	
	The connection the \f$ a_{ijk} \f$ 's and the coefficients 
	that are used here is 
	\f{eqnarray*}
	\frac{b M}{3} g_{\sigma}^3 \sigma^3 &=& a_{300}~\sigma^3
	\nonumber \\
	\frac{c}{4} g_{\sigma}^4 \sigma^4 &=& a_{400}~\sigma^4
	\nonumber \\
	\frac{\zeta}{24} g_{\omega}^4 \omega^4 &=& a_{020}~\omega^4
	\nonumber \\
	\frac{\xi}{24} g_{\rho}^4 \rho^4 &=& a_{002}~\rho^4
	\nonumber \\
	b_1 g_{\rho}^2 \omega^2 \rho^2 &=& a_{011}~\omega^2 \rho^2 
	\nonumber \\
	b_2 g_{\rho}^2 \omega^4 \rho^2 &=& a_{021}~\omega^4 \rho^2 
	\nonumber \\
	b_3 g_{\rho}^2 \omega^6 \rho^2 &=& a_{031}~\omega^6 \rho^2 
	\nonumber \\
	a_1 g_{\rho}^2 \sigma^1 \rho^2 &=& a_{101}~\sigma^1 \rho^2 
	\nonumber \\
	a_2 g_{\rho}^2 \sigma^2 \rho^2 &=& a_{201}~\sigma^2 \rho^2 
	\nonumber \\
	a_3 g_{\rho}^2 \sigma^3 \rho^2 &=& a_{301}~\sigma^3 \rho^2 
	\nonumber \\
	a_4 g_{\rho}^2 \sigma^4 \rho^2 &=& a_{401}~\sigma^4 \rho^2 
	\nonumber \\
	a_5 g_{\rho}^2 \sigma^5 \rho^2 &=& a_{501}~\sigma^5 \rho^2 
	\nonumber \\
	a_6 g_{\rho}^2 \sigma^6 \rho^2 &=& a_{601}~\sigma^6 \rho^2 
	\nonumber
	\f}

	Note that Muller and Serot use the notation 
	\f[
	\frac{\bar{\kappa} g_s^3 }{2} = \frac{\kappa}{2} = b M 
	g_s^3 \qquad \mathrm{and} \qquad
	\frac{\bar{\lambda} g_s^4}{6} = \frac{\lambda}{6}
	= c g_s^4
	\f]
	which differs slightly from the "standard" notation above.

	We need to compare the values of
	\f{eqnarray*}
	&\frac{m_s^2}{2 g_s^2 M^2}, \frac{m_v^2}{2 g_v^2 M^2} 
	\frac{m_{\rho}^2}{8 g_{\rho}^2 M^2},\frac{b}{3},
	\frac{c}{4}
	&
	\nonumber \\
	&\frac{\zeta}{24}, \frac{\xi}{384},
	\frac{b_1}{4 g_{\omega}^2},
	\frac{b_2 M^2}{4 g_{\omega}^4},
	\frac{b_3 M^4}{4 g_{\omega}^6},
	\frac{a_1}{4 g_{\sigma} M},&
	\nonumber \\
	&\frac{a_2}{4 g_{\sigma}^2},
	\frac{a_3 M}{4 g_{\sigma}^3},
	\frac{a_4 M^2}{4 g_{\sigma}^4},
	\frac{a_5 M^3}{4 g_{\sigma}^5},~\mathrm{and}~\frac{a_6 M^4}
	{4 g_{\sigma}^6}\, .&
	\f}

	These values are stored in the variables cs, cw, cr, b, c,
	zeta, xi, b1, etc. in the specified \ref rmf_eos object. All
	of the numbers should be around 0.001 or 0.002.

	For the scale \f$ M \f$, \ref mnuc is used.

	\todo I may have ignored some signs in the above, which are
	unimportant for this application, but it would be good to fix
	them for posterity.

    */
    int check_naturalness(rmf_eos &re) {
      
      double gs=cs*ms;
      double gw=cw*mw;
      double gr=cr*mr;

      re.cs=0.5/cs/cs/mnuc/mnuc;
      re.cw=0.5/cw/cw/mnuc/mnuc;
      re.cr=0.125/cr/cr/mnuc/mnuc;
      re.b=b/3.0;
      re.c=c/4.0;

      re.zeta=zeta/24.0;
      re.xi=xi/384.0;

      re.b1=b1/gw/gw/4.0;
      re.b2=b2/pow(gw,4.0)/4.0*mnuc*mnuc;
      re.b3=b3/pow(gw,6.0)/4.0*pow(mnuc,4.0);

      re.a1=a1/gs/4.0/mnuc;
      re.a2=a2/pow(gs,2.0)/4.0;
      re.a3=a3/pow(gs,3.0)/4.0*mnuc;
      re.a4=a4/pow(gs,4.0)/4.0*mnuc*mnuc;
      re.a5=a5/pow(gs,5.0)/4.0*pow(mnuc,3.0);
      re.a6=a6/pow(gs,6.0)/4.0*pow(mnuc,4.0);

      return 0;
    }
    
    /** \brief Provide the maximum values of the couplings assuming
	a limit on naturalness

	The limits for the couplings are function of the nucleon and
	meson masses, except for the limits on \c b, \c c, \c zeta,
	and \c xi which are independent of the masses because of the
	way that these four couplings are defined.
    */
    int naturalness_limits(double value, rmf_eos &re) {
      
      double gs=cs*ms;
      double gw=cw*mw;
      double gr=cr*mr;

      re.cs=value*2.0*mnuc*mnuc;
      re.cw=value*2.0*mnuc*mnuc;
      re.cr=value*8.0*mnuc*mnuc;
      re.b=value*3.0;
      re.c=value*4.0;

      re.zeta=value*24.0;
      re.xi=value*384.0;
      
      re.b1=value*gw*gw*4.0;
      re.b2=value*pow(gw,4.0)*4.0/mnuc/mnuc;
      re.b3=value*pow(gw,6.0)*4.0/pow(mnuc,4.0);

      re.a1=value*gs*4.0*mnuc;
      re.a2=value*pow(gs,2.0)*4.0;
      re.a3=value*pow(gs,3.0)*4.0/mnuc;
      re.a4=value*pow(gs,4.0)*4.0/mnuc/mnuc;
      re.a5=value*pow(gs,5.0)*4.0/pow(mnuc,3.0);
      re.a6=value*pow(gs,6.0)*4.0/pow(mnuc,4.0);
      
      return 0;
    }

    /** \brief Set class mroot object for use calculating saturation density
     */
    virtual int set_sat_mroot(mroot<void *,mm_funct<void *> > &mrx) {
      sat_mroot=&mrx;
      return 0;
    }

    /** \brief The default solver for calculating the saturation 
	density
	
	Used by fn0() (which is called by saturation()) to solve
	saturation_matter_e() (1 variable).
    */
    gsl_mroot_hybrids<void *,mm_funct<void *> > def_sat_mroot;

#ifndef DOXYGEN_INTERNAL

  protected:
    
    /** 
	\brief Temporary charge density 

	\todo Should use hadronic_eos::proton_frac instead?
    */
    double n_charge;

    /// \name The meson fields
    //@{
    double sigma, omega, rho;
    //@}

    /// Temperature for solving field equations at finite temperature
    double fe_temp;

    /// For calc_e(), if true, then solve for neutron matter
    bool ce_neut_matter;
    
    /// For calc_e(), if true, then solve for proton matter
    bool ce_prot_matter;

    /// True if a guess for the fields has been given
    bool guess_set;

    /// The solver to compute saturation properties
    mroot<void *,mm_funct<void *> > *sat_mroot;

    /// The function for fix_saturation()
    int fix_saturation_fun(size_t nv, const ovector_view &x, ovector_view &y, 
		   void *&pa);
    
    /// Compute matter at zero pressure (for saturation())
    virtual int zero_pressure(size_t nv, const ovector_view &ex, 
			      ovector_view &ey, void *&pa);

    /// The function for calc_e()
    virtual int calc_e_solve_fun(size_t nv, const ovector_view &ex, 
				 ovector_view &ey, void *&pa);
    /// The function for calc_temp_e()
    virtual int calc_temp_e_solve_fun(size_t nv, const ovector_view &ex, 
				 ovector_view &ey, void *&pa);

    /** \brief Calculate the \c cr coupling given \c sig and \c ome 
	at the density 'nb'.
	
	Used by fix_saturation().
    */
    int calc_cr(double sig, double ome, double nb);

    double ce_temp;

#endif

  };

  template<> int io_tlate<rmf_eos>::input
    (cinput *co, in_file_format *ins, rmf_eos *sk);
  template<> int io_tlate<rmf_eos>::output
    (coutput *co, out_file_format *outs, rmf_eos *sk);
  template<> const char *io_tlate<rmf_eos>::type();

  typedef io_tlate<rmf_eos> rmf_eos_io_type;

#ifndef DOXYGENP
}
#endif

#endif
