/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/test_mgr.h>
#include <o2scl/nonrel_fermion.h>
#include <o2scl/apr_eos.h>
#include <o2scl/gsl_mroot_hybrids.h>
#include <o2scl/cern_mroot.h>
#include <o2scl/tov_eos.h>
#include <o2scl/tov_solve.h>

using namespace std;
using namespace o2scl;
using namespace o2scl_const;

nonrel_fermion neut(939.0/hc_mev_fm,2.0),p(939.0/hc_mev_fm,2.0);
fermion e(0.511/hc_mev_fm,2.0),zero(0.0,0.0);
thermo ht, hb, l;
apr_eos ap;
double barn;
int nstarfun(size_t nv, const ovector_view &x, ovector_view &y, void *&vp);

const double xmev_kg=1.782661731e-30; 
const double mev_per_fm3_msun_km3=xmev_kg/gsl_mks::solar_mass*1.0e54; 

int main(void) {
  test_mgr t;
  t.set_output_level(2);
  gsl_mroot_hybrids<void *,mm_funct<void *> > nd;
  cern_mroot<void *,mm_funct<void *> > nd2;
  ovector x(4);
  void *vp=0;
  table eos(200);
  collection co;

  cout.setf(ios::scientific);

  // --------------------------------------------------------------
  // Calculate APR EOS, with naive phase transition and no muons

  ap.select(1);
  ap.pion=0;

  x[0]=0.09663;
  x[1]=0.003365;
  x[2]=0.4636;

  eos.line_of_names("ed pr nb");
  mm_funct_fptr<void *> nf(nstarfun);
  for(barn=0.1;barn<=2.0001;barn+=0.01) {
    int ret=nd.msolve(3,x,vp,nf);
    // If the GSL solver fails, use CERNLIB
    if (ret!=0) {
      int ret2=nd2.msolve(3,x,vp,nf);
      if (ret2!=0) {
	cout << "Solver failed." << endl;
      }
    }
    double line[3]={ht.ed*mev_per_fm3_msun_km3*hc_mev_fm,
		    ht.pr*mev_per_fm3_msun_km3*hc_mev_fm,
		    barn};
    eos.line_of_data(3,line);
  }
  
  tov_interp_eos te;
  te.verbose=3;
  te.set_low_density_eos(true,"");
  te.read_table(eos);
  
  t.test_gen(te.baryon_column==true,"baryon column");
  vector<string> auxp;
  size_t np;
  te.get_names(np,auxp);
  t.test_gen(np==0,"no aux parms");

  // --------------------------------------------------------------
  // Test maximum mass, radius, and radius of 1.4 solar mass star.

  tov_solve at;
  at.calcgpot=true;
  at.verbose=0;
  at.set_eos(te);
  at.fixed(1.4);
  double rad14=at.rad;
  t.test_rel(at.mass,1.4,1.0e-6,"APR Maximum mass.");
  t.test_rel(at.rad,11.4,0.03,"APR Radius of 1.4 solar mass star.");
  t.test_rel(at.bmass,1.56,0.01,"APR baryonic mass of 1.4 solar mass star.");
  t.test_rel(at.gpot,-0.22,0.01,"APR grav. pot. of 1.4 solar mass star.");
  cout << at.gpot << endl;
  at.max();
  double massmax=at.mass;
  double radmax=at.rad;
  t.test_rel(at.mass,2.20,0.03,"APR Maximum mass.");
  t.test_rel(at.rad,10.0,0.02,"APR Radius of maximum mass star.");
  t.test_rel(at.bmass,2.68,0.01,"APR baryonic mass of 1.4 solar mass star.");
  t.test_rel(at.gpot,-0.52,0.01,"APR grav. pot. of 1.4 solar mass star.");
  cout << at.gpot << endl;
  at.mvsr();
  table tab=at.get_results();
  tab.summary(&cout);

  // --------------------------------------------------------------
  // Test the Buchdahl EOS 

  tov_solve bts;
  bts.verbose=0;
  tov_buchdahl_eos buch;
  bts.set_eos(buch);
  bts.fixed(1.4);
  double beta=gsl_mks::schwarzchild_radius/2.0e3*bts.mass/bts.rad;
  table &br=bts.get_results();
  t.test_rel(36.0*buch.Pstar*beta*beta,br.get("pr",0),1.0e-5,"Buch Pc");
  t.test_rel(72.0*buch.Pstar*beta*(1.0-2.5*beta),
	     br.get("ed",0),5.0e-6,"Buch rho_c");
  
  t.report();

  return 0;
}

int nstarfun(size_t nv, const ovector_view &x, ovector_view &y, void *&vp) {
  neut.n=x[0];
  p.n=x[1];
  e.mu=x[2];

  if (x[0]<0.0 || x[1]<0.0) return 1;

  ap.calc_e(neut,p,hb);
  e.calc_mu_zerot();

  l.ed=e.ed;
  l.pr=e.pr;
  ht=l+hb;

  y[0]=neut.n+p.n-barn;
  y[1]=p.n-e.n;
  y[2]=neut.mu-p.mu-e.mu;

  return 0;
}


