/*
  -------------------------------------------------------------------

  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner

  This file is part of O2scl.

  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_FIT_FIX_H
#define O2SCL_FIT_FIX_H

#include <o2scl/fit_base.h>
#include <o2scl/gsl_fit.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief Multidimensional fitting fixing some parameters and 
      varying others
  */
  template<class param_t, class bool_vec_t>
    class fit_fix_pars : public fit_base<param_t,fit_funct<param_t> > {
    
  public:
    
    /** \brief Specify the member function pointer
     */
    fit_fix_pars() {
      fitp=&def_fit;
    }
    
    virtual ~fit_fix_pars() {}
    
    /** \brief Fit the data specified in (xdat,ydat) to
        the function \c fitfun with the parameters in \c par.

        The covariance matrix for the parameters is returned in \c covar
        and the value of \f$ \chi^2 \f$ is returned in \c chi2.

    */
    virtual int fit(size_t ndat, ovector_view &xdat, ovector_view &ydat, 
		    ovector_view &yerr, size_t npar, ovector_view &par, 
		    omatrix_view &covar, double &chi2, param_t &pa, 
		    fit_funct<param_t> &fitfun) {
      
      xp=&par;
      funcp=&fitfun;
      unv=npar;
      nv_new=npar;
      ovector xnew(nv_new);
      for(size_t i=0;i<npar;i++) {
	xnew[i]=par[i];
      }
      fit_funct_mfptr<fit_fix_pars,param_t> 
	mfm(this,&fit_fix_pars::fit_func);
      fitp->fit(ndat,xdat,ydat,yerr,nv_new,xnew,covar,chi2,pa,mfm);
      for(size_t i=0;i<npar;i++) {
	par[i]=xnew[i];
      }
      return 0;
    }

    /** \brief Fit function \c func while fixing some parameters as
	specified in \c fix.
    */
    virtual int fit_fix(size_t ndat, ovector_view &xdat, ovector_view &ydat, 
			ovector_view &yerr, size_t npar, ovector_view &par, 
			bool_vec_t &fix, omatrix_view &covar, double &chi2, 
			param_t &pa, fit_funct<param_t> &fitfun) {
      xp=&par;
      funcp=&fitfun;
      unv=npar;
      fixp=&fix;
      nv_new=0;
      for(size_t i=0;i<npar;i++) {
	if (fix[i]==false) nv_new++;
      }
      if (nv_new==0) return 0;
      size_t j=0;
      ovector xnew(nv_new);
      for(size_t i=0;i<npar;i++) {
	if (fix[i]==false) {
	  xnew[j]=par[i];
	  j++;
	}
      }
      fit_funct_mfptr<fit_fix_pars,param_t> 
	mfm(this,&fit_fix_pars::fit_func);
      fitp->fit(ndat,xdat,ydat,yerr,nv_new,xnew,covar,chi2,pa,mfm);
      j=0;
      for(size_t i=0;i<npar;i++) {
	if (fix[i]==false) {
	  par[i]=xnew[j];
	  j++;
	}
      }
      return 0;
    }

    /// Change the base minimizer
    int set_fit(fit_base<param_t,fit_funct_mfptr
		<fit_fix_pars,param_t> > &fitter) {
      fitp=&fitter;
      return 0;
    }

    /// The default base minimizer
    gsl_fit<param_t,fit_funct_mfptr<fit_fix_pars,param_t> > def_fit;
    
#ifndef DOXYGEN_INTERNAL
    
  protected:
    
    /** \brief The new function to send to the minimizer
     */
    virtual int fit_func(size_t nv, ovector_view &x, double xx,
			 double &y, param_t &pa) {
      ovector tmp(unv);
      size_t j=0;
      for(size_t i=0;i<unv;i++) {
	if (nv_new<unv && (*fixp)[i]==true) {
	  tmp[i]=(*xp)[i];
	} else {
	  tmp[i]=x[j];
	  j++;
	}
      }
      int ret=(*funcp)(unv,tmp,xx,y,pa);
      return ret;
    }
    
    /// The minimizer
    fit_base<param_t,fit_funct_mfptr<fit_fix_pars,param_t> > *fitp;

    /// The user-specified function
    fit_funct<param_t> *funcp;

    /// The user-specified number of variables
    size_t unv;

    /// The new number of variables
    size_t nv_new;
    
    /// Specify which parameters to fix
    bool_vec_t *fixp;
    
    /// The user-specified initial vector
    ovector_view *xp;

#ifndef DOXYGENP
#endif

  private:
 
    fit_fix_pars(const fit_fix_pars &);
    fit_fix_pars& operator=(const fit_fix_pars&);

#endif

  };


#ifndef DOXYGENP
}
#endif

#endif
