/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_GSL_FIT_H
#define O2SCL_GSL_FIT_H

#include <o2scl/fit_base.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_multifit_nlin.h>
#include <o2scl/omatrix_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Non-linear least-squares fitting class (GSL)
    
      The GSL-based fitting class using a Levenberg-Marquardt type
      algorithm. The algorithm stops when
      \f[
      |dx_i| < \mathrm{epsabs}+\mathrm{epsrel}\times|x_i|
      \f]
      where \f$dx\f$ is the last step and \f$x\f$ is the current
      position. If test_gradient is true, then additionally fit()
      requires that
      \f[
      \sum_i |g_i| < \mathrm{epsabs}
      \f]
      where \f$g_i\f$ is the \f$i\f$-th component of the gradient of
      the function \f$\Phi(x)\f$ where
      \f[
      \Phi(x) = || F(x) ||^2
      \f]
      
      \todo Properly generalize other vector types than 
      ovector_view
      \todo Allow the user to specify the derivatives
      \todo Fix so that the user can specify automatic
      scaling of the fitting parameters, where the initial
      guess are used for scaling so that the fitting parameters
      are near unity.
  */
  template<class param_t, class func_t, class vec_t=ovector_view, 
    class mat_t=omatrix_view, class bool_vec_t=bool *> class gsl_fit : 
  public fit_base<param_t,func_t,vec_t,mat_t> {
    
    public:
    
    gsl_fit() {
      gsl_set_error_handler(err_hnd->gsl_hnd);
      max_iter=500;
      epsabs=1.0e-4;
      epsrel=1.0e-4;
      use_scaled=true;
      test_gradient=false;
    }

    virtual ~gsl_fit() {}
  
    /** \brief Fit the data specified in (xdat,ydat) to
	the function \c fitfun with the parameters in \c par.
	
	The covariance matrix for the parameters is returned in \c covar
	and the value of \f$ \chi^2 \f$ is returned in \c chi2.
    */
    virtual int fit(size_t ndat, vec_t &xdat, vec_t &ydat, vec_t &yerr,
		    size_t npar, vec_t &par, mat_t &covar, double &chi2,
		    param_t &pa, func_t &fitfun) {

      int status, iter=0;
      size_t i, j;

      gsl_vector *x=gsl_vector_alloc(npar);
      for(i=0;i<npar;i++) gsl_vector_set(x,i,par[i]);

      // Create fitting function and parameters
      gsl_multifit_function_fdf f;
      gsl_vector *grad=gsl_vector_alloc(npar);
      func_par fpar={fitfun,&pa,ndat,&xdat,&ydat,&yerr,npar};
      f.f=func;
      f.df=dfunc;
      f.fdf=fdfunc;
      f.n=ndat;
      f.p=npar;
      f.params=&fpar;
     
      // Create fitting object
      gsl_multifit_fdfsolver *s;
      if (use_scaled) {
	s=gsl_multifit_fdfsolver_alloc
	  (gsl_multifit_fdfsolver_lmsder,ndat,npar);
      } else {
	s=gsl_multifit_fdfsolver_alloc
	(gsl_multifit_fdfsolver_lmder,ndat,npar);
      }
      gsl_multifit_fdfsolver_set(s,&f,x);
  
      // Perform the fit
      do {
	iter++;
	status=gsl_multifit_fdfsolver_iterate(s);
    
	if (status) {
	  break;
	}
    
	status=gsl_multifit_test_delta(s->dx,s->x,epsabs,epsrel);
	if (test_gradient && status==gsl_continue) {
	  gsl_multifit_gradient(s->J,s->x,grad);
	  status=gsl_multifit_test_gradient(grad,epsabs);
	}

	print_iter(npar,s->dx,s->x,iter,epsabs,epsrel);
    
      } while (status == gsl_continue && iter < max_iter);
  
      // Create covariance matrix and copy results to the 
      // user-specified vector
      gsl_matrix *gcovar=gsl_matrix_alloc(npar,npar);
      gsl_multifit_covar(s->J,0.0,gcovar);
      for(i=0;i<npar;i++) {
	par[i]=gsl_vector_get(s->x,i);
	for(j=0;j<npar;j++) {
	  covar.set(i,j,gsl_matrix_get(gcovar,i,j));
	}
      }
      gsl_matrix_free(gcovar);
  
      // Compute chi squared
      chi2=gsl_blas_dnrm2(s->f);
      chi2*=chi2;

      // Free memory
      gsl_multifit_fdfsolver_free(s);
      gsl_vector_free(grad);
  
      return 0;
    }
    
    /// (default 500)
    int max_iter;

    /// (default 1.0e-4)
    double epsabs;

    /// (default 1.0e-4)
    double epsrel;

    /// If true, test the gradient also (default false)
    bool test_gradient;

    /** \brief Use the scaled routine if true (default true)
     */
    bool use_scaled;

    /// Return string denoting type ("gsl_fit")
    virtual const char *type() { return "gsl_fit"; }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// Print the progress in the current iteration
    virtual int print_iter(int nv, gsl_vector *x, gsl_vector *dx, int iter,
			   double l_epsabs, double l_epsrel) {
      if (this->verbose<=0) return 0;
  
      int i;
      char ch;
      double dxmax, epsmin;

      std::cout << "Iteration: " << iter <<  std::endl;
      text_out_file outs(&std::cout,79);
      outs.word_out("x:");
      dxmax=gsl_vector_get(dx,0);
      epsmin=l_epsabs+l_epsrel*gsl_vector_get(x,0);
      for(i=0;i<nv;i++) {
	if (dxmax<gsl_vector_get(dx,i)) dxmax=gsl_vector_get(dx,i);
	if (epsmin>l_epsabs+l_epsrel*gsl_vector_get(x,i)) {
	  epsmin=l_epsabs+l_epsrel*gsl_vector_get(x,i);
	}
	outs.double_out(gsl_vector_get(x,i));
      }
      outs.end_line();
      std::cout << " Val: " << dxmax << " Lim: " << epsmin << std::endl;
      if (this->verbose>1) {
	std::cout << "Press a key and type enter to continue. ";
	std::cin >> ch;
      }

      return 0;
    }

    /** \brief A structure for passing to the functions func(), dfunc(),
	and fdfunc()
    */
    typedef struct {
      /// The function object 
      func_t &f;
      /// The user-specified parameter
      param_t *vp;
      /// The number 
      int ndat;
      /// The x values 
      vec_t *xdat;
      /// The y values 
      vec_t *ydat;
      /// The y uncertainties
      vec_t *yerr;
      /// The number of parameters
      int npar;
    } func_par;

    /// Evaluate the function
    static int func(const gsl_vector *x, void *pa, gsl_vector *f) {
      func_par *fp=(func_par *)pa;
      int i, j;
      double yi;
      ovector xp(fp->npar);
      for(j=0;j<fp->npar;j++) xp[j]=gsl_vector_get(x,j);
      for(i=0;i<fp->ndat;i++) {
	fp->f(fp->npar,xp,(*fp->xdat)[i],yi,*(fp->vp));
	gsl_vector_set(f,i,(yi-(*fp->ydat)[i])/(*fp->yerr)[i]);
      }

      return 0;
    }

    /// Evaluate the jacobian
    static int dfunc(const gsl_vector *x, void *pa, gsl_matrix *jac) {
      func_par *fp=(func_par *)pa;
      int i, j;
      double yi, ylo, yhi, xtemp, eps=1.0e-4;
      ovector xp(fp->npar);
  
      for(j=0;j<fp->npar;j++) xp[j]=gsl_vector_get(x,j);
      for(j=0;j<fp->npar;j++) {
	for(i=0;i<fp->ndat;i++) {
	  xtemp=xp[j];
	  xp[j]+=eps;
	  fp->f(fp->npar,xp,(*fp->xdat)[i],yi,*(fp->vp));
	  yhi=(yi-(*fp->ydat)[i])/(*fp->yerr)[i];
	  xp[j]=xtemp;
	  fp->f(fp->npar,xp,(*fp->xdat)[i],yi,*(fp->vp));
	  ylo=(yi-(*fp->ydat)[i])/(*fp->yerr)[i];
	  gsl_matrix_set(jac,i,j,(yhi-ylo)/eps);
	}
      }

      return 0;
    }

    /// Evaluate the function and the jacobian
    static int fdfunc(const gsl_vector *x, void *pa, gsl_vector *f,
		      gsl_matrix *jac) {
      func(x,pa,f);
      dfunc(x,pa,jac);
      return 0;
    }

#endif
  
  };

#ifndef DOXYGENP
}
#endif

#endif
