/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_NAIVE_METROPOLIS_H
#define O2SCL_NAIVE_METROPOLIS_H

#include <iostream>
#include <o2scl/collection.h>
#include <o2scl/funct.h>
#include <o2scl/mcarlo_inte.h>
#include <o2scl/gsl_rnga.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Naive metropolis monte carlo integration

      Very experimental.
      
      This returns the ratio of n-dimensional integrals
      \f[
      I = \frac{\int f(\vec{x}) \exp\left[-\beta~w(\vec{x})\right] d \vec{x}}
      {\int \exp\left[-\beta~w(\vec{x})\right] d \vec{x}}
      \f]
      for function  \f$ f \f$  specified in \c func and  \f$ w \f$ 
      specified in \c weight. 

      A "Metropolis step" (or simply a "step") is defined by 
      the following procedure: The step \f$ \vec{x}_0 \rightarrow 
      \vec{x}_1 \f$ is accepted if
      \f[
      w(\vec{x}_1)<w_(\vec{x}_0)
      \f]
      or if 
      \f[
      e^{-\beta \left[w(\vec{x}_1)-w_(\vec{x}_0)\right]}>r
      \f]
      where  \f$ r \f$  is a uniform random number  \f$ [0,1) \f$  newly
      generated for each step.

      \ref nstart steps are taken initially to ensure the first point
      is created with the correct probability. Afterwards, the
      average value of the function is stored over \ref niter steps.
      This average value is computed \ref nblock times to get
      an overall average and an estimate of the uncertainity.

      \todo Talk about how accurate this is with \f$ \beta \f$ .
      \todo Redo statistics, talk about how many times the integral
      is evaluated. 
      \todo Offer an option of giving more results 
      than just the final averange and error?
      
      \b Minimization

      There is a connection between integration over functions of this
      type and minimization. The expression
      \f[
      \lim_{T \rightarrow 0} 
      \frac{\int_{x=a}^{x=b} f(x) e^{-f(x)/T}}
      {\int_{x=a}^{x=b} e^{-f(x)/T}}
      \f]
      gives the minimum of the function  \f$ f(x) \f$  in the region
       \f$ x \in (a,b) \f$ .

  */
  template<class param_t, class func_t, class rng_t,
    class vec_t, class alloc_vec_t> class naive_metropolis :
    public mcarlo_inte<param_t,func_t,rng_t,vec_t>

    {
    public:

      rng_t rng;
      
      naive_metropolis() {
	nblock=10;
	niter=10000;
	nstart=1000;
      }
      
      virtual ~naive_metropolis() {}
    
      /** 
	  \brief Calculate the integral returning result in \c value

	  Calculates the integral given the functions \c func and \c weight
	  over the region specified by \c a and \c b, both of side \c ndim.
	  The value of  \f$ \beta \f$  is specified in \c beta, and the
	  result is returned in \c value, with the uncertainty given
	  in \c err.
      */
      virtual int minteg_err(func_t &func, func_t &energy, size_t ndim, 
			     const vec_t &a, const vec_t &b, 
			     const double beta, double &value, double &err, 
			     param_t &pa) 
	{
	  double value2=0.0, bvalue;
	  alloc_vec_t x, xnew;
	  
	  value=0.0;
	  
	  // Choose a random starting point
	  for(size_t i=0;i<ndim;i++) {
	    x[i]=rng.random()*(b[i]-a[i])+a[i];
	  }

	  // Take several steps to ensure independence
	  // from the starting point
	  for(size_t n=0;n<nstart;n++) {

	    // Pick a new point
	    for(size_t i=0;i<ndim;i++) {
	      xnew[i]=rng.random()*(b[i]-a[i])+a[i];
	    }

	    // Take a step
	    double dw=energy(ndim,xnew,pa)-energy(ndim,x,pa);
	    if (dw<0 || exp(-beta*dw)>rng.random()) {
	      for(size_t i=0;i<ndim;i++) {
		x[i]=xnew[i];
	      }
	    }

	  }

	  for(unsigned long int ib=0;ib<nblock;ib++) {

	    bvalue=0.0;

	    // Integrate the function
	    for(size_t n=0;n<niter;n++) {
	      double t=func(ndim,x,pa);
	      bvalue+=t;
      
	      // Pick a new point
	      for(size_t i=0;i<ndim;i++) {
		xnew[i]=rng.random()*(b[i]-a[i])+a[i];
	      }
      
	      // Decide to accept or reject
	      double dw=energy(ndim,xnew,pa)-energy(ndim,x,pa);
	      if (dw<0 || exp(-beta*dw)>rng.random()) {
		for(size_t i=0;i<ndim;i++) {
		  x[i]=xnew[i];
		}
	      }
	    }
    
	    bvalue/=((double)niter);
	    value+=bvalue;
	    value2+=bvalue*bvalue;

	  }

	  value/=((double)nblock);
	  value2/=((double)nblock);
	  err=sqrt(fabs(value2-value*value));

	  return 0;
	}

      /** 
	  \brief Calculate the integral returning result in \c value

	  Calculates the integral given the functions \c func and \c weight
	  over the region specified by \c a and \c b, both of side \c ndim.
	  The value of  \f$ \beta \f$  is specified in \c beta, and the
	  result is returned in \c value, with the uncertainty given
	  in \c err.
      */
      virtual int minteg_array(func_t &func, func_t &energy, size_t ndim, 
			       const vec_t &a, const vec_t &b, 
			       const double beta, vec_t &results, param_t &pa) 
	{
	  double value2=0.0, bvalue;
	  alloc_vec_t x, xnew;

	  unsigned long int ix=0;

	  // Choose a random starting point
	  for(size_t i=0;i<ndim;i++) {
	    x[i]=rng.random()*(b[i]-a[i])+a[i];
	  }

	  // Take several steps to ensure independence
	  // from the starting point
	  for(size_t n=0;n<nstart;n++) {

	    // Pick a new point
	    for(size_t i=0;i<ndim;i++) {
	      xnew[i]=rng.random()*(b[i]-a[i])+a[i];
	    }

	    // Take a step
	    double dw=energy(ndim,xnew,pa)-energy(ndim,x,pa);
	    if (dw<0 || exp(-beta*dw)>rng.random()) {
	      for(size_t i=0;i<ndim;i++) {
		x[i]=xnew[i];
	      }
	    }

	  }

	  for(unsigned long int ib=0;ib<nblock;ib++) {

	    bvalue=0.0;

	    // Integrate the function
	    for(size_t n=0;n<niter;n++) {
	      double t=func(ndim,x,pa);
	      bvalue+=t;
      
	      // Pick a new point
	      for(size_t i=0;i<ndim;i++) {
		xnew[i]=rng.random()*(b[i]-a[i])+a[i];
	      }
      
	      // Decide to accept or reject
	      double dw=energy(ndim,xnew,pa)-energy(ndim,x,pa);
	      if (dw<0 || exp(-beta*dw)>rng.random()) {
		for(size_t i=0;i<ndim;i++) {
		  x[i]=xnew[i];
		}
	      }
	    }

	    results[ix]=bvalue;
	    ix++;

	  }
	  
	  return 0;
	}
    
      /// The number of iterations (default 10000)
      unsigned long int niter;

      /// The number of warm-up iterations (default 1000)
      unsigned long int nstart;

      /// Number of blocks (default 10)
      unsigned long int nblock;
    };

#ifndef DOXYGENP
}
#endif

#endif






