/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_MULTI_MIN_H
#define O2SCL_MULTI_MIN_H

#include <o2scl/text_file.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/multi_funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /// Base class for a gradient function
  template<class param_t, class vec_t=ovector_view>
    class grad_funct {

    public:

    virtual ~grad_funct() {}

    /// Compute the gradient \c g at the point \c x
    virtual int operator()(size_t nv, vec_t &x, vec_t &g, param_t &pa) {
      return 0;
    }

  };

  /** \brief Function pointer to a gradient
   */
  template <class param_t, class vec_t=ovector_view>
    class grad_funct_fptr : public grad_funct<param_t,vec_t> {
    
    public:

    /** \brief Specify the function pointer
     */
    grad_funct_fptr(int (*fp)(size_t nv, vec_t &x, vec_t &g, param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~grad_funct_fptr() {};

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nv, vec_t &x, vec_t &g, param_t &pa) {
      return (*fptr)(nv,x,g,pa);
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// The function pointer
    int (*fptr)(size_t nv, vec_t &x, vec_t &g, param_t &pa);

#endif

#ifndef DOXYGENP

    private:

    grad_funct_fptr(const grad_funct_fptr &);
    grad_funct_fptr& operator=(const grad_funct_fptr&);

#endif

  };

  /** \brief Member function pointer to a gradient
   */
  template <class tclass, class param_t, class vec_t=ovector_view>
    class grad_funct_mfptr : public grad_funct<param_t,vec_t> {
    
    public:

    /** \brief Specify the member function pointer
     */
    grad_funct_mfptr(tclass *tp, int (tclass::*fp)
		     (size_t nv, vec_t &x, vec_t &g, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~grad_funct_mfptr() {};

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nv, vec_t &x, vec_t &g, param_t &pa) {
      return (*tptr.*fptr)(nv,x,g,pa);
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// Member function pointer
    int (tclass::*fptr)(size_t nv, vec_t &x, vec_t &g, param_t &pa);
    
    /// Class pointer
    tclass *tptr;

#endif

#ifndef DOXYGENP

    private:

    grad_funct_mfptr(const grad_funct_mfptr &);
    grad_funct_mfptr& operator=(const grad_funct_mfptr&);

#endif

  };

  /// Base class for automatically computing gradients
  template<class param_t, class func_t, class vec_t=ovector_view> 
    class gradient : public grad_funct<param_t,vec_t> {

    public:

    virtual ~gradient() {}

    /// Set the function to compute the gradient of
    virtual int set_function(func_t &f) {
      func=&f;
      return 0;
    }

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nv, vec_t &x, vec_t &g, param_t &pa) {
      return 0;
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// A pointer to the user-specified function
    func_t *func;

#endif

  };

  /// Simple automatic computation of gradient by finite differencing
  template<class param_t, class func_t, class vec_t> class simple_grad :
  public gradient<param_t,func_t,vec_t> {

  public:

    simple_grad() {
      epsrel=1.0e-4;
      epsmin=1.0e-15;
    }
    
    virtual ~simple_grad() {}

    /** \brief The relative stepsize for finite-differencing
        (default \f$ 10^{-4} \f$ )
    */
    double epsrel;

    /// The minimum stepsize (default \f$ 10^{-15} \f$)
    double epsmin;

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nv, vec_t &x, vec_t &g, param_t &pa) {
      double fv1, fv2, h;

      (*this->func)(nv,x,fv1,pa);
      
      for(size_t i=0;i<nv;i++) {
	
	h=epsrel*fabs(x[i]);
	if (fabs(h)<=epsmin) h=epsrel;

	x[i]+=h;
	(*this->func)(nv,x,fv2,pa);
	x[i]-=h;
	g[i]=(fv2-fv1)/h;
      }
      
      return 0;
    }

  };
    
  /// Base class for a gradient function using arrays
  template<class param_t, size_t nv> class grad_vfunct {

    public:

    virtual ~grad_vfunct() {}

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nvar, double x[nv], double g[nv], 
			   param_t &pa) {
      return 0;
    }

  };

  /** \brief Function pointer to a gradient
   */
  template <class param_t, size_t nv>
    class grad_vfunct_fptr : public grad_vfunct<param_t,nv> {
    
    public:

    /** \brief Specify the member function pointer
     */
    grad_vfunct_fptr(int (*fp)(size_t nv, double x[nv], double g[nv], 
			       param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~grad_vfunct_fptr() {};

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nvar, double x[nv], double g[nv], 
			   param_t &pa) {
      return (*fptr)(nvar,x,g,pa);
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// Function pointer
    int (*fptr)(size_t nvar, double x[nv], double g[nv], param_t &pa);

#endif

#ifndef DOXYGENP

    private:

    grad_vfunct_fptr(const grad_vfunct_fptr &);
    grad_vfunct_fptr& operator=(const grad_vfunct_fptr&);

#endif

  };

  /** \brief Member function pointer to a gradient
   */
  template <class tclass, class param_t, size_t nv>
    class grad_vfunct_mfptr : public grad_vfunct<param_t,nv> {
    
    public:

    /** \brief Specify the member function pointer
     */
    grad_vfunct_mfptr(tclass *tp, int (tclass::*fp)
		      (size_t nvar, double x[nv], double g[nv], param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
    
    virtual ~grad_vfunct_mfptr() {};

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nvar, double x[nv], double g[nv], 
			   param_t &pa) {
      return (*tptr.*fptr)(nvar,x,g,pa);
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// Member function pointer
    int (tclass::*fptr)(size_t nvar, double x[nv], double g[nv], param_t &pa);
    
    /// Class pointer
    tclass *tptr;

#endif

#ifndef DOXYGENP

    private:

    grad_vfunct_mfptr(const grad_vfunct_mfptr &);
    grad_vfunct_mfptr& operator=(const grad_vfunct_mfptr&);

#endif

  };

  /// Base class for automatically computing gradients with arrays
  template<class param_t, class func_t, size_t nv> 
    class gradient_array : public grad_vfunct<param_t,nv> {

    public:

    virtual ~gradient_array() {}

    /// Set the function to compute the gradient of
    virtual int set_function(func_t &f) {
      func=&f;
      return 0;
    }

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nvar, double x[nv], double g[nv], 
			   param_t &pa) {
      return 0;
    }

#ifndef DOXYGEN_INTERNAL

    protected:

    /// A pointer to the user-specified function
    func_t *func;

#endif

  };

  /** \brief Simple automatic computation of gradient by finite 
      differencing with arrays
  */
  template<class param_t, class func_t, size_t nv> class simple_grad_array :
  public gradient_array<param_t,func_t,nv> {

  public:

    simple_grad_array() {
      epsrel=1.0e-4;
      epsmin=1.0e-15;
    }
    
    virtual ~simple_grad_array() {}

    /** \brief The relative stepsize for finite-differencing
        (default \f$ 10^{-4} \f$ )
    */
    double epsrel;

    /// The minimum stepsize (default \f$ 10^{-15} \f$)
    double epsmin;

    /** \brief Compute the gradient \c g at the point \c x
     */
    virtual int operator()(size_t nvar, double x[nv], double g[nv], 
			   param_t &pa) {
      double fv1, fv2, h;

      (*this->func)(nv,x,fv1,pa);
      
      for(size_t i=0;i<nvar;i++) {
	
	h=epsrel*fabs(x[i]);
	if (fabs(h)<=epsmin) h=epsrel;

	x[i]+=h;
	(*this->func)(nvar,x,fv2,pa);
	x[i]-=h;
	g[i]=(fv2-fv1)/h;
      }
      
      return 0;
    }

  };
    
  /** 
      \brief Multidimensional minimization base

      <b>The template parameters:</b>
      The template parameter \c func_t specifies the function to 
      minimize and should be a class containing a definition 
      \code
      func_t::operator()(size_t nv, const vec_t &x, double &f, param_t &pa);
      \endcode
      where \c f is the value of the function at \c x with parameter \c pa
      where \c x is a array-like class defining \c operator[] of size \c nv.
      The parameter \c dfunc_t (if used) should provide the gradient with
      \code
      func_t::operator()(size_t nv, const vec_t &x, vec_t &g, param_t &pa);
      \endcode
      where \c g is the gradient of the function at \c x. 
  */
  template<class param_t, class func_t, class dfunc_t=func_t, 
    class vec_t=ovector_view> class multi_min {
      
      public:
      
      /// Output control
      int verbose;
      
      /// Maximum number of iterations
      int ntrial;
      
      /// Tolerance
      double tolf;
      
      /// The minimum allowable stepsize
      double tolx;

      /// The number of iterations for in the most recent minimization
      int last_ntrial;
      
      multi_min() {
	verbose=0;
	ntrial=100;
	tolf=1.0e-4;
	tolx=1.0e-4;
	last_ntrial=0;
      }

      virtual ~multi_min() {}
      
      /** \brief Calculate the minimum \c min of \c func w.r.t. the
	  array \c x of size \c nvar.
      */
      virtual int mmin(size_t nvar, vec_t &x, double &fmin, param_t &pa,
		       func_t &func) 
      {
	return 0;
      }
      
      /** \brief Calculate the minimum \c min of \c func
	  w.r.t. the array \c x of size \c nvar with gradient
	  \c dfunc
      */
      virtual int mmin_de(size_t nvar, vec_t &x, double &fmin, param_t &pa,
			  func_t &func, dfunc_t &dfunc)
      {
	return mmin(nvar,x,fmin,pa,func);
      }
      
      /** 
	  \brief Print out iteration information.
	  
	  Depending on the value of the variable verbose, this prints out
	  the iteration information. If verbose=0, then no information is
	  printed, while if verbose>1, then after each iteration, the
	  present values of x and y are output to std::cout along with the
	  iteration number. If verbose>=2 then each iteration waits for a
	  character.
      */
      template<class vec2_t> 
      int print_iter(size_t nv, vec2_t &x, double y, int iter,
		     double value, double limit,
		     std::string comment) 
      {
	if (verbose<=0) return 0;
	
	int i;
	char ch;
	
	std::cout << comment << " Iteration: " << iter << std::endl;
	text_out_file outs(&std::cout,79);
	outs.word_out("x:");
	for(i=0;i<((int)nv);i++) outs.double_out(x[i]);
	outs.end_line();
	std::cout << "y: " << y << " Val: " << value << " Lim: " 
	<< limit << std::endl;
	if (verbose>1) {
	  std::cout << "Press a key and type enter to continue. ";
	  std::cin >> ch;
	}
	
	return 0;
      }
      
      /// Return string denoting type ("multi_min")
      const char *type() { return "multi_min"; }
      
  };
  
#ifndef DOXYGENP
}
#endif

#endif

