/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  Some of the source code in this file has been adapted from OOL
  library written by Sergio Drumond Ventura, Luis Alberto D'Afonseca,
  and Ricardo Biloti available from http://ool.sourceforge.net.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_OOL_MMIN_PGRAD_H
#define O2SCL_OOL_MMIN_PGRAD_H

#include <gsl/gsl_math.h>

#include <o2scl/text_file.h>
#include <o2scl/ovector_tlate.h>
#include <o2scl/multi_funct.h>
#include <o2scl/ool_constr_mmin.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Constrained minimization by the projected gradient method (OOL)

      \todo Complete the mmin() interface with automatic gradient
  */
  template<class param_t, class func_t, class dfunc_t, 
    class vec_t=ovector_view, class alloc_vec_t=ovector,
    class alloc_t=ovector_alloc> class ool_mmin_pgrad : 
  public ool_constr_mmin<param_t,func_t,dfunc_t,ool_hfunct<void *>,vec_t,
    alloc_vec_t,alloc_t> {
    
#ifndef DOXYGEN_INTERNAL

    protected:

    /// A convenient typedef for the unused Hessian product type
    typedef ool_hfunct<void *> hfunc_t;

    /// Temporary vector
    alloc_vec_t xx;
    
    /// Project into feasible region
    int proj(vec_t &xt) {
      size_t ii, n=this->dim;
      
      for(ii=0;ii<n;ii++) {
	xt[ii]=GSL_MAX(this->L[ii],GSL_MIN(xt[ii],this->U[ii]));
      }
      return 0;
    }

    /// Line search
    int line_search() {

      double t, tnew, fx, fxx, dif2, gtd;

      fx=this->f;

      tnew=1;
      
      do {
	
	t = tnew;
	/* xx = x - t df */
	o2scl_arith::vector_copy(this->dim,this->x,xx);
	for(size_t i=0;i<this->dim;i++) xx[i]-=t*this->gradient[i];
	proj(xx);
	
	(*this->func)(this->dim,xx,fxx,*this->param);
	
	o2scl_arith::vector_copy(this->dim,xx,this->dx);
	for(size_t i=0;i<this->dim;i++) this->dx[i]-=this->x[i];

	dif2=0.0;
	for(size_t i=0;i<this->dim;i++) dif2+=this->dx[i]*this->dx[i];
	
	gtd=0.0;
	for(size_t i=0;i<this->dim;i++) {
	  gtd+=this->gradient[i]*this->dx[i];
	}
	
	tnew = - t * t * gtd / (2*(fxx -  this->f - gtd));

	double arg1=sigma1*t;
	double arg2=GSL_MIN(sigma2*t,tnew);
	tnew=GSL_MAX(arg1,arg2);
	
	// sufficient decrease condition (Armijo)
      } while(fxx > fx - (alpha/t) * dif2 );
      
      o2scl_arith::vector_copy(this->dim,xx,this->x);
      this->f = fxx;
      
      (*this->dfunc)(this->dim,this->x,this->gradient,*this->param);
      
      return 0;
    }
    
#endif

    public:

    ool_mmin_pgrad() {
      fmin=-1.0e99;
      tol=1.0e-4;
      alpha=1.0e-4;
      sigma1=0.1;
      sigma2=0.9;
      this->ntrial=1000;
    }

    /** 
	\brief Minimum function value (default \f$ 10^{-99} \f$)
	
	If the function value is below this value, then the algorithm
	assumes that the function is not bounded and exits.
    */
    double fmin;
    /// Tolerance on infinite norm
    double tol;
    /**
       \brief Constant for the sufficient decrease condition 
       (default \f$ 10^{-4} \f$)
    */
    double alpha;
    /// Lower bound to the step length reduction
    double sigma1;
    /// Upper bound to the step length reduction
    double sigma2;
    
    /// Allocate memory
    virtual int allocate(const size_t n) {
      if (this->dim>0) free();
      this->ao.allocate(xx,n);
      return ool_constr_mmin<param_t,func_t,dfunc_t,hfunc_t,vec_t,alloc_vec_t,
	alloc_t>::allocate(n);
    }
    
    /// Free previously allocated memory
    virtual int free() {
      if (this->dim>0) this->ao.free(xx);
      return ool_constr_mmin<param_t,func_t,dfunc_t,hfunc_t,vec_t,alloc_vec_t,
      alloc_t>::free();
    }
    
    /// Set the function, the initial guess, and the parameters
    virtual int set(func_t &fn, dfunc_t &dfn, vec_t &init, param_t &par) {
      
      int ret=ool_constr_mmin<param_t,func_t,dfunc_t,hfunc_t,vec_t,
      alloc_vec_t,alloc_t>::set(fn,dfn,init,par);
      
      // Turn initial guess into a feasible point
      proj(this->x);

      // Evaluate function and gradient
      (*this->func)(this->dim,this->x,this->f,*this->param);
      (*this->dfunc)(this->dim,this->x,this->gradient,*this->param);
      
      return 0;
    }

    /// Restart the minimizer
    virtual int restart() {
      // Turn x into a feasible point
      proj(this->x);

      // Evaluate function and gradient
      (*this->func)(this->dim,this->x,this->f,*this->param);
      (*this->dfunc)(this->dim,this->x,this->gradient,*this->param);

      return 0;
    }

    /// Perform an iteration
    virtual int iterate() {
      double t;

      line_search();
      
      /* Infinite norm of g1 = d <- [P(x - g) - x] */
      o2scl_arith::vector_copy(this->dim,this->x,xx);
      for(size_t i=0;i<this->dim;i++) {
	xx[i]-=this->gradient[i];
      }
      proj(xx);
      for(size_t i=0;i<this->dim;i++) xx[i]-=this->x[i];
      
      double maxval=fabs(xx[0]);
      for(size_t i=1;i<this->dim;i++) {
	if (fabs(xx[i])>maxval) {
	  maxval=fabs(xx[i]);
	}
      }
      this->size=fabs(maxval);

      return GSL_SUCCESS;
    }

    /// See if we're finished
    virtual int is_optimal() {
      if (this->size>tol && this->f>fmin) {
	return GSL_CONTINUE;
      }
      return GSL_SUCCESS;
    }

    /// Return string denoting type ("ool_mmin_pgrad")
    const char *type() { return "ool_mmin_pgrad"; }
      
  };
  
#ifndef DOXYGENP
}
#endif

#endif

