/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_ADAPT_STEP_H
#define O2SCL_ADAPT_STEP_H

#include <o2scl/odestep.h>
#include <o2scl/gsl_rkck.h>
#include <o2scl/ode_funct.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Adaptive stepper base
      
      The adaptive stepper routines are based on several applications
      of ordinary ODE steppers (implemented in \ref odestep). Each
      adaptive stepper (gsl_astep or nonadapt_step) can be used with
      any of the ODE stepper classes (e.g. gsl_rkck). By default,
      gsl_rkck is used. To modify the ODE stepper which is used, use
      the member function set_step() documented below.

      \note If you use gsl_rkck_fast or gsl_rk8pd_fast, you'll 
      need to make sure that the argument \c n to astep() or
      astep_derivs() below matches the template size parameter 
      given in the ODE stepper. 
  */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_view, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class adapt_step {
    
      public:
    
      adapt_step() {
	stepp=&def_step;
	verbose=0;
      }

      virtual ~adapt_step() {}
      
      /** 
	  \brief Make an adaptive integration step of the system 
	  \c derivs 
	  
	  This attempts to take a step of size \c h from the point \c
	  x of an \c n-dimensional system \c derivs starting with \c
	  y. On exit, \c x and \c y contain the new values at the end
	  of the step, \c h contains the size of the step.
      */
      virtual int astep(double &x, double &h, double xmax, size_t n, 
			vec_t &y, param_t &pa, func_t &derivs) 
      {
	
	set_err_ret("Empty function in adapt_step::astep().",
		    o2scl::gsl_nobase);
      }

      /** 
	  \brief Make an adaptive integration step of the system 
	  \c derivs with derivatives
	  
	  This attempts to take a step of size \c h from the point \c
	  x of an \c n-dimensional system \c derivs starting with \c y
	  and given the initial derivatives \c dydx. On exit, \c x, \c
	  y and \c dydx contain the new values at the end of the step,
	  \c h contains the size of the step.
      */
      virtual int astep_derivs(double &x, double &h, double xmax, size_t n, 
			       vec_t &y, vec_t &dydx, param_t &pa, 
			       func_t &derivs) 
      {
	
	set_err_ret("Empty function in adapt_step::astep_derivs().",
		    o2scl::gsl_nobase);
      }
    
      /// Set output level
      int verbose;

      /** 
	  \brief Set stepper

	  This sets the stepper for use in the adaptive step 
	  routine. If no stepper is specified, then the default
	  (gsl_rkck) is used.
      */
      int set_step(odestep<param_t,func_t,vec_t> &step) {
	stepp=&step;
	return 0;
      }
  
      /// The default stepper
      gsl_rkck<param_t,func_t,vec_t,alloc_vec_t,alloc_t> def_step;
      
#ifndef DOXYGEN_INTERNAL

    protected:
      
      /// Pointer to the stepper being used
      odestep<param_t,func_t,vec_t> *stepp;
    
#endif

    };

#ifndef DOXYGENP
}
#endif

#endif
