/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/

#ifndef O2SCL_NONADAPT_STEP_H
#define O2SCL_NONADAPT_STEP_H

#include <o2scl/adapt_step.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_odeiv.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief An non-adaptive stepper implementation of \ref adapt_step.

      This class simply calls the specified ODE stepper without any
      attempt to modify the size of the step, and is primarily useful
      to allow for simple comparisons between adaptive and
      non-adaptive solution. To modify the ODE stepper which is used,
      use the adapt_step::set_step().
      
      \future Modify so that memory allocation/deallocation is only
      performed when necessary
   */
  template<class param_t, class func_t=ode_funct<param_t>, 
    class vec_t=ovector_view, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class nonadapt_step : 
  public adapt_step<param_t,func_t,vec_t,alloc_vec_t,alloc_t> {
    
    public:
    
    nonadapt_step()  {
    }
      
    virtual ~nonadapt_step() {}

    /** \brief Make an adaptive integration step of the system 
	\c derivs 
	  
	This attempts to take a step of size \c h from the point \c x
	of an \c n-dimensional system \c derivs starting with \c y. On
	exit, \c x and \c y contain the new values at the end of the
	step and \c h contains the size of the step. 
    */
    virtual int astep(double &x, double &h, double xmax, size_t n, 
		      vec_t &y, param_t &pa, func_t &derivs) 
    {
      alloc_vec_t dydx, yerr;

      ao.allocate(dydx,n);
      ao.allocate(yerr,n);
	
      derivs(x,n,y,dydx,pa);
      int ret=this->stepp->step(x,h,n,y,dydx,y,yerr,dydx,pa,derivs);
      x+=h;

      ao.free(dydx);
      ao.free(yerr);

      return ret;
    }
      
    /** \brief Make an adaptive integration step of the system 
	\c derivs 
	  
	This attempts to take a step of size \c h from the point \c x
	of an \c n-dimensional system \c derivs starting with \c y. On
	exit, \c x and \c y contain the new values at the end of the
	step and \c h contains the size of the step. 
    */
    virtual int astep_derivs(double &x, double &h, double xmax, size_t n, 
			     vec_t &y, vec_t &dydx, param_t &pa, 
			     func_t &derivs) 
    {
      alloc_vec_t yerr;

      ao.allocate(yerr,n);
	
      int ret=this->stepp->step(x,h,n,y,dydx,y,yerr,dydx,pa,derivs);
      x+=h;

      ao.free(yerr);

      return ret;
    }
      
#ifndef DOXYGEN_INTERNAL

    protected:

    /// Memory allocator for objects of type \c alloc_vec_t
    alloc_t ao;
      
#endif
      
  };
  
#ifndef DOXYGENP
}
#endif

#endif
