/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_ODE_IV_SOLVE_H
#define O2SCL_ODE_IV_SOLVE_H

#include <string>
#include <o2scl/collection.h>
#include <o2scl/adapt_step.h>
#include <o2scl/gsl_astep.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Solve an initial-value ODE problems given an 
      adaptive ODE stepper

      \todo (10/8/07) It's not clear to me that \c astepper shouldn't
      be a member pointer instead of having the type \c adapt_step be
      a template parameter. The present code certainly works, but
      prevents the user from changing the adaptive stepper for the ODE
      solver at runtime.
      
      \todo Decide what to do if the adaptive stepper fails. (1/18/07: 
      Two approaches: continue no matter what, or just stop. 10/8/07:
      should probably let the class user decide?)

      \future Convert to using \c astep_derivs()?

      \future Consider modifying so that this can handle tables
      which are two small by removing half the rows and doubling
      the stepsize.
      
  */
  template<class param_t, class func_t=ode_funct<void *>, 
    class vec_t=ovector_view, 
    class alloc_vec_t=ovector, class alloc_t=ovector_alloc, 
    class adapt_step=gsl_astep<param_t,func_t,vec_t,alloc_vec_t,alloc_t> > 
    class ode_iv_solve 
    
    {
      public:
      
      ode_iv_solve() {
	verbose=0;
	ntrial=1000;
      }
      
      virtual ~ode_iv_solve() {}
      
      /** 
	  \brief Solve the initial-value problem and output a table 

	  Initially, \c xsol should be a vector of size \c nsol, and
	  \c ysol should be a two-dimensional array
	  (i.e. omatrix_view) of size \c [nsol][n]. On exit, \c nsol
	  will will be the size of the solution table, less than or
	  equal to the original value of \c nsol.
      */
      template<class mat_t>
      int solve_table(double x0, double x1, double h, size_t n,
		      vec_t &ystart, size_t &nsol, vec_t &xsol, 
		      mat_t &ysol, param_t &pa, func_t &derivs) 
      {
	int ret=0, nmax=nsol, i;
	size_t j;
  
	xsol[0]=x0;
	for(j=0;j<n;j++) ysol[0][j]=ystart[j];
	if (verbose>0) print_iter(xsol[0],n,ystart);
  
	for(i=1;i<nmax && xsol[i-1]<x1 && ret==0;i++) {
    
	  xsol[i]=xsol[i-1];
	  for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];
    
	  omatrix_row ar(ysol,i);
	  ret=astepper.astep(xsol[i],h,x1,n,ar,pa,derivs);
    
	  if (verbose>0) print_iter(xsol[i],n,ar);
	}

	nsol=i;

	if (ret!=0) return ret;

	if (i==nmax && xsol[i-1]<x1) {
	  set_err("Ran out of space in ode_iv_solve::solve_table().",
		  gsl_etable);
	  return gsl_etable;
	}
	return 0;
      }

      /** \brief Solve the initial-value problem from \c x0 
	  to \c x1 over a grid 

	  Initially, \c xsol should be an array of size \c nsol, and \c
	  ysol should be a \c omatrix of size \c [nsol][n].

	  This function never takes a step larger than the grid
	  size. This could cause inaccuracy if the grid size is too
	  fine.
      */
      template<class mat_t>
      int solve_grid(double x0, double x1, double h, size_t n,
		     vec_t &ystart, size_t nsol, vec_t &xsol, mat_t &ysol,
		     param_t &pa, func_t &derivs) 
      {
	double x=x0, xnext;
	int ret=0;
	size_t j;

	xsol[0]=x0;
	for(j=0;j<n;j++) ysol[0][j]=ystart[j];
	if (verbose>0) print_iter(xsol[0],n,ystart);

	if (x0<x1) {
	  for(size_t i=1;i<nsol && ret==0;i++) {
	    
	    xsol[i]=xsol[i-1];
	    for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];
	    
	    xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
	    
	    while(xsol[i]<xnext && ret==0) {
	      omatrix_row ar(ysol,i);
	      ret=astepper.astep(xsol[i],h,xnext,n,ar,pa,derivs);
	      if (verbose>0) print_iter(xsol[i],n,ar);
	    }
	    
	  }
	} else {
	  for(size_t i=1;i<nsol && ret==0;i++) {
	    
	    xsol[i]=xsol[i-1];
	    for(j=0;j<n;j++) ysol[i][j]=ysol[i-1][j];
	    
	    xnext=x0+(x1-x0)*((double)i)/((double)(nsol-1));
	    
	    while(xsol[i]>xnext && ret==0) {
	      omatrix_row ar(ysol,i);
	      ret=astepper.astep(xsol[i],h,xnext,n,ar,pa,derivs);
	      if (verbose>0) print_iter(xsol[i],n,ar);
	    }
	    
	  }
	}
	if (ret!=0) return ret;
	return 0;
      }

      /** \brief Solve the initial-value problem 
	  to get the final value 

	  For a partiular adaptive stepper, this will likely be the
	  fastest approach, but it provides no information about 
	  the solution other than the final value at x=x1.
      */
      int solve_final_value(double x0, double x1, double h, size_t n,
			    vec_t &ystart, vec_t &yend, 
			    param_t &pa, func_t &derivs) 
      {
	if ((x1>x0 && h<=0.0) || (x0>x1 && h>=0.0)) {
	  set_err_ret("Interval ordering and sign of h don't match.",
		      gsl_einval);
	}

	double x=x0;
	int ret=0, it=0;
	if (verbose>0) print_iter(x0,n,ystart);
	
	for(size_t i=0;i<n;i++) yend[i]=ystart[i];
	
	if (x1>x0) {
	  while(x<x1 && ret==0) {
	    ret=astepper.astep(x,h,x1,n,yend,pa,derivs);
	    if (verbose>0) print_iter(x,n,yend);
	    it++;
	    if (it>ntrial) {
	      set_err_ret("Exceed maximum iterations.",gsl_emaxiter);
	    }
	  }
	} else {

	  while(x>x1 && ret==0) {
	    ret=astepper.astep(x,h,x1,n,yend,pa,derivs);
	    if (verbose>0) print_iter(x,n,yend);
	    it++;
	    if (it>ntrial) {
	      set_err_ret("Exceed maximum iterations.",gsl_emaxiter);
	    }
	  }
	}

	if (ret!=0) return ret;
	return 0;
      }
      
      /// Set output level
      int verbose;
      
      /// Maximum number of steps for solve_final_value() (default 1000)
      int ntrial;

      /// The adaptive stepper utilized
      adapt_step astepper;

      /// Return the type, \c "ode_iv_solve".
      virtual const char *type() { return "ode_iv_solve"; }
      
#ifndef DOXYGEN_INTERNAL

      protected:
      
      /// Print out iteration information
      virtual int print_iter(double x, size_t nv, vec_t &y) {
	std::cout << type() << " x: " << x << " y: ";
	for(size_t i=0;i<nv;i++) std::cout << y[i] << " ";
	std::cout << std::endl;
	if (verbose>1) {
	  char ch;
	  std::cin >> ch;
	}
	return 0;
      }
  
#endif

    };

#ifndef DOXYGENP
}
#endif

#endif
