
/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_PERMUTATION_H
#define O2SCL_PERMUTATION_H

/** \file permutation.h
    \brief File for definitions of vectors
*/

#include <iostream>
#include <cstdlib>
#include <string>
#include <fstream>
#include <sstream>
#include <vector>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_permutation.h>
#include <o2scl/err_hnd.h>
#include <o2scl/string_conv.h>
#include <o2scl/uvector_tlate.h>
#include <o2scl/array.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief A permutation
  */
  class permutation : public gsl_permutation_struct {
    
  public:
    
    permutation(size_t dim=0) {
      gsl_permutation_struct::size=dim;
      if (dim>0) {
	data=(size_t *)malloc(dim*sizeof(size_t));
      }
    }

    /// \name Copy constructors
    //@{
    permutation(const permutation &v) {
      gsl_permutation_struct::size=v.size();
      data=new size_t[gsl_permutation_struct::size];
      for(size_t i=0;i<gsl_permutation_struct::size;i++) {
	data[i]=v.data[i];
      }
    }
    
    permutation& operator=(const permutation &v) {
      gsl_permutation_struct::size=v.size();
      data=new size_t[gsl_permutation_struct::size];
      for(size_t i=0;i<gsl_permutation_struct::size;i++) {
	data[i]=v.data[i];
      }
      return *this;
    }
    //@}

    ~permutation() {
      if (gsl_permutation_struct::size>0) {
	std::free(data);
	gsl_permutation_struct::size=0;
      }
    };

    /** \brief Get (with optional range-checking) */
    size_t get(size_t i) const {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=gsl_permutation_struct::size) {
	set_err((((std::string)"Permutation index ")+itos(i)+" out of bounds"
		 +" in permutation::get(). Size: "+
		 itos(gsl_permutation_struct::size)+
		 " (index should be less than size).").c_str(),gsl_index);
	return 0;
      }
#endif
      return data[i];
    }
    
    /** \brief Set (with optional range-checking) */
    int set(size_t i, size_t val) {
#if O2SCL_NO_RANGE_CHECK
#else
      if (i>=gsl_permutation_struct::size) {
	set_err_ret((((std::string)"Permutation index ")+itos(i)+
		     " out of bounds"+" in permutation::set(). Size: "+
		     itos(gsl_permutation_struct::size)+
		     " (index should be less than size).").c_str(),gsl_index);
      }
#endif
      data[i]=val;
      return 0;
    }

    /// Initialize permutation to the identity
    int init() {
      for(size_t i=0;i<gsl_permutation_struct::size;i++) data[i]=i;
      return 0;
    }

    /** 
	\brief Return permutation size 
	
	If no memory has been allocated, this will quietly 
	return zero.
    */
    size_t size() const {
      return gsl_permutation_struct::size;
    }

    /** \brief Allocate memory for a permutation of size \c dim
    */
    int allocate(size_t dim) {
      if (gsl_permutation_struct::size!=dim && 
	  gsl_permutation_struct::size>0) free();
      gsl_permutation_struct::size=dim;
      if (dim>0) {
	data=(size_t *)malloc(dim*sizeof(size_t));
      }
      return 0;
    }

    /** 
	\brief Free the memory 
	
	This function will safely do nothing if used without first
	allocating memory or if called multiple times in succession.
    */
    int free() {
      if (gsl_permutation_struct::size>0) {
	std::free(data);
	gsl_permutation_struct::size=0;
      }
      return 0;
    }
    //@}

    /// Swap two elements of a permutation
    int swap(const size_t i, const size_t j) {
      size_t tmp=data[i];
      data[i]=data[j];
      data[j]=tmp;
      return 0;
    }

    /// Check to see that a permutation is valid
    bool valid() {
      for(size_t i=0;i<gsl_permutation_struct::size;i++) {
	if (data[i]>gsl_permutation_struct::size) return false;
	for(size_t j=0;j<i;j++) {
	  if (data[i]==data[j]) return false;
	}
      }
      return true;
    }

    /// Reverse the permutation
    int reverse() {
      size_t i;
      for (i = 0; i < (gsl_permutation_struct::size / 2); i++){
	size_t j = gsl_permutation_struct::size - i - 1;
	
	size_t tmp = this->data[i] ;
	this->data[i] = this->data[j] ;
	this->data[j] = tmp ;
      }
      return 0;
    }

    /// Compute the inverse of a permutation
    permutation inverse() {
      permutation p(gsl_permutation_struct::size);
      for(size_t i=0;i<gsl_permutation_struct::size;i++) {
	p.data[data[i]]=i;
      }
      return p;
    }
    
    /// Apply the permutation to a vector
    template<class vec_t> int apply(vec_t &v) {
      size_t i, k, pk;
      for(i=0;i<gsl_permutation_struct::size;i++) {
	k=data[i];
	while (k>i) k=data[k];
	if (k<i) continue;
	/// Now have k==i, i.e. the least in its cycle
	pk=data[k];
	if (pk==i) continue;
	// Shuffle the elements of the cycle
	{
	  double t=v[i];
	  while (pk!=i) {
	    double r1=v[pk];
	    v[k]=r1;
	    k=pk;
	    pk=data[k];
	  }
	  v[k]=t;
	}
      }
      return 0;
    }

    /// Apply the permutation to a vector
    template<class vec_t> int apply_inverse(vec_t &v) {
      size_t i, k, pk;
      for(i=0;i<gsl_permutation_struct::size;i++) {
	k=data[i];
	while (k>i) k=data[k];
	if (k<i) continue;
	/// Now have k==i, i.e. the least in its cycle
	pk=data[k];
	if (pk==i) continue;
	// Shuffle the elements of the cycle
	{
	  double t=v[k];
	  while (pk!=i) {
	    double r1=v[pk];
	    v[pk]=t;
	    t=r1;
	    k=pk;
	    pk=data[k];
	  }
	  v[pk]=t;
	}
      }
      return 0;
    }

  };

#ifndef DOXYGENP
}
#endif

#endif

