/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_QUAD_INTP_H
#define O2SCL_QUAD_INTP_H

#include <iostream>
#include <string>
#include <cmath>
#include <map>
#include <gsl/gsl_linalg.h>
#include <o2scl/err_hnd.h>
#include <o2scl/ovector_tlate.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Interpolate a function of two independent variables with 
      a quadratic polynomial
      
      This is a "conic-section" interpolation for two
      dimensions, using the function
      \f[
      z(x,y) = a_1 x^2 + a_2 x y + a_3 y^2 + a_4 x + a_5 y + a_6
      \f]
      For a set of data \f${x_i,y_i,z_i}\f$, a value of z is predicted
      given a value of x and y. This is done by finding the conic
      section obeying the above relation that goes through six closest
      points in the data set.
      
      This procedure does not always succeed. It fails when the 
      6 closest points are somehow degenerate, for example, if they
      are all colinear.
      
      The vector and matrix types can be any types which have
      suitably defined functions \c operator[]. 

      There is no caching so the numeric values of the data may be
      freely changed between calls to interp().

      \bug This class doesn't seem to work at present.
  */
  template<class vec_t, class mat_t> class quad_intp {
  public:
    
    quad_intp() {
      data_set=false;
    }
    
    typedef struct {
      int i;
      double c;
    } point;
    
    int compare(const void *x, const void *y) {
      point *p=(point *)x;
      point *q=(point *)y;
      return (p->c)>(q->c);
    }
    
    /** 
	\brief Initialize the data for the quad interpolation
    */
    int set_data(size_t n_points, vec_t &x, vec_t &y,
		 size_t n_dat, mat_t &dat) {
      if (n_points<6) {
	set_err_ret("Must provide at least six points in set_data()",
		    gsl_efailed);
      }
      np=n_points;
      nd=n_dat;
      ux=&x;
      uy=&y;
      udat=&dat;
      data_set=true;
      return 0;
    }
    
    /** 
	\brief Perform the quadratic interpolation 
	
	It is assumed that \c ip is properly allocated beforehand.
    */
    int interp(double x, double y, vec_t &ip) {
      int i, j, mat_ret=0;
  
      if (data_set==false) {
	set_err("Data not set in quad_intp::interp().",gsl_einval);
	return gsl_einval;
      }
      if (ip.is_owner()==false) {
	set_err("Provided ovector does not own its data. Proceeding anyway.",
		gsl_einval);
      }
      
      std::map<double,int> pmap;
      for (i=0;i<np;i++) {
	pmap.insert(make_pair(sqrt(pow(x-(*ux)[i],2.0)+
				   pow(y-(*uy)[i],2.0)),i));
      }
      std::map<double,int>::iterator pit;

      gsl_matrix *gm=gsl_matrix_alloc(6,6);
      gsl_vector *gx=gsl_vector_alloc(6), *gb=gsl_vector_alloc(6);
      
      for(j=0;j<nd;j++) {
	
	pit=pmap.begin();
	for(i=0;i<6;i++) {
	  gsl_matrix_set(gm,i,0,(*ux)[pit->second]*(*ux)[pit->second]);
	  gsl_matrix_set(gm,i,1,(*ux)[pit->second]*(*uy)[pit->second]);
	  gsl_matrix_set(gm,i,2,(*uy)[pit->second]*(*uy)[pit->second]);
	  gsl_matrix_set(gm,i,3,(*ux)[pit->second]);
	  gsl_matrix_set(gm,i,4,(*uy)[pit->second]);
	  gsl_matrix_set(gm,i,5,1.0);
	  gsl_vector_set(gb,i,(*(udat[j]))[i]);
	  pit++;
	}
	int ret=gsl_linalg_HH_solve(gm,gb,gx);
	if (ret!=0) mat_ret=ret;
	ip[j]=gsl_vector_get(gx,0)*x*x+
	  gsl_vector_get(gx,1)*x*y+
	  gsl_vector_get(gx,2)*y*y+
	  gsl_vector_get(gx,3)*x+
	  gsl_vector_get(gx,4)*y+
	  gsl_vector_get(gx,5);
      }
  
      if (mat_ret!=0) {
	add_err("Matrix solver failed in quad_intp::interp().",
		gsl_efailed);
      }

      gsl_matrix_free(gm);
      gsl_vector_free(gx);
      gsl_vector_free(gb);
      
      return mat_ret;
    }
    
#ifndef DOXYGEN_INTERNAL

  protected:

    /// The number of grid points
    int np;
    /// The number of functions
    int nd;
    /// The x-values
    vec_t *ux;
    /// The y-values
    vec_t *uy;
    /// The data
    mat_t *udat;
    /// True if the data has been given by the user
    bool data_set;
    
    // Swap points 1 and 2
    int swap(int &i1, double &c1, int &i2, double &c2);
    
#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif



