/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <o2scl/twod_intp.h>

using namespace std;
using namespace o2scl;

twod_intp::twod_intp() {
  nuit=0;
  it_set=false;
  data_set=false;
}

twod_intp::~twod_intp() {
  if (data_set) {
    int i;
    if (xfirst==true) {
      for(i=0;i<ny;i++)	{
	delete si[i];
	delete ari[i];
      }
      delete[] si;
      delete[] ari;
    } else {
      for(i=0;i<nx;i++)	{
	delete si[i];
	delete aci[i];
      }
      delete[] si;
      delete[] aci;
    }
  }
}

int twod_intp::set_data(int n_x, int n_y, ovector &x_fun,
			ovector &y_fun, omatrix &u_data, 
			bool x_first) {
  int i, j;
  
  // Free interpolation objects from previous data
  if (data_set) {
    if (xfirst==true) {
      for(i=0;i<ny;i++)	{
	delete si[i];
	delete ari[i];
      }
      delete[] si;
      delete[] ari;
    } else {
      for(i=0;i<nx;i++)	{
	delete si[i];
	delete aci[i];
      }
      delete[] si;
      delete[] aci;
    }
  }

  // Set new data
  xfirst=x_first;
  nx=n_x;
  ny=n_y;
  xfun=&x_fun;
  yfun=&y_fun;
  data=&u_data;

  // Allocate new interpolation objects
  if (x_first==true) {
    si=new sm_interp_vec *[ny];
    ari=new omatrix_row *[ny];
    for(i=0;i<ny;i++) {
      ari[i]=new omatrix_row(*data,i);
      if (it_set && i<nuit) {
	si[i]=new sm_interp_vec(uit1[i],uit2[i],nx,*xfun,*(ari[i]));
      } else {
	si[i]=new sm_interp_vec(nx,*xfun,*(ari[i]));
      }
    }
  } else {
    si=new sm_interp_vec *[nx];
    aci=new omatrix_col *[nx];
    for(i=0;i<nx;i++) {
      aci[i]=new omatrix_col(*data,i);
      if (it_set && i<nuit) {
	si[i]=new sm_interp_vec(uit1[i],uit2[i],ny,*yfun,*aci[i]);
      } else {
	si[i]=new sm_interp_vec(ny,*yfun,*aci[i]);
      }
    }
  }
  data_set=true;
  return 0;
}

int twod_intp::reset_interp() {
  if (data_set) {
    int i, j;

    if (xfirst==true) {
      for(i=0;i<ny;i++)	{
	delete si[i];
	delete ari[i];
      }
      delete[] si;
      delete[] ari;
    } else {
      for(i=0;i<nx;i++)	{
	delete si[i];
	delete aci[i];
      }
      delete[] si;
      delete[] aci;
    }

    // Allocate new interpolation objects
    if (xfirst==true) {
      si=new sm_interp_vec *[ny];
      ari=new omatrix_row *[ny];
      for(i=0;i<ny;i++) {
	ari[i]=new omatrix_row(*data,i);
	if (it_set && i<nuit) {
	  si[i]=new sm_interp_vec(uit1[i],uit2[i],nx,*xfun,*ari[i]);
	} else {
	  si[i]=new sm_interp_vec(nx,*xfun,*ari[i]);
	}
      }
    } else {
      si=new sm_interp_vec *[nx];
      aci=new omatrix_col *[nx];
      for(i=0;i<nx;i++) {
	aci[i]=new omatrix_col(*data,i);
	if (it_set && i<nuit) {
	  si[i]=new sm_interp_vec(uit1[i],uit2[i],ny,*yfun,*aci[i]);
	} else {
	  si[i]=new sm_interp_vec(ny,*yfun,*aci[i]);
	}
      }
    }
  } else {
    set_err("Data not set in twod_intp::reset_interp().",gsl_einval);
    return gsl_einval;
  }
  return 0;
}

double twod_intp::interp(double x, double y) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

double twod_intp::deriv_x(double x, double y) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->deriv(x);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.deriv(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.deriv(x);
    }
  }
  return result;
}

double twod_intp::deriv_xy(double x, double y) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->deriv(x);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.deriv(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.deriv(y);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->deriv(y);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.deriv(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.deriv(x);
    }
  }
  return result;
}

double twod_intp::deriv_y(double x, double y) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.deriv(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.deriv(y);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->deriv(y);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

double twod_intp::deriv2_x(double x, double y) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->deriv2(x);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.deriv2(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.deriv2(x);
    }
  }
  return result;
}

double twod_intp::deriv2_y(double x, double y) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.deriv2(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.deriv2(y);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->deriv2(y);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

double twod_intp::integ_x(double x0, double x1, double y) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->integ(x0,x1);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.interp(y);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.interp(y);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->interp(y);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.integ(x0,x1);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.integ(x0,x1);
    }
  }
  return result;
}

double twod_intp::integ_y(double x, double y0, double y1) {
  if (data_set==false) {
    set_err("Data not set in twod_intp::interp().",1);
    return 0.0;
  }
  double result;
  if (xfirst==true) {
    ovector icol(ny);
    for(int i=0;i<ny;i++) {
      icol[i]=si[i]->interp(x);
    }
    if (it_set) {
      sm_interp_vec siy(*uit3,*uit4,ny,*yfun,icol);
      result=siy.integ(y0,y1);
    } else {
      sm_interp_vec siy(ny,*yfun,icol);
      result=siy.integ(y0,y1);
    }
  } else {
    ovector icol(nx);
    for(int i=0;i<nx;i++) {
      icol[i]=si[i]->integ(y0,y1);
    }
    if (it_set) {
      sm_interp_vec six(*uit3,*uit4,nx,*xfun,icol);
      result=six.interp(x);
    } else {
      sm_interp_vec six(nx,*xfun,icol);
      result=six.interp(x);
    }
  }
  return result;
}

