/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_PART_H
#define O2SCL_PART_H

#include <string>
#include <iostream>
#include <cmath>
#include <o2scl/constants.h>
#include <o2scl/inte.h>
#include <o2scl/collection.h>
#include <o2scl/funct.h>
#include <o2scl/mroot.h>

/** \file part.h
    \brief File for definitions for \ref thermo and \ref part 
*/

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** \brief A class for the thermodynamical variables (energy
      density, pressure, entropy density)
  */
  class thermo {

  public:

    /// pressure
    double pr;
    /// energy density
    double ed;
    /// entropy density
    double en;

    /// Return string denoting type ("thermo")
    const char *type() { return "thermo"; }

  };

  /** \brief Addition operator
  */
  extern thermo operator+(const thermo &left, const thermo &right);

  /** \brief Subtraction operator
  */
  extern thermo operator-(const thermo &left, const thermo &right);

  template<> int io_tlate<thermo>::input
    (cinput *co, in_file_format *ins, thermo *t);
  template<> int io_tlate<thermo>::output
    (coutput *co, out_file_format *outs, thermo *t);
  template<> const char *io_tlate<thermo>::type();

  /** 
      \brief Particle base class 

      Calculate the properties of particles from their chemical
      potential (calc_mu() and pair_mu()) or from the density
      (calc_density() and pair_density()).

      When non-interacting is false, the thermodynamic integrals need
      both a value of "mu" and "nu". "nu" is an effective chemical
      potential which appears in the argument of the exponential of
      the Fermi-function.

      Keep in mind, that the pair functions use anti(), which assumes
      that nu -> -nu and mu -> -mu for the anti-particles, which might
      not be true for interacting particles.  When non-interacting is
      true, then "ms" is set equal to "m", and "nu" is set equal to
      "mu", everywhere.
      
      The "density" functions use the value of nu (or mu when
      non_interacting is true) for an initial guess. Zero is very
      likely a bad guess, but these functions will not warn you about
      this.

  */
  class part {

  public:

    /// degeneracy
    double g;
    /// mass
    double m;
    /// density
    double n;
    /// energy density
    double ed;
    /// pressure
    double pr;
    /// chemical potential
    double mu;
    /// entropy
    double en;
    /// effective mass (Dirac unless otherwise specified)
    double ms;
    /// effective chemical potential
    double nu;
    /// derivative of energy with respect to effective mass
    bool inc_rest_mass;
    /** \brief When this is true, \ref nu and \ref ms are set equal
	to \ref mu and \ref m by \ref calc_mu(), etc.. (default true)
    */
    bool non_interacting;
    /// The name usually defaults to the class name.
    std::string name;
    
    /// make a particle of mass \c m and degeneracy \c g.
    part(double m=0.0, double g=0.0);
    virtual ~part();
  
    /// Set the mass \c m and degeneracy \c g.
    virtual int init(double m, double g);

    /** 
	\brief Calculate properties as function of chemical potential
     */
    virtual int calc_mu(const double temper);

    /** \brief Calculate properties as function of density
     */
    virtual int calc_density(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	chemical potential
    */
    virtual int pair_mu(const double temper);

    /** \brief Calculate properties with antiparticles as function of
	density
    */
    virtual int pair_density(const double temper);
  
    /// Make an anti-particle
    virtual int anti(part &ax);

    /// Return string denoting type ("part")
    virtual const char *type() { return "part"; }
    
    friend class io_tlate<part>;
    
  };

  /** \brief Addition operator
  */
  extern thermo operator+(const thermo &left, const part &right);

  /** \brief Subtraction operator
  */
  extern thermo operator-(const thermo &left, const part &right);

  template<> int io_tlate<part>::input
    (cinput *co, in_file_format *ins, part *p);
  template<> int io_tlate<part>::output
    (coutput *co, out_file_format *outs, part *p);
  template<> const char *io_tlate<part>::type();

  //---------------------------------------------

  typedef io_tlate<thermo> thermo_io_type;
  typedef io_tlate<part> part_io_type;

#ifndef DOXYGENP
}
#endif

#endif
