/*
  -------------------------------------------------------------------
  
  Copyright (C) 2006, 2007, 2008, Andrew W. Steiner
  
  This file is part of O2scl.
  
  O2scl is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.
  
  O2scl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with O2scl. If not, see <http://www.gnu.org/licenses/>.

  -------------------------------------------------------------------
*/
#ifndef O2SCL_JACOBIAN_H
#define O2SCL_JACOBIAN_H

#include <string>
#include <o2scl/collection.h>
#include <o2scl/omatrix_tlate.h>
#include <o2scl/mm_funct.h>
#include <o2scl/gsl_deriv.h>

#ifndef DOXYGENP
namespace o2scl {
#endif

  /** 
      \brief Base for a square Jacobian where J is computed at x given y=f(x)

      Compute
      \f[
      J_(i,j) = \frac{\partial f_i}{\partial x_j}
      \f]

      The \c vec_t objects in operator() could have been written to be
      \c const, but they are not \c const so that they can be
      used as temporary workspace. They are restored to their 
      original values before operator() exits.

  */
  template<class param_t, class vec_t=ovector_view, 
    class mat_t=omatrix_view> class jac_funct {

    public:  

    jac_funct() {}

    virtual ~jac_funct() {}
    
    /** \brief The operator()
     */
    virtual int operator()(size_t nv, vec_t &x, 
			   vec_t &y, mat_t &j, param_t &pa) {
      set_err_ret("Missing base in mm_funct::operator().",gsl_nobase);
    }
    
#ifndef DOXYGENP

    private:

    jac_funct(const jac_funct &);
    jac_funct& operator=(const jac_funct&);

#endif

  };

  /** \brief Function pointer to jacobian
   */
  template<class param_t, class vec_t=ovector_view, 
    class mat_t=omatrix_view> class jac_funct_fptr : 
  public jac_funct<param_t,vec_t,mat_t> {
    
    public:
    
    /** \brief Specify the function pointer
     */
    jac_funct_fptr(int (*fp)(size_t nv, vec_t &x, vec_t &y, 
			     mat_t &j, param_t &pa)) {
      fptr=fp;
    }
    
    virtual ~jac_funct_fptr();
    
    /** \brief The operator()
     */
    virtual int operator()(size_t nv, vec_t &x, 
			   vec_t &y, mat_t &j, param_t &pa) {
      return fptr(nv,x,y,j,pa);
    }

#ifndef DOXYGENP

    protected:
  
    jac_funct_fptr() {};
    
    /// Function pointer
    int (*fptr)(size_t nv, vec_t &x, vec_t &y,
		mat_t &j, param_t &pa);
    
    private:

    jac_funct_fptr(const jac_funct_fptr &);
    jac_funct_fptr& operator=(const jac_funct_fptr&);

#endif

  };

  /** \brief Member function pointer to a Jacobian
   */
  template <class tclass, class param_t, class vec_t=ovector_view,
    class mat_t=omatrix_view> class jac_funct_mfptr : 
  public jac_funct<param_t,vec_t,mat_t> {

    public:
  
    /** \brief Specify the member function pointer
     */
    jac_funct_mfptr(tclass *tp, int (tclass::*fp)
		    (size_t nv, vec_t &x, vec_t &y,
		     mat_t &j, param_t &pa)) {
      tptr=tp;
      fptr=fp;
    }
  
    virtual ~jac_funct_mfptr() {};
  
    /** \brief The operator()
     */
    virtual int operator()(size_t nv, vec_t &x, 
			   vec_t &y, mat_t &j, param_t &pa) {
      return (*tptr.*fptr)(nv,x,y,j,pa);
    }
  
#ifndef DOXYGEN_INTERNAL

    protected:
  
    /// Member function pointer
    int (tclass::*fptr)(size_t nv, vec_t &x, vec_t &y,
			mat_t &j, param_t &pa);

    /// Class pointer
    tclass *tptr;

#endif

#ifndef DOXYGENP
  
    private:

    jac_funct_mfptr(const jac_funct_mfptr &);
    jac_funct_mfptr& operator=(const jac_funct_mfptr&);

#endif

  };

  /** \brief Base for providing a numerical jacobian
   */
  template<class param_t, class func_t, class vec_t=ovector_view, 
    class mat_t=omatrix_view>
    class jacobian : public jac_funct<param_t,vec_t,mat_t> {
    public:
    
    jacobian() {
    };
    
    virtual ~jacobian() {};
    
    /// Set the function to compute the Jacobian of
    virtual int set_function(func_t &f) {
      func=&f;
      return 0;
    }
    
    /** \brief The operator()
     */
    virtual int operator()(size_t nv, vec_t &x, vec_t &y,
			   mat_t &j, param_t &pa) {
      return 0;
    }
    
#ifndef DOXYGEN_INTERNAL
    
    protected:
    
    /// A pointer to the user-specified function
    func_t *func;
    
    private:
    
    jacobian(const jacobian &);
    jacobian& operator=(const jacobian&);
    
#endif
    
  };
  
  /** 
      \brief Simple automatic Jacobian
      
      This simple routine is nearly equivalent to GSL as given in \c
      multiroots/fdjac.c. It has an additional test to ensure that the
      finite-differencing stepsize does not vanish, and returns an
      error if the Jacobian is singular. To obtain the GSL behavior,
      set \ref epsrel to \c GSL_SQRT_DBL_EPSILON and set epsmin to
      zero. 

      This class does not separately check the vector and matrix sizes
      to ensure they are commensurate. 

      \todo Double check that this class works with arrays
  */
  template<class param_t, class func_t, class vec_t=ovector_view, 
    class mat_t=omatrix_view, class alloc_vec_t=ovector, 
    class alloc_t=ovector_alloc> class simple_jacobian : 
  public jacobian<param_t,func_t,vec_t,mat_t> {

    public:
    
    simple_jacobian() {
      epsrel=1.0e-4;
      epsmin=1.0e-15;
    }
    
    /** \brief The relative stepsize for finite-differencing 
	(default \f$ 10^{-4} \f$ )
    */
    double epsrel;
    
    /// The minimum stepsize (default \f$ 10^{-15} \f$)
    double epsmin;

    /// For memory allocation
    alloc_t ao;

    /** \brief The operator()
     */
    virtual int operator()(size_t nv, vec_t &x, vec_t &y, 
			   mat_t &jac, param_t &pa) {
      
      size_t i,j;
      double h,temp;
      alloc_vec_t f;
      bool success=true;

      ao.allocate(f,nv);
      
      for (j=0;j<nv;j++) {
	
	h=epsrel*fabs(x[j]);
	if (fabs(h)<=epsmin) h=epsrel;
	
	x[j]+=h;
	(*this->func)(nv,x,f,pa);
	x[j]-=h;

	bool nonzero=false;
	for (i=0;i<nv;i++) {
	  temp=(f[i]-y[i])/h;
	  if (temp!=0.0) nonzero=true;
	  jac[i][j]=temp;
	}
	if (nonzero==false) success=false;
      }
      
      ao.free(f);
      
      if (success==false) {
	set_err_ret("At least one row of the Jacobian is zero.",gsl_efailed);
      }
      return 0;
    }
  };

  /** 
      \brief A direct calculation of the jacobian using a \ref deriv object
      
      Note that it is sometimes wasteful to use this Jacobian in a
      root-finding routine and using more approximate Jacobians is
      more efficient. This class is mostly useful for demonstration
      purposes.
  */
  template<class param_t, class func_t, class vec_t=ovector_view, 
    class mat_t=omatrix_view> class exact_jacobian : 
  public jacobian<param_t,func_t,vec_t,mat_t> {
    
    public:
    
    exact_jacobian() {
      def_deriv.h=1.0e-4;
      dptr=&def_deriv;
    }
    
    /** 
	\brief Parameter structure for passing information
	
	This class is primarily useful for specifying derivatives
	for using the jacobian::set_deriv() function.

	\comment
	This type needs to be publicly available so that the
	user can properly specify a base 1-dimensional derivative
	object. 
	\endcomment
    */
    typedef struct {
      /// The number of variables
      size_t nv;
      /// The current x value
      size_t xj;
      /// The current y value
      size_t yi;
      /// The x vector
      vec_t *x;
      /// The y vector
      vec_t *y;
      /// The parameters
      param_t *pa;
    } ej_parms;

    /// The default derivative object
    gsl_deriv<ej_parms,funct<ej_parms> > def_deriv;

    /// Set the derivative object
    int set_deriv(deriv<ej_parms,funct<ej_parms> > &de) {
      dptr=&de;
      return 0;
    }
    
    /** \brief The operator()
     */
    virtual int operator()(size_t nv, vec_t &x, vec_t &y, 
			   mat_t &jac, param_t &pa) {
      size_t i,j;
      double h,temp;

      ej_parms ejp;
      ejp.nv=nv;
      ejp.x=&x;
      ejp.y=&y;
      ejp.pa=&pa;
      
      funct_mfptr<exact_jacobian,ej_parms> dfnp(this,&exact_jacobian::dfn);
      
      for (j=0;j<nv;j++) {
	ejp.xj=j;
	for (i=0;i<nv;i++) {
	  ejp.yi=i;
	  double tmp=(*ejp.x)[j];
	  jac[i][j]=dptr->calc(tmp,ejp,dfnp);
	  (*ejp.x)[j]=tmp;
	}
      }

      return 0;
    }
    

#ifndef DOXYGEN_INTERNAL

    protected:

    /// Pointer to the derivative object
    deriv<ej_parms,funct<ej_parms> > *dptr;
    
    /// Function for the derivative object
    int dfn(double x, double &y, ej_parms &ejp) {
      (*ejp.x)[ejp.xj]=x;
      (*this->func)(ejp.nv,*ejp.x,*ejp.y,*ejp.pa);
      y=(*ejp.y)[ejp.yi];
      return 0;
    }

#endif

  };
  
#ifndef DOXYGENP
}
#endif

#endif
